/******************************************************************************/
/*                                                                            */
/*  Title       : satcalc.c                                                   */
/*  Author      : Manfred Bester, DL5KR                                       */
/*  Date        : 03Mar92                                                     */
/*  Last change : 12Sep94                                                     */
/*                                                                            */
/*  Synopsis    : Auxiliary routines for the satellite tracking program       */
/*                'sattrack'.                                                 */
/*                                                                            */
/*                                                                            */
/*  Copyright (c) 1992, 1993, 1994 Manfred Bester. All Rights Reserved.       */
/*                                                                            */
/*  Permission to use, copy, modify, and distribute this software and its     */
/*  documentation for educational, research and non-profit purposes, without  */
/*  fee, and without a written agreement is hereby granted, provided that the */
/*  above copyright notice and the following three paragraphs appear in all   */
/*  copies.                                                                   */
/*                                                                            */
/*  Permission to incorporate this software into commercial products may be   */
/*  obtained from the author, Dr. Manfred Bester, 1636 M. L. King Jr. Way,    */
/*  Berkeley, CA 94709, USA.                                                  */
/*                                                                            */
/*  IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, */
/*  SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF   */
/*  THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE AUTHOR HAS BEEN ADVISED  */
/*  OF THE POSSIBILITY OF SUCH DAMAGE.                                        */
/*                                                                            */
/*  THE AUTHOR SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT      */
/*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A   */
/*  PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS"      */
/*  BASIS, AND THE AUTHOR HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT, */
/*  UPDATES, ENHANCEMENTS, OR MODIFICATIONS.                                  */
/*                                                                            */
/******************************************************************************/

#include <stdio.h>
#include <math.h>

#include "satglobalsx.h"
#include "sattrack.h"
 
#ifndef SUNOS4
#include <stdlib.h>
#endif

/******************************************************************************/
/*                                                                            */
/*  reduce: reduces number into specified interval (e.g. -PI, +PI)            */
/*                                                                            */
/******************************************************************************/

double reduce(value,rangeMin,rangeMax)

double value, rangeMin, rangeMax;

{
    double range, rangeFrac, fullRanges, retval;

    range     = rangeMax - rangeMin;
    rangeFrac = (rangeMax - value) / range;

    modf(rangeFrac,&fullRanges);

    retval = value + fullRanges * range;

    if (retval > rangeMax)
        retval -= range;

    return(retval);
}

/******************************************************************************/
/*                                                                            */
/* getTrueAnomaly: gets mean and true anomaly                                 */
/*                                                                            */
/******************************************************************************/

void getTrueAnomaly(tmArg)

double tmArg;

{
    getOrbitNumber(tmArg);                          /* also gets mean anomaly */
    getSatPrecession();
    kepler();

    return;
}

/******************************************************************************/
/*                                                                            */
/* getOrbitNumber: gets the mean anomaly and the orbit number                 */
/*                                                                            */
/******************************************************************************/

void getOrbitNumber(tArg)

double tArg;

{
    double dT, dT2, orbFract;

    dT           = tArg - epochDay;
    dT2          = dT * dT;
    curMotion    = epochMeanMotion + 2.0*decayRate*dT + 6.0*decayRateDot*dT2;

    refOrbit     = (double) epochOrbitNum + epochMeanAnomaly / TWOPI;
    curOrbit     = refOrbit + curMotion * dT 
                            - (curMotion - epochMeanMotion) * dT;

    orbitNum     = (long) curOrbit + 1L;

    orbFract     = modf(curOrbit,&dummyd);
    orbitFract   = orbFract * 100.0;
    meanAnomaly  = orbFract * TWOPI;

    return;
}

/******************************************************************************/
/*                                                                            */
/* getSatPrecession: calculates precession of the satellite's orbit           */
/*                                                                            */
/*                   for reference see:                                       */
/*                   "The Satellite Experimenter's Handbook", M. Davidoff,    */
/*                   The ARRL, Newington                                      */
/*                                                                            */
/******************************************************************************/

void getSatPrecession()

{
    double satPrecFact;

    getSemiMajorAxis();
    checkPerigeeHeight();

    satPrecFact = pow((EARTHRADIUS / semiMajorAxis), 3.5) /
                          SQR(1.0 - SQR(eccentricity)) * CDR;

    raanPrec    = -9.95 * satPrecFact * cosInclination;
    perigeePrec =  4.97 * satPrecFact * (5.0 * SQR(cosInclination) - 1.0);

    return;
}
 
/******************************************************************************/
/*                                                                            */
/* kepler: solves Kepler's equation iteratively                               */
/*                                                                            */
/******************************************************************************/

void kepler()
 
{
    register double eccen, eccAnom, meanAnom, diffAnom;

    eccen    = eccentricity;
    meanAnom = meanAnomaly;
    eccAnom  = meanAnomaly;                              /* initial guess     */

    do
    {
        diffAnom = (eccAnom - eccen * sin(eccAnom) - 
                    meanAnom) / (1.0 - eccen * cos(eccAnom));
        eccAnom -= diffAnom;
    }
    while (fabs(diffAnom) >= CASR);                     /* 1 arcsec precision */

    if (fabs(eccAnom - PI) < CASR)
        trueAnomaly = PI;
    else
        trueAnomaly = 2.0 * atan(sqrt((1.0 + eccen) / (1.0 - eccen)) * 
                             tan(eccAnom / 2.0));

    trueAnomaly = reduce(trueAnomaly,ZERO,TWOPI);

    return;
}
 
/******************************************************************************/
/*                                                                            */
/* getSubSatPoint: calculates sub-satellite point                             */
/*                                                                            */
/*  EARTHRADIUS = equatorial radius of the Earth (m)                          */
/*                from "Astronomical Almanac", 1984, p. S7                    */
/*                                                                            */
/*  F,FF,FFF    = flattening factors for the Earth (Hayford's spheroid)       */
/*                from "Astronomical Almanac", 1984, p. S7                    */
/*  F1,2,3      = coefficients for the latitude correction formula            */
/*                from "Explanatory Supplement to the Ephemeris", 1961, p.58  */
/*                                                                            */
/*  REFF0,1,2,3 = coefficient for the Earth's radius correction formula       */
/*                from "Explanatory Supplement to the Ephemeris", 1961, p.58  */
/*                                                                            */
/*  effEarthRad = effective radius of the Earth at the location of the        */
/*                sub-satellite point with respect to sea level (m)           */
/*                                                                            */
/*  for reference see:  a) "Methods of Experimental Physics", Vol. 12C, p.308 */
/*                      b) "Explanatory Supplement to the Ephemeris", 1961,   */
/*                          p.58                                              */
/*                                                                            */
/******************************************************************************/

void getSubSatPoint()

{
    double satPosSSP[3], satPosEquat[3];
    double rSSP, rEquat, b2, b4, b6, lat, lng, arg, effEarthRad;
    int    i;

    for (i = 0; i <= 2; i++)
        satPosSSP[i] = satPosGl[i];

    rSSP = absol(satPosSSP);

    for (i = 0; i <= 2; i++)
        satPosSSP[i] /= rSSP;

    arg = satPosSSP[2];

    if (arg >  1.0) arg =  1.0;
    if (arg < -1.0) arg = -1.0;

    lat = asin(arg);

    b2  = 2.0 * (lat);
    b4  = 4.0 * (lat);
    b6  = 6.0 * (lat);

    satLat = lat + F1 * sin(b2) + F2 * sin(b4) + F3 * sin(b6);
    satLat = lat;

    satPosEquat[0] = satPosSSP[0];
    satPosEquat[1] = satPosSSP[1];
    satPosEquat[2] = 0.0;

    rEquat = absol(satPosEquat);

    for (i = 0; i <= 2; i++)
        satPosEquat[i] /= rEquat;

    lng          = gasTime - atan2(satPosEquat[1],satPosEquat[0]);
    satLong      = reduce(lng,-PI,PI);

    effEarthRad  = (REFF0 + REFF1 * cos(b2) + REFF2 * cos(b4) + REFF3 * cos(b6))
                   * EARTHRADIUS;

    satHeight    = rSSP- effEarthRad;

    satCrashFlag = (satHeight < 0.0) ? TRUE : FALSE;

    return;
}
 
/******************************************************************************/
/*                                                                            */
/* getStateVector: calculates satellite's state vector, i.e. position and     */
/*                 velocity in the mean equatorial coordinate system (M50)    */
/*                                                                            */
/*                 This function also provides essential calculations for the */
/*                 orbit number and squint angle calculations and therefore   */
/*                 needs to be called even when the state vector has been     */
/*                 calculated with the SGP4 model.                            */
/*                                                                            */
/******************************************************************************/

void getStateVector(timeArg)

double timeArg;
 
{
    double pVec[3], qVec[3];
    double xOrb, yOrb, zOrb, vxOrb, vyOrb, vzOrb;
    double xObs, yObs, zObs, xObsOrb, yObsOrb, zObsOrb;
    double tmp, vectLen, cosArgPerigee, sinArgPerigee, cosRaan, sinRaan;
    double phi, cosPhi, sinPhi;
    int    i;

    if (propModelType == TLEMEAN)
    {
        getTrueAnomaly(timeArg);
        curArgPerigee = epochArgPerigee + (timeArg - epochDay) * perigeePrec;
        curRaan       = epochRaan + (timeArg - epochDay) * raanPrec;
    }

    else
    {
        trueAnomaly   = trueAnomalyX;
        curArgPerigee = curArgPerigeeX;
        curRaan       = curRaanX;
    }

    cosTrueAnomaly = cos(trueAnomaly);
    sinTrueAnomaly = sin(trueAnomaly);

    cosArgPerigee  = cos(curArgPerigee);
    sinArgPerigee  = sin(curArgPerigee);

    cosRaan        = cos(curRaan);
    sinRaan        = sin(curRaan);

    satRadius      = semiMajorAxis * (1.0 - SQR(eccentricity))
                     / (1.0 + eccentricity * cosTrueAnomaly);

    if (propModelType == TLEMEAN || attitudeFlag)
    {
        xOrb    =  satRadius * cosTrueAnomaly;   /* position in orbital plane */
        yOrb    =  satRadius * sinTrueAnomaly;
        zOrb    =  ZERO;
    }

    /* calculate state vector if not yet calculated from SGP4/SDP4 model */

    if (propModelType == TLEMEAN)
    {
        tmp     =  sqrt(GMSGP / (semiMajorAxis * (1.0 - SQR(eccentricity))));

        vxOrb   = -tmp * sinTrueAnomaly;         /* velocity in orbital plane */
        vyOrb   =  tmp * (cosTrueAnomaly + eccentricity);
        vzOrb   =  ZERO;

        pVec[0] =  cosArgPerigee*cosRaan - sinArgPerigee*sinRaan*cosInclination;
        pVec[1] =  cosArgPerigee*sinRaan + sinArgPerigee*cosRaan*cosInclination;
        pVec[2] =  sinArgPerigee*sinInclination;

        qVec[0] = -sinArgPerigee*cosRaan - cosArgPerigee*sinRaan*cosInclination;
        qVec[1] = -sinArgPerigee*sinRaan + cosArgPerigee*cosRaan*cosInclination;
        qVec[2] =  cosArgPerigee*sinInclination;

        for (i = 0; i <= 2; i++)
        {
            satPosGl[i] =  xOrb * pVec[i] +  yOrb * qVec[i];
            satVelGl[i] = vxOrb * pVec[i] + vyOrb * qVec[i];
        }
    }

    /* calculate squint angle if satellite's attitude is specified */

    if (attitudeFlag)
    {
        phi     = reduce(lasTime - curRaan - HALFPI,ZERO,TWOPI);
        cosPhi  = cos(phi);
        sinPhi  = sin(phi);

        xObs    = siteCtyVecGl[0];
        yObs    = siteCtyVecGl[1];
        zObs    = siteCtyVecGl[2];

        xObsOrb = (-sinArgPerigee*cosInclination*cosPhi + 
	            cosArgPerigee*sinPhi) * xObs + 
	          ( sinArgPerigee*cosInclination*sinPhi + 
	            cosArgPerigee*cosPhi) * yObs -
	            sinArgPerigee*sinInclination * zObs;

        yObsOrb = (-cosArgPerigee*cosInclination*cosPhi - 
	            sinArgPerigee*sinPhi) * xObs + 
	          ( cosArgPerigee*cosInclination*sinPhi - 
	            sinArgPerigee*cosPhi) * yObs - 
	            cosArgPerigee*sinInclination * zObs;

        zObsOrb =  -sinInclination*cosPhi * xObs +
                    sinInclination*sinPhi * yObs + 
	            cosInclination * zObs;

        satSquintGl[0] = -xOrb - xObsOrb;
        satSquintGl[1] = -yOrb - yObsOrb;
        satSquintGl[2] = -zOrb - zObsOrb;
    }

    return;
}

/******************************************************************************/
/*                                                                            */
/* getSiteVector: computes the site postion and velocity in the mean          */
/*                equatorial coordinate system. The matrix 'siteRotMat' is    */
/*                used by getTopocentric to convert geocentric coordinates    */
/*                to topocentric (observer-centered) coordinates.             */
/*                                                                            */
/******************************************************************************/

void getSiteVector()

{
    double sb2, sb4, sb6, g1, g2, siteRA, cosSiteRA, sinSiteRA;
    double effSiteRad;

    sb2 = 2.0 * siteLat;
    sb4 = 4.0 * siteLat;
    sb6 = 6.0 * siteLat;

/*
    siteLat += F1*sin(sb2) + F2*sin(sb4) + F3*sin(sb6);
*/

    siteRA     = lasTime;
    cosSiteRA  = cos(siteRA);
    sinSiteRA  = sin(siteRA);

    effSiteRad = (REFF0 + REFF1 * cos(sb2) + REFF2 * cos(sb4) + 
                          REFF3 * cos(sb6)) * EARTHRADIUS + siteAlt;

    g1 = effSiteRad * cosSiteLat;
    g2 = effSiteRad * sinSiteLat;

    sitePosGl[0] =  g1 * cosSiteRA;
    sitePosGl[1] =  g1 * sinSiteRA;
    sitePosGl[2] =  g2;

    siteVelGl[0] = -SIDRATE * sitePosGl[1];
    siteVelGl[1] =  SIDRATE * sitePosGl[0];
    siteVelGl[2] =  ZERO;

    siteRotMatGl[0][0] =  sinSiteLat * cosSiteRA;
    siteRotMatGl[0][1] =  sinSiteLat * sinSiteRA;
    siteRotMatGl[0][2] = -cosSiteLat;
    siteRotMatGl[1][0] = -sinSiteRA;
    siteRotMatGl[1][1] =  cosSiteRA;
    siteRotMatGl[1][2] =  ZERO;
    siteRotMatGl[2][0] =  cosSiteRA * cosSiteLat;
    siteRotMatGl[2][1] =  sinSiteRA * cosSiteLat;
    siteRotMatGl[2][2] =  sinSiteLat;

    return;
}

/******************************************************************************/
/*                                                                            */
/* getRange: calculates satellite range                                       */
/*                                                                            */
/******************************************************************************/

void getRange()

{
    double satDist[3];
    int    i;

    for (i = 0; i <= 2; i++)
        satDist[i] = satPosGl[i] - sitePosGl[i];

    satRange    = absol(satDist);

    rangeRate   = ((satVelGl[0] - siteVelGl[0]) * satDist[0] + 
                   (satVelGl[1] - siteVelGl[1]) * satDist[1] +
                   (satVelGl[2] - siteVelGl[2]) * satDist[2]) / satRange;

    satVelocity = absol(satVelGl);

    return;
}

/******************************************************************************/
/*                                                                            */
/* getTopocentric: converts from geocentric mean equatorial coordinates to    */
/*                 topocentric coordinates                                    */
/*                                                                            */
/*                 topocentric coordinate system used: right-handed           */
/*                 +X = south,  +Y = east,  +Z = up                           */
/*                                                                            */
/******************************************************************************/

void getTopocentric(objGeo,pobjTopo)

double objGeo[3], (*pobjTopo)[3];

{
    double objDistT[3], objTopo[3];
    double objDistNorm;
    int    i;

    for (i = 0; i <= 2; i++)
        objDistT[i] = objGeo[i] - sitePosGl[i];

    objDistNorm = absol(objDistT);

    for (i = 0; i <= 2; i++)
        objDistT[i] /= objDistNorm;

    multMatVec(objDistT,objTopo,siteRotMatGl);

    for (i = 0; i <= 2; i++)
        (*pobjTopo)[i] = objTopo[i];

    return;
}

/******************************************************************************/
/*                                                                            */
/* getAziElev: calculates object's position in the local geodetic frame, and  */
/*             azimuth and elevation                                          */
/*                                                                            */
/******************************************************************************/

void getAziElev(switchFlag)

int switchFlag;

{
    double  localVec[3], aziVec[3], aziPerp[3], refrVec[3];
    double  vectLen, azimuth, elevation, arg;
    double  refIndex, refCorr, diffCorr, airMass;
    int     i;

    if (switchFlag == SAT)
    {
/*
        getPrecMatrix();
        multMatVec(satPosGl,satPosPrec,precMatrix);
        getTopocentric(satPosPrec,localVec);
*/

        getTopocentric(satPosGl,localVec);
    }

    if (switchFlag == SUN)
        getTopocentric(sunPosGl,localVec);

    if (switchFlag == MOON)
        getTopocentric(moonPosGl,localVec);

    aziVec[0]  = localVec[0];
    aziVec[1]  = localVec[1];
    aziVec[2]  = ZERO;

    vectLen    = absol(aziVec);

    aziVec[0] /= vectLen;
    aziVec[1] /= vectLen;

    if (aziVec[0] == ZERO)
        azimuth = (aziVec[1] > ZERO) ? HALFPI : THREEHALFPI;
    else
        azimuth = PI - atan2(aziVec[1],aziVec[0]);

    if (azimuth < ZERO)
        azimuth += PI;

    arg = localVec[2];

    if (arg >  1.0) arg =  1.0;      /* protect against out-of-range problems */
    if (arg < -1.0) arg = -1.0;

    elevation = asin(arg);

    /* apply correction for atmospheric refraction for visible tracking */

    getRefraction(atmPressure,ambTemperature,relHumidity,
                  WAVEL,REFWAVEL,EARTHRADIUS,SCALEHEIGHT,
                  siteAlt,elevation,&refIndex,&refCorr,&diffCorr,&airMass);

    aziPerp[0] =  aziVec[1];
    aziPerp[1] = -aziVec[0];
    aziPerp[2] = ZERO;

    cross(aziPerp,localVec,refrVec,1);

    vectLen = refCorr;

    for (i = 0; i <=2; i++)
        localVec[i] += vectLen * refrVec[i];

    vectLen = absol(localVec);

    for (i = 0; i <=2; i++)
        localVec[i] /= vectLen;

    arg = localVec[2];

    if (arg >  1.0) arg =  1.0;      /* protect against out-of-range problems */
    if (arg < -1.0) arg = -1.0;

    elevation = asin(arg);

    if (switchFlag == SAT)
    {
        for (i = 0; i <= 2; i++)
            localVecSatGl[i] = localVec[i];

        satAzimuth   = azimuth;
        satElevation = elevation;
    }

    if (switchFlag == SUN)
    {
        for (i = 0; i <= 2; i++)
            localVecSunGl[i] = localVec[i];

        sunAzimuth   = azimuth;
        sunElevation = elevation;
    }

    if (switchFlag == MOON)
    {
        for (i = 0; i <= 2; i++)
            localVecMoonGl[i] = localVec[i];

        moonAzimuth   = azimuth;
        moonElevation = elevation;
    }

    return;
}

/******************************************************************************/
/*                                                                            */
/* satEclipse: calculates the eclipses of the satellite                       */
/*                                                                            */
/* this function assumes that the Sun is point-like rather than extended      */
/* and that the Earth is perfectly spherical                                  */
/*                                                                            */
/******************************************************************************/

int satEclipse(elev)

double elev;

{
    double satPara, satPerp, satDist, alpha, beta;

    satDist = absol(satPosGl);

    satPara = scalar(satPosGl,sunPosGl) / sunDist;
    satPerp = sqrt(SQR(satDist) - SQR(satPara));

    /* case 1: satellite is on side of the Earth facing the Sun */

    if (satPara > ZERO)   /* satellite is always in daylight */
    {
        if (elev < ZERO)
            return(DAY);
        
        if (elev >= ZERO && sunElevation <= TWILIGHT)
            return(VISIBLE);
        else
            return(DAY);
    }

    alpha = asin(EARTHRADIUS / sunDist);
    beta  = atan(satPerp / (satPara + sunDist));

    /* case 2: satellite is on side of the Earth facing away from the Sun */

    if (beta >= alpha)    /* satellite can see the Sun */
    {
        if (elev >= ZERO && sunElevation <= TWILIGHT)
            return(VISIBLE);
        else
            return(DAY);
    }

    else                  /* satellite is in the Earth's shadow */
        return(NIGHT);
}

/******************************************************************************/
/*                                                                            */
/* getDoppler: calculates Doppler shift [Hz]                                  */
/*                                                                            */
/******************************************************************************/

void getDoppler()

{
    dopplerShift = -beaconFreq * rangeRate / CVAC;
    return;
}

/******************************************************************************/
/*                                                                            */
/* getPathLoss: calculates path loss for given frequency [Hz], range [km],    */
/*              and velocity of light [km/s]                                  */
/*                                                                            */
/******************************************************************************/

void getPathLoss()

{
    pathLoss = 20.0 * log10(FOURPI * satRange / CVAC * beaconFreq);
    return;
}

/******************************************************************************/
/*                                                                            */
/* getSquintAngle: calculates squint angle                                    */
/*                                                                            */
/******************************************************************************/

void getSquintAngle()

{
    double vectLen;

    if (!attitudeFlag)
    {
        squintAngle = ZERO;
        return;
    }

    vectLen     = absol(satSquintGl);
    squintAngle = acos(-(scalar(satAttGl,satSquintGl))/vectLen);

    return;
}

/******************************************************************************/
/*                                                                            */
/* getPhase: gets orbital phase of the satellite, i.e. the mean anomaly in    */
/*           units of usually 0 to 256                                        */
/*                                                                            */
/******************************************************************************/

void getPhase()

{
    satPhase = reduce(meanAnomaly/TWOPI*maxPhase + perigeePhase,ZERO,maxPhase);
    return;
}
 
/******************************************************************************/
/*                                                                            */
/* getMode: gets satellite mode                                               */
/*                                                                            */
/******************************************************************************/

void getMode()

{
    int curMode;

    sprintf(modeString,"  ");

    if (numModes > 0)
    {
        for (curMode = 0; curMode < numModes; curMode++)
        {
            if (satPhase >= modes[curMode].minPhase && 
                satPhase <  modes[curMode].maxPhase)
            {
                sprintf(modeString,"%s",modes[curMode].modeStr);
            }
        }
    }

    return;
}

/******************************************************************************/
/*                                                                            */
/* getShuttleOrbit: gets orbit number for the space shuttle                   */
/*                                                                            */
/* Convention for element sets from NASA  : add -1                            */
/*                             from NORAD : add +1                            */
/*                                                                            */
/******************************************************************************/

void getShuttleOrbit()

{
    if (!stsFlag)
        return;

    stsOrbit     = curOrbit + curArgPerigee * CRREV;
    stsOrbit    += (!strcmp(elementType,"NASA")) ? -1.0 : 1.0;

    stsOrbitNum  = (long) stsOrbit;

    if (propModelType == TLEMEAN)
        stsOrbitFract = modf(stsOrbit,&dummyd) * 100.0;
    else
        stsOrbitFract = curArgNodeX * CRREV * 100.0;

    if (preOrbitFlag)
        tleInvalidFlag = FALSE;
    else
        tleInvalidFlag = (fabs(stsOrbit - curOrbit) > TLELIMIT) ? TRUE : FALSE;

    return;
}

/******************************************************************************/
/*                                                                            */
/* End of function block satcalc.c                                            */
/*                                                                            */
/******************************************************************************/
