/******************************************************************************/
/*                                                                            */
/*  Title       : satcity.c                                                   */
/*  Author      : Manfred Bester, DL5KR                                       */
/*  Date        : 17Sep92                                                     */
/*  Last change : 11May94                                                     */
/*                                                                            */
/*  Synopsis    : Calculates distance and direction to nearest city.          */
/*                                                                            */
/*                                                                            */
/*  Copyright (c) 1992, 1993, 1994 Manfred Bester. All Rights Reserved.       */
/*                                                                            */
/*  Permission to use, copy, modify, and distribute this software and its     */
/*  documentation for educational, research and non-profit purposes, without  */
/*  fee, and without a written agreement is hereby granted, provided that the */
/*  above copyright notice and the following three paragraphs appear in all   */
/*  copies.                                                                   */
/*                                                                            */
/*  Permission to incorporate this software into commercial products may be   */
/*  obtained from the author, Dr. Manfred Bester, 1636 M. L. King Jr. Way,    */
/*  Berkeley, CA 94709, USA.                                                  */
/*                                                                            */
/*  IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, */
/*  SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF   */
/*  THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE AUTHOR HAS BEEN ADVISED  */
/*  OF THE POSSIBILITY OF SUCH DAMAGE.                                        */
/*                                                                            */
/*  THE AUTHOR SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT      */
/*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A   */
/*  PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS"      */
/*  BASIS, AND THE AUTHOR HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT, */
/*  UPDATES, ENHANCEMENTS, OR MODIFICATIONS.                                  */
/*                                                                            */
/******************************************************************************/

#include <stdio.h>
#include <math.h>
#include <malloc.h>

#include "satglobalsx.h"
#include "sattrack.h"

#ifndef SUNOS4
#include <stdlib.h>
#endif

/******************************************************************************/
/*                                                                            */
/* ground track vector segments                                               */
/*                                                                            */
/******************************************************************************/

char *gndTrkVector[] = { "N", "NNE", "NE", "ENE", "E", "ESE", "SE", "SSE",
                         "S", "SSW", "SW", "WSW", "W", "WNW", "NW", "NNW" };

/******************************************************************************/
/*                                                                            */
/* getGroundTrack: calculates the distance and direction of the nearest city  */
/*                                                                            */
/******************************************************************************/

void getGroundTrack(latSat,lngSat,groundTrkDist,groundTrkDir,groundTrkCity)

double latSat, lngSat, *groundTrkDist;
char   *groundTrkDir, *groundTrkCity;

{
    double cityVec[3], sspVec[3], distVec[3], northVec[3];
    double eastVec[3];
    double aziX, aziY, azi, distR, distRad, latLimit, lngLimit;
    double latSite, lngSite, cosLatSite, sinLatSite, cosLngSite, sinLngSite;
    double cosLatSat, sinLatSat, cosLngSat, sinLngSat;
    int    i, j, dir, cityPointer;

    cityPointer = 0;                       /* default setting: the North Pole */
                                           /* (or first entry in cities.dat)  */

    cosLatSat = cos(latSat);
    sinLatSat = sin(latSat);
    cosLngSat = cos(lngSat);
    sinLngSat = sin(lngSat);

    sspVec[0] = cosLngSat*cosLatSat;
    sspVec[1] = sinLngSat*cosLatSat;
    sspVec[2] = sinLatSat;

    distRad   = ONEMEG;

    if (fabs(latSat) > 70*CDR)         /* search limits for polar regions     */
    {
        latLimit =  30.1*CDR;
        lngLimit = 360.1*CDR;
    }

    if (fabs(latSat) <= 70*CDR)        /* search limits for rest of the world */
    {
        latLimit =  45.0*CDR;
        lngLimit =  45.0*CDR;
    }

    if (latSat > 30*CDR && latSat <  50*CDR &&
        lngSat > 70*CDR && lngSat < 130*CDR)
    {
        latLimit =  20.0*CDR;          /* search limits for the United States */
        lngLimit =  20.0*CDR;
    }

    for (i = 0; i < numCities; i++)
    {
        if (fabs(latSat - city[i].lat) < latLimit &&
            fabs(lngSat - city[i].lng) < lngLimit)
        {
            distVec[0] = city[i].X - sspVec[0];
            distVec[1] = city[i].Y - sspVec[1];
            distVec[2] = city[i].Z - sspVec[2];

            distR = absol(distVec);

            if (distR < distRad - ONEPPM)
            {
                distRad     = distR;
                cityPointer = i;
            }
        }

        newCityNum = cityPointer;
    }

    sprintf(groundTrkCity,"%s",city[cityPointer].cty);

    cityVec[0] = city[cityPointer].X;
    cityVec[1] = city[cityPointer].Y;
    cityVec[2] = city[cityPointer].Z;

    for (j = 0; j <= 2; j++)
        distVec[j] = cityVec[j] - sspVec[j];

    distR = absol(distVec);

    *groundTrkDist = EARTHRADIUS * distR;

    cosLatSite  = cos(city[cityPointer].lat);
    sinLatSite  = sin(city[cityPointer].lat);
    cosLngSite  = cos(city[cityPointer].lng);
    sinLngSite  = sin(city[cityPointer].lng);

    northVec[0] = -cosLngSite*sinLatSite;
    northVec[1] = -sinLngSite*sinLatSite;
    northVec[2] =  cosLatSite;

    cross(cityVec,northVec,eastVec,-1);

    aziX = scalar(northVec,distVec);
    aziY = scalar(eastVec,distVec);

    if (aziX == 0.0)
        azi = (aziY > 0.0) ? HALFPI : THREEHALFPI;
    else
        azi = PI - atan2(aziY,aziX);

    if (azi < 0.0)
        azi += PI;

    azi *= CRD;

    dir  = (int) ((azi + 11.25) / 22.5) % 16;

    if (cityPointer == 0) dir = 8;                          /* the North Pole */
    if (cityPointer == 1) dir = 0;                          /* the South Pole */

    sprintf(groundTrkDir,"%s",gndTrkVector[dir]);
    return;
}

/******************************************************************************/
/*                                                                            */
/* readCities: reads the city data base into a structure                      */
/*                                                                            */
/******************************************************************************/

void readCities()

{
    double lat, lng, alt;
    int    i;
    char   cityDat[80], cityName[50], str[100];
    FILE   *cityFile;

    if (verboseFlag)
        printf("\nreading city data base ...\n");

    sprintf(cityDat,"%s/%s/%s",strpHome,DATA,CITIES);

    if ((cityFile = fopen(cityDat,"r")) == 0)
    {
        printf("cannot open input file '%s'\n",cityDat);
        exit(-1);
    }

    numCities = 0;

    while (fgets(str,80,cityFile) && numCities < MAXCITIES)
    {
        sscanf(str,"`%[^`]`,%lf,%lf,%lf",cityName,&lat,&lng,&alt);

        city[numCities].lat = lat;
        city[numCities].lng = lng;
        city[numCities].alt = alt;

        if (strlen(cityName) > MAXCITYLEN)
        {
            i = MAXCITYLEN;

            do
            {
                i--;
            } while (cityName[i] != ' ');

            if (cityName[i-1] == ',')
                i--;

            cityName[i] = '\0';
        }

        strcpy(city[numCities].cty,cityName);

        city[numCities].lat *= CDR;
        city[numCities].lng *= CDR;

        numCities++;
    }

    fclose(cityFile);

    if (verboseFlag)
        printf("data base contains %d locations\n",numCities);

    return;
}

/******************************************************************************/
/*                                                                            */
/* calcCityVectors: calculates city vectors for the nearest city display      */
/*                                                                            */
/******************************************************************************/

void calcCityVectors()

{
    double citycoslat, citysinlat, citycoslng, citysinlng;
    int    i;

    for (i = 0; i < numCities; i++)
    {
        citycoslat = cos(city[i].lat);
        citysinlat = sin(city[i].lat);
        citycoslng = cos(city[i].lng);
        citysinlng = sin(city[i].lng);

        city[i].X  = citycoslng * citycoslat;
        city[i].Y  = citysinlng * citycoslat;
        city[i].Z  = citysinlat;
    }

    return;
}

/******************************************************************************/
/*                                                                            */
/* End of function block satcity.c                                            */
/*                                                                            */
/******************************************************************************/
