/******************************************************************************/
/*                                                                            */
/*  Title       : satctrl.c                                                   */
/*  Author      : Manfred Bester, DL5KR                                       */
/*  Date        : 23Jan94                                                     */
/*  Last change : 01Aug94                                                     */
/*                                                                            */
/*  Synopsis    : Routines for controlling satellite communications hardware  */
/*                (antenna interface, radio).                                 */
/*                                                                            */
/*                                                                            */
/*  Copyright (c) 1992, 1993, 1994 Manfred Bester. All Rights Reserved.       */
/*                                                                            */
/*  Permission to use, copy, modify, and distribute this software and its     */
/*  documentation for educational, research and non-profit purposes, without  */
/*  fee, and without a written agreement is hereby granted, provided that the */
/*  above copyright notice and the following three paragraphs appear in all   */
/*  copies.                                                                   */
/*                                                                            */
/*  Permission to incorporate this software into commercial products may be   */
/*  obtained from the author, Dr. Manfred Bester, 1636 M. L. King Jr. Way,    */
/*  Berkeley, CA 94709, USA.                                                  */
/*                                                                            */
/*  IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, */
/*  SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF   */
/*  THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE AUTHOR HAS BEEN ADVISED  */
/*  OF THE POSSIBILITY OF SUCH DAMAGE.                                        */
/*                                                                            */
/*  THE AUTHOR SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT      */
/*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A   */
/*  PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS"      */
/*  BASIS, AND THE AUTHOR HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT, */
/*  UPDATES, ENHANCEMENTS, OR MODIFICATIONS.                                  */
/*                                                                            */
/******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "satglobalsx.h"
#include "sattrack.h"

#ifndef SUNOS4
#include <stdlib.h>
#endif

/******************************************************************************/
/*                                                                            */
/* trackAntenna: controls movement of the satellite antenna                   */
/*                                                                            */
/* THIS FUNCTION NEEDS TO BE ADAPTED TO THE PARTICULAR ANTENNA CONTROLLER     */
/* USED. SPECIFY THE ANTENNA CONTROLLER TYPE IN 'sattrack.h'.                 */
/*                                                                            */
/******************************************************************************/

void trackAntenna()

{
    double sendEle, sendEleRate;
    int    error;
    char   syscomm[80];

    trackAziRate = trackAzimuth   - lastTrackAzi;
    trackEleRate = trackElevation - lastTrackEle;

    if (fabs(trackAzimuth - lastTrackAzi) > PI)          /* pass across north */
        trackAziRate += (trackAzimuth < lastTrackAzi) ? TWOPI : -TWOPI;

    if (fabs(trackAziRate) > MAXAZIRATE)
        trackAziRate = (trackAziRate > 0.0) ? MAXAZIRATE : -MAXAZIRATE;

    if (fabs(trackEleRate) > MAXELERATE)
        trackEleRate = (trackEleRate > 0.0) ? MAXELERATE : -MAXELERATE;

    lastTrackAzi = trackAzimuth;
    lastTrackEle = trackElevation;
    
    sendEle      = (trackElevation > HALFDEG) ? trackElevation : ZERO;
    sendEleRate  = (trackElevation > HALFDEG) ? trackEleRate   : ZERO;

    if (!trackCtrlFlag)
        return;

    switch(ANTENNATYPE) 
    {
        case NONE:
            break;

        case GENERIC:
            sprintf(syscomm,
                    "echo 'satAntenna: %8ld %8.3f %7.3f %7.3f %7.3f' > %s\n",
                    curYearSec,trackAzimuth*CRD,sendEle*CRD,
                    trackAziRate*CRD,sendEleRate*CRD,antennaIO);
            system(syscomm);
            break;

        case ISITELESCOPE:
            error = controlISI();
            break;

        case KANSASCITY:
            break;

        default:
            break;
    }

    return;
}

/******************************************************************************/
/*                                                                            */
/* trackRadio: control satellite radio (Doppler shift, etc.)                  */
/*                                                                            */
/* THIS FUNCTION NEEDS TO BE ADAPTED TO THE PARTICULAR RADIO USED.            */
/* SPECIFY THE RADIO TYPE IN 'sattrack.h'.                                    */
/*                                                                            */
/******************************************************************************/

void trackRadio()

{
    double radioFreq;
    char   syscomm[80];

    radioFreq = (beaconFreq + dopplerShift) * CHZMHZ;

    if (!trackCtrlFlag)
        return;

    switch(RADIOTYPE)
    {
        case NONE:
            break;

        case GENERIC:
            sprintf(syscomm,"echo 'satRadio: %8ld %13.6f' > %s\n",
                    curYearSec,radioFreq,radioIO);
            system(syscomm);
            break;

        case FT736:
            break;

        case IC970:
            break;

        default:
            break;
    }

    return;
}

/******************************************************************************/
/*                                                                            */
/* trackingControl: controls communications hardware                          */
/*                                                                            */
/*               'lastTrackYearSec' is used to avoid duplicate messages when  */
/*               switching between SINGLE and MULTISAT live displays          */
/*                                                                            */
/*               'trackCtrlFlag' is used to inhibit the first message after   */
/*               changing the object to avoid big azimuth and elevation rates */
/*                                                                            */
/******************************************************************************/

void trackingControl()

{
    double trackLimit;

    switch(trackObject)
    {
        case SUN:
            trackAzimuth   = sunAzimuth;
            trackElevation = sunElevation;
            break;

        case MOON:
            trackAzimuth   = moonAzimuth;
            trackElevation = moonElevation;
            break;

        case SAT:
            trackAzimuth   = satAzimuth;
            trackElevation = satElevation;
            break;

        default:
            break;
    }

    trackLimit = (ANTENNATYPE == ISITELESCOPE) ? -PI : TRACKLIMIT;
    trackLimit = (trackingFlag == AUTOTRK) ? ZERO : trackLimit;

    if (!preOrbitFlag && trackingFlag && trackElevation > trackLimit)
    {
        if (curYearSec != lastTrackYearSec)
        {
            trackAntenna();
            trackRadio();

            lastTrackYearSec = curYearSec; 
            trackCtrlFlag    = TRUE;
        }
    }

    return;
}

/******************************************************************************/
/*                                                                            */
/* End of function block satctrl.c                                            */
/*                                                                            */
/******************************************************************************/
