/******************************************************************************/
/*                                                                            */
/*  Title       : satdisp.c                                                   */
/*  Author      : Manfred Bester, DL5KR                                       */
/*  Date        : 24Feb92                                                     */
/*  Last change : 29Sep94                                                     */
/*                                                                            */
/*  Synopsis    : This function block performs the live display part of the   */
/*                satellite tracking program 'sattrack'.                      */
/*                                                                            */
/*                                                                            */
/*  Copyright (c) 1992, 1993, 1994 Manfred Bester. All Rights Reserved.       */
/*                                                                            */
/*  Permission to use, copy, modify, and distribute this software and its     */
/*  documentation for educational, research and non-profit purposes, without  */
/*  fee, and without a written agreement is hereby granted, provided that the */
/*  above copyright notice and the following three paragraphs appear in all   */
/*  copies.                                                                   */
/*                                                                            */
/*  Permission to incorporate this software into commercial products may be   */
/*  obtained from the author, Dr. Manfred Bester, 1636 M. L. King Jr. Way,    */
/*  Berkeley, CA 94709, USA.                                                  */
/*                                                                            */
/*  IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, */
/*  SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF   */
/*  THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE AUTHOR HAS BEEN ADVISED  */
/*  OF THE POSSIBILITY OF SUCH DAMAGE.                                        */
/*                                                                            */
/*  THE AUTHOR SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT      */
/*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A   */
/*  PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS"      */
/*  BASIS, AND THE AUTHOR HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT, */
/*  UPDATES, ENHANCEMENTS, OR MODIFICATIONS.                                  */
/*                                                                            */
/******************************************************************************/

#include <stdio.h>                /* standard C I/O                           */
#include <math.h>                 /* library of mathematical functions        */

#include "satglobalsx.h"          /* external global variables                */
#include "sattrack.h"             /* definition of various constants          */
#include "vt100.h"                /* definitions of VT100 macros              */

#ifndef SUNOS4
#include <stdlib.h>
#endif

static int launchMsgA, launchMsgB, launchMsgC, launchMsgD;
static int lastYearDay, reverseFlag;
static int firstMultiSatDisp;
static int nx, ny;

/******************************************************************************/
/*                                                                            */
/*  updateSingleSatLiveDisp: updates the tracking live display                */
/*                                                                            */
/******************************************************************************/

void updateSingleSatLiveDisp(curTime,nextRiseTime,nextSetTime,
            riseAzimuth,maxAzimuth,setAzimuth,maxElevation,maxRange)

double curTime, nextRiseTime, nextSetTime;
double riseAzimuth, maxAzimuth, setAzimuth, maxElevation, maxRange;

{
    double MET, fabsMET, timeArg, dTime, cUnit, dispAzimuth, dispElevation;
    long   tDayNum;
    int    tDay, tMonth, tYear, tYearDay, tHour, tMin, tSec;
    int    METday, METhour, METmin, METsec;
    int    i, lastHour, lastMin, lastSec;

    if (preOrbitFlag)                       /* do not display negative orbits */
    {
        for (i = 0; i <= 2; i++)
        {
            satPosGl[i] = ZERO;
            satVelGl[i] = ZERO;
        }

        orbitNum      = FALSE;
        orbitFract    = ZERO;
        stsOrbitNum   = FALSE;
        stsOrbitFract = ZERO;
        dopplerShift  = ZERO;
        pathLoss      = ZERO;
        squintAngle   = ZERO;
        satPhase      = ZERO;
        satAzimuth    = ZERO;
        satElevation  = -ONEPPM / 2.0;
        satRange      = ZERO;
        rangeRate     = ZERO;
        satLat        = ZERO;
        satLong       = ZERO;
        satHeight     = ZERO;
        satVelocity   = ZERO;
    }

    if (trackObject)
    {
        satRange      = ZERO;
        rangeRate     = ZERO;
    }

    cUnit = (stsFlag) ? CKMNM : 1.0;

    if (satElevation > ZERO && !elevationFlag)
    {
        gotoXY(nx+headerCol,ny+1);
        alarm();
        reverse();
        printf("%s",header);
        normal();
        elevationFlag = TRUE;
    }

    if (satElevation <= ZERO && elevationFlag)
    {
        gotoXY(nx+headerCol,ny+1);
        alarm();
        underline();
        printf("%s",header);
        normal();
        elevationFlag = FALSE;
    }

    timeArg = curTime + timeZone;
    tDayNum = (long) timeArg;
    convertTime(timeArg,&dummyi,&tHour,&tMin,&tSec);
    getDate(tDayNum,&tYear,&tMonth,&tDay,&tYearDay);

    if (tYearDay != lastYearDay)
    {
        gotoXY(nx+38,ny+3);
        printf("%3s  %02d%3s%02d", 
            dayName(tDayNum%7),tDay,monthName(tMonth),tYear%100);
        gotoXY(nx+38,ny+4);
        printf("%-3d %02d:%02d:%02d",tYearDay,tHour,tMin,tSec);
        lastYearDay = tYearDay;
    }

    else
    {
        gotoXY(nx+42,ny+4);
        printf("%02d:%02d:%02d",tHour,tMin,tSec);
    }

    if (debugFlag)
    {
        gotoXY(nx+42,ny+5);
        printf("%8ld",curYearSec);

        gotoXY(nx+42,ny+6);
        convertTime(lasTime*CRREV,&dummyi,&lastHour,&lastMin,&lastSec);
        printf("%02d:%02d:%02d",lastHour,lastMin,lastSec);
    }

    if (trackDispFlag)
    {
        gotoXY(nx+46,ny+8);

        if (trackingFlag)
        {
            reverse();

            if (trackingFlag == ON)
                printf("  ON");

            if (trackingFlag == AUTOTRK)
                printf("AUTO");

            normal();
        }

        else
        {
            printf(" OFF");
        }

        trackDispFlag = FALSE;
    }

    if (stsFlag)
    {
        gotoXY(nx+14,ny+6);
        printf("%5ld",stsOrbitNum);
        gotoXY(nx+20,ny+6);
        printf("%5.1f",stsOrbitFract);
    }

    else
    {
        gotoXY(nx+14,ny+6);
        printf("%5ld",orbitNum);
        gotoXY(nx+20,ny+6);
        printf("%5.1f",orbitFract);
    }

    gotoXY(nx+28,ny+6);

    if (orbitNum > 0 || stsOrbitNum > 0)
    {
        if (eclipseCode == VISIBLE) reverse();
        printf("%s",visibCode[eclipseCode]);
        if (eclipseCode == VISIBLE) normal();
    }

    else
        printf("%s",visibCode[BLANK]);

    if (newTrackObjectFlag)
    {
        gotoXY(nx+1,ny+7);
        if (trackObject == SUN) reverse();
        printf("Sun ");
        if (trackObject == SUN) normal();
    }

    gotoXY(nx+14,ny+7);
    printf("%5.1f %5.1f",sunAzimuth*CRD,sunElevation*CRD);

    if (newTrackObjectFlag)
    {
        gotoXY(nx+1,ny+8);
        if (trackObject == MOON) reverse();
        printf("Moon");
        if (trackObject == MOON) normal();
    }

    gotoXY(nx+14,ny+8);
    printf("%5.1f %5.1f",moonAzimuth*CRD,moonElevation*CRD);

    /* this is useful for solar eclipses */
/*
    gotoXY(nx+14,ny+9);
    printf("%7.3f %7.3f",(sunAzimuth - moonAzimuth)*CRD,
                         (sunElevation - moonElevation)*CRD);
*/

    /* the next lines are needed for debugging puposes */

    if (debugFlag)
    {
        gotoXY(nx+0,ny+9);
        printf("dH:%9.3f  M0:%9.6f M:%9.6f dT:%6.3f  Rev:%9.3f %9.3f",
            satRadius-perigeeHeight-EARTHRADIUS,
            epochMeanMotion,curMotionX,deltaTimeX,curOrbit,stsOrbit);
    }

    if (satElevation > ZERO) reverse();
    gotoXY(nx+69,ny+4);
    printf("%+8.3f",dopplerShift*CHZKHZ);
    gotoXY(nx+69,ny+5);
    printf("%8.3f",(beaconFreq + dopplerShift)*CHZMHZ);
    gotoXY(nx+69,ny+6);
    printf("%7.2f0",pathLoss);
    gotoXY(nx+69,ny+7);
    printf("%8.3f",squintAngle*CRD);
    gotoXY(nx+69,ny+8);
    printf("%7.2f0",satPhase);

    if (!numModes && satElevation > ZERO) normal();
    gotoXY(nx+78,ny+8);
    printf("%-2s",modeString);
    if (numModes && satElevation > ZERO) normal();

    if (trackObject == SAT)
    {
        dispAzimuth   = satAzimuth;
        dispElevation = satElevation;
    }

    if (trackObject == SUN)
    {
        dispAzimuth   = sunAzimuth;
        dispElevation = sunElevation;
    }

    if (trackObject == MOON)
    {
        dispAzimuth   = moonAzimuth;
        dispElevation = moonElevation;
    }

    gotoXY(nx+20,ny+10);
    if (dispElevation > ZERO) reverse();
    printf("%10.3f",dispAzimuth*CRD);
    gotoXY(nx+20,ny+11);
    printf("%10.3f",dispElevation*CRD);
    gotoXY(nx+35,ny+11);

    if (newTrackObjectFlag)
    {
        printf("_");
    }

    else
    {
        if (fabs(dispElevation) < ONEPPM && fabs(dispAzimuth) < ONEPPM)
            printf("/");

        else
        {
            if (prevDispEle > -PI - ONEPPM)
            {
                (dispElevation > prevDispEle) ? printf("/") : printf("\\");
            }
        }
    }

    prevDispEle = dispElevation;

    if (dispElevation > ZERO && trackObject) normal();
    gotoXY(nx+20,ny+12);
    printf("%10.3f",satRange*cUnit);
    gotoXY(nx+20,ny+13);
    printf("%10.3f",rangeRate*cUnit);
    if (dispElevation > ZERO && !trackObject) normal();

    newTrackObjectFlag = FALSE;

    gotoXY(nx+65,ny+10);
    printf("%6.3f",fabs(satLat*CRD));
    gotoXY(nx+76,ny+10);

    (satLat > ZERO) ? printf("N") : printf("S");
    
    if (fabs(satLat) < ONEPPM && fabs(satLong) < ONEPPM)
        prevSatLat = -2.0*ONEPPM;

    if (prevSatLat > -PI - ONEPPM)
    {
        (satLat > prevSatLat) ? printf(" /") : printf(" \\");
    }

    prevSatLat = satLat;
    gotoXY(nx+64,ny+11);
    printf("%7.3f",fabs(satLong*CRD));
    gotoXY(nx+75,ny+11);

    (satLong > ZERO) ? printf(" W") : printf(" E");
    
    gotoXY(nx+61,ny+12);
    if (satHeight < ZERO) reverseBlink();
    printf("%10.3f",satHeight*cUnit);
    if (satHeight < ZERO) normal();
    gotoXY(nx+62,ny+13);
    printf("%9.3f",satVelocity*cUnit);

/*
    gotoXY(nx+18,ny+14);
    printf("%+10.3f",sitePosGl[0]*cUnit);
    gotoXY(nx+42,ny+14);
    printf("%+10.3f",sitePosGl[1]*cUnit);
    gotoXY(nx+66,ny+14);
    printf("%+10.3f",sitePosGl[2]*cUnit);
*/

    gotoXY(nx+18,ny+15);
    printf("%+10.3f",satPosGl[0]*cUnit);
    gotoXY(nx+42,ny+15);
    printf("%+10.3f",satPosGl[1]*cUnit);
    gotoXY(nx+66,ny+15);
    printf("%+10.3f",satPosGl[2]*cUnit);

/*
    gotoXY(nx+18,ny+16);
    printf("%+10.3f",satPosS[0]*cUnit);
    gotoXY(nx+42,ny+16);
    printf("%+10.3f",satPosS[1]*cUnit);
    gotoXY(nx+66,ny+16);
    printf("%+10.3f",satPosS[2]*cUnit);
*/

    gotoXY(nx+18,ny+16);
    printf("%+10.3f",satVelGl[0]*cUnit);
    gotoXY(nx+42,ny+16);
    printf("%+10.3f",satVelGl[1]*cUnit);
    gotoXY(nx+66,ny+16);
    printf("%+10.3f",satVelGl[2]*cUnit);

/*
    gotoXY(nx+18,ny+16);
    printf("%+10.3f",satVelS[0]*cUnit);
    gotoXY(nx+42,ny+16);
    printf("%+10.3f",satVelS[1]*cUnit);
    gotoXY(nx+66,ny+16);
    printf("%+10.3f",satVelS[2]*cUnit);
*/

/*
    gotoXY(nx+6,ny+27);
    printf("Satellite:");
    gotoXY(nx+18,ny+27);
    printf("%+12.9f",localVecSatGl[0]);
    gotoXY(nx+33,ny+27);
    printf("%+12.9f",localVecSatGl[1]);
    gotoXY(nx+48,ny+27);
    printf("%+12.9f",localVecSatGl[2]);

    gotoXY(nx+6,ny+28);
    printf("Sun:");
    gotoXY(nx+18,ny+28);
    printf("%+12.9f",localVecSunGl[0]);
    gotoXY(nx+33,ny+28);
    printf("%+12.9f",localVecSunGl[1]);
    gotoXY(nx+48,ny+28);
    printf("%+12.9f",localVecSunGl[2]);

    gotoXY(nx+6,ny+29);
    printf("Moon:");
    gotoXY(nx+18,ny+29);
    printf("%+12.9f",localVecMoonGl[0]);
    gotoXY(nx+33,ny+29);
    printf("%+12.9f",localVecMoonGl[1]);
    gotoXY(nx+48,ny+29);
    printf("%+12.9f",localVecMoonGl[2]);
*/

    if (geoSyncFlag != GEOSTAT)
    {
        if (newRiseFlag)
        {
            gotoXY(nx+6,ny+18);

            if (satElevation <= ZERO)
            {
                reverse();
                printf("Next AOS  :");
                normal();
            }

            else
                printf("Last AOS  :");

            timeArg = nextRiseTime + timeZone;
            tDayNum = (long) timeArg;
            convertTime(timeArg,&dummyi,&tHour,&tMin,&tSec);
            getDate(tDayNum,&dummyi,&dummyi,&dummyi,&tYearDay);
            gotoXY(nx+18,ny+18);
            printf("%3d/%02d:%02d:%02d",tYearDay,tHour,tMin,tSec);
        }

        convertTime(ZERO,&tYearDay,&tHour,&tMin,&tSec);

        if (satElevation <= ZERO && nextRiseTime - curTime >= ZERO)
            convertTime(nextRiseTime-curTime,&tYearDay,&tHour,&tMin,&tSec);

        if (satElevation > ZERO && nextSetTime - curTime >= ZERO)
            convertTime(nextSetTime-curTime,&tYearDay,&tHour,&tMin,&tSec);

        if (tYearDay == 0 && tHour == 0 && tMin < 1)
        {
            countdownFlag = TRUE;
            alarm();
        }

        else
            countdownFlag = FALSE;

        if (countdownFlag && checkCountdown)
        {
            gotoXY(nx+6,ny+21);
            reverseBlink();
            printf("Countdown :");
            normal();
            checkCountdown = FALSE;
        }

        if (!countdownFlag && !checkCountdown)
        {
            gotoXY(nx+6,ny+21);
            reverse();
            printf("Countdown :");
            normal();
            checkCountdown = TRUE;
        }

        gotoXY(nx+18,ny+21);
        if (countdownFlag || satElevation > ZERO) reverse();
        printf("%3d/%02d:%02d:%02d",tYearDay,tHour,tMin,tSec);
        if (countdownFlag || satElevation > ZERO) normal();

        if (satElevation > ZERO && passDispFlag)
        {
            gotoXY(nx+6,ny+18);
            printf("Last AOS  :");
            gotoXY(nx+6,ny+20);
            reverse();
            printf("Next LOS  :");
            normal();
            passDispFlag = FALSE;
        }

        if (newRiseFlag)
        {
            convertTime(nextSetTime-nextRiseTime,&tYearDay,&tHour,&tMin,&tSec);
            gotoXY(nx+18,ny+19);
            printf("%3d/%02d:%02d:%02d",tYearDay,tHour,tMin,tSec);

            if (satElevation <= ZERO)
            {
                gotoXY(nx+6,ny+20);
                printf("Next LOS  :");
            }

            timeArg = nextSetTime + timeZone;
            tDayNum = (long) timeArg;
            convertTime(timeArg,&dummyi,&tHour,&tMin,&tSec);
            getDate(tDayNum,&dummyi,&dummyi,&dummyi,&tYearDay);
            gotoXY(nx+18,ny+20);
            printf("%3d/%02d:%02d:%02d",tYearDay,tHour,tMin,tSec);

            gotoXY(nx+64,ny+18);
            printf("%5.1f00",riseAzimuth*CRD);
            gotoXY(nx+77,ny+18);
            printf("%s",visibCode[eclipseRise]);

            gotoXY(nx+64,ny+19);
            printf("%5.1f00",maxAzimuth*CRD);
            gotoXY(nx+77,ny+19);
            printf("%s",visibCode[eclipseMax]);

            gotoXY(nx+64,ny+20);
            printf("%5.1f00",setAzimuth*CRD);
            gotoXY(nx+77,ny+20);
            printf("%s",visibCode[eclipseSet]);

            gotoXY(nx+61,ny+21);
            if (maxElevation > MAXELEV) reverse();
            printf("%8.1f00",maxElevation*CRD);
            gotoXY(nx+61,ny+22);
            printf("%8.1f00",maxRange*cUnit);
            if (maxElevation > MAXELEV) normal();

            newRiseFlag  = FALSE;
            passDispFlag = TRUE;
        }
    }

    if (launchFlag)
    {
        MET = curTime - launchEpoch;

        if (preLaunchFlag)
            MET *= -1.0;

        convertTime(MET,&METday,&METhour,&METmin,&METsec);
        gotoXY(nx+17,ny+22);

        if (preLaunchFlag)
        {
            if (METday)
                printf("%4d/%02d:%02d:%02d",-METday,METhour,METmin,METsec);
            else
                printf("  -%d/%02d:%02d:%02d",-METday,METhour,METmin,METsec);
        }

        else
            printf("%4d/%02d:%02d:%02d",METday,METhour,METmin,METsec);

        fabsMET = fabs(curTime - launchEpoch);

        if (fabsMET <= NINEMIN && fabsMET > ONEMIN && preLaunchFlag && 
            !launchMsgA)
        {
            gotoXY(nx+6,ny+23);
            alarm();
            reverse();
            printf("Final Launch Countdown  ");
            normal();
            launchMsgA = TRUE;
        }

        if (fabsMET <= ONEMIN && preLaunchFlag)
        {
            alarm();

            if (!launchMsgB)
            {
                gotoXY(nx+6,ny+23);
                reverseBlink();
                printf("Final Launch Countdown  ");
                normal();
                launchMsgB = TRUE;
            }
        }

        if (!preLaunchFlag && preOrbitFlag && !launchMsgC)
        {
            gotoXY(nx+6,ny+23);
            alarm();
            reverse();
            printf("Ascending to Orbit      ");
            normal();
            launchMsgC = TRUE;
        }

        if (!preLaunchFlag && !preOrbitFlag && launchMsgC && !launchMsgD)
        {
            gotoXY(nx+1,ny+23);
            alarm();
            clearCurs();
            launchMsgD = TRUE;
        }
    }

    getGroundTrack(satLat,satLong,&gndTrkDist,gndTrkDir,gndTrkCity);

    if (stsFlag)
    {
        if (preLaunchFlag)
        {
            newCityNum = -2;
            gndTrkDist = 0.0;
            sprintf(gndTrkDir,"N");
            sprintf(gndTrkCity,"Launch Complex 39, KSC, FL, USA");
        }

        if (!preLaunchFlag && preOrbitFlag)
        {
            newCityNum = -3;
            gndTrkDist = 0.0;
            sprintf(gndTrkDir,"NE");
            sprintf(gndTrkCity,"the Launch Pad --- Ascending to Orbit");
        }
    }

    if (!stsFlag && preOrbitFlag)
    {
        newCityNum = -2;
        gndTrkDist = 0.0;
        sprintf(gndTrkDir,"N");
        sprintf(gndTrkCity,"the Launch Pad");
    }

    gotoXY(nx+24,ny+24);
    printf("%7.1f",gndTrkDist*cUnit);
    gotoXY(nx+35,ny+24);
    printf("%3s",gndTrkDir);

    if (oldCityNum != newCityNum)
    {
        gotoXY(nx+42,ny+24);

        if (newCityNum == 0 || newCityNum == 1)        /* North or South Pole */
            printf("the %-37s",gndTrkCity);
        else
            printf("%-37s",gndTrkCity);
    }

    gotoXY(nx+80,ny+24);
    oldCityNum = newCityNum;
    fflush(stdout);
    return;
}

/******************************************************************************/
/*                                                                            */
/*  updateMultiSatLiveDisp: updates the tracking live display for multiple    */
/*                          satellites                                        */
/*                                                                            */
/******************************************************************************/

void updateMultiSatLiveDisp(curTime)

double curTime;

{
    double timeArg, nxtRiseT, nxtSetT, elev;
    long   tDayNum;
    int    i, j, k, n, tYear, tMonth, tDay, tYearDay, tHour, tMin, tSec;
    char   str[80];

    n = (gndTrkFlag) ? XDISPCOLUMNS - DISPCOLUMNS : 0;

    if (trackDispFlag)
    {
        if (xTermFlag)
            gotoXY(69,3);
        else
            gotoXY(65,1);

        if (trackingFlag)
        {
            reverse();

            if (trackingFlag == ON)
                printf("  ON");

            if (trackingFlag == AUTOTRK)
                printf("AUTO");

            normal();
        }

        else
        {
            if (!xTermFlag)
            {
                underline();
                printf("    ");
                normal();
            }

            else
                printf(" OFF");
        }

        trackDispFlag = FALSE;
    }

    timeArg = curTime + timeZone;
    tDayNum = (long) timeArg;
    convertTime(timeArg,&dummyi,&tHour,&tMin,&tSec);
    getDate(tDayNum,&tYear,&tMonth,&tDay,&tYearDay);

    if (tYearDay != lastYearDay)
    {
        gotoXY(46+n,3);
        printf("%3s  %02d%3s%02d",
            dayName(tDayNum%7),tDay,monthName(tMonth),tYear%100);
        gotoXY(67+n,3);
        printf("%3d  %02d:%02d:%02d",tYearDay,tHour,tMin,tSec);
        lastYearDay = tYearDay;
    }

    else
    {
        gotoXY(72+n,3);
        printf("%02d:%02d:%02d",tHour,tMin,tSec);
    }

    if (numSats > numSatLines)
    {
        gotoXY(3,6);
        printf("%3d-%3d (%3d)",satLinePtr+1,satLinePtr+numSatLines,numSats);
    }

    sortSats();

    for (i = 0; i < numSatLines; i++)
    {
        j = satOrder[satLinePtr+i];

        elev     = sat[j].satEle;
        nxtRiseT = sat[j].nextRiseT;
        nxtSetT  = sat[j].nextSetT;

        gotoXY(1,i+7);

        if (sat[j].statusFl == SELECT)
        {
            reverse();
            printf(">");
            normal();
            reverseFlag = FALSE;
        }

        if (sat[j].statusFl == MARK)
            printf("-");

        if (sat[j].statusFl == FALSE)
            bl();

        if (sat[j].geosyncFl == GEOSTAT)
        {
            if (elev > ZERO)
            {
                reverse();
                reverseFlag = TRUE;
            }
        }

        else
        {
            if (!sat[j].preorbitFl && !sat[j].invalidFl)
            {
                if (elev <= ZERO && nxtRiseT-curTime < ONEMIN) 
                {
                    if (sat[j].statusFl == SELECT)
                        alarm();

                    reverseBlink();
                    reverseFlag  = TRUE;
                    justAcquFlag = TRUE;
                }

                if (elev > ZERO && fabs(nxtSetT-curTime) < ONEMIN) 
                {
                    if (sat[j].statusFl == SELECT)
                        alarm();

                    reverseBlink();
                    reverseFlag = TRUE;
                }

                if (elev > ZERO && fabs(nxtSetT-curTime) >= ONEMIN) 
                {
                    reverse();
                    reverseFlag = TRUE;
                }

                if (elev > TRACKLIMIT && nxtRiseT-curTime > TENMIN)
                {
                    reverse();
                    reverseFlag  = TRUE;
                    justLostFlag = TRUE;
                }
            }
        }

        if (objectNumFlag)
            sprintf(str,"%ld",sat[j].satIdNum);
        else
            sprintf(str,"%s",sat[j].satellite);

        printf("%-s",str);

        if (strlen(str) < MAXSATNAMELEN)
            bl();

        for (k = 0; k < MAXSATNAMELEN - strlen(str) - 1; k++)
            printf(".");

        if (debugFlag)
        {
            gotoXY(11,i+7);
            if (sat[j].stsFl)
                printf("%5.1f",sat[j].stsOrbFrac);
            else
                printf("%5.1f",sat[j].satOrbFrac);
        }

        if (justAcquFlag || justLostFlag)
        {
             normal();
             reverseFlag  = FALSE;
             justAcquFlag = FALSE;
             justLostFlag = FALSE;
        }

        if (sat[j].geosyncFl == GEOSTAT || sat[j].crashFl || sat[j].invalidFl 
                                                          || sat[j].nopassFl)
        {
            if (sat[j].geosyncFl == GEOSTAT)
                printf("  Geostationary satellite                     ");

            if (sat[j].crashFl)
            {
                printf("  Satellite has crashed already");
                clearCurs();
            }

            if (sat[j].invalidFl && !sat[j].crashFl)
            {
                printf("  TLE set is out of date                      ");
                clearCurs();
            }

            if (sat[j].geosyncFl != GEOSTAT && !sat[j].crashFl && 
                !sat[j].invalidFl && sat[j].nopassFl)
            {
                printf("  No pass for %6.3f days                     ",
                    nxtRiseT - curTime);
            }
        }

        else
        {
            convertTime(ZERO,&tYearDay,&tHour,&tMin,&tSec);

            if (elev <= ZERO && nxtRiseT-curTime >= ZERO)
                convertTime(nxtRiseT-curTime,&tYearDay,&tHour,&tMin,&tSec);

            if (elev > ZERO && nxtSetT-curTime >= ZERO)
                convertTime(nxtSetT-curTime,&tYearDay,&tHour,&tMin,&tSec);

            printf("%3d/%02d:%02d:%02d  ",tYearDay,tHour,tMin,tSec);

            if (elev <= ZERO)
            {
                tDayNum = (long) (nxtRiseT + timeZone);
                convertTime(nxtRiseT+timeZone,&dummyi,&tHour,&tMin,&tSec);
            }

            else
            {
                tDayNum = (long) (nxtSetT + timeZone);
                convertTime(nxtSetT+timeZone,&dummyi,&tHour,&tMin,&tSec);
            }

            getDate(tDayNum,&tYear,&tMonth,&tDay,&tYearDay);
            printf("%3d/%02d:%02d:%02d ",tYearDay,tHour,tMin,tSec);

            convertTime(nxtSetT-nxtRiseT,&tDay,&tHour,&tMin,&tSec);
            printf("%3d:%02d:%02d  ",tDay*24+tHour,tMin,tSec);

            printf("%2.0f",sat[j].maxEle*CRD);

            (sat[j].maxEle > MAXELEV) ? printf("* ") : printf("  ");

            printf("%s%s%s ",visibCode[sat[j].eclRise],
                             visibCode[sat[j].eclMax],
                             visibCode[sat[j].eclSet]);
        }

        if (sat[j].preorbitFl)
        {
            reverse();

            if (sat[j].prelaunchFl)
                printf("  not yet launched");
            else
                printf("ascending to orbit");

            normal();
            reverseFlag = FALSE;
            clearCurs();
        }

        if (!sat[j].preorbitFl && !sat[j].crashFl && !sat[j].invalidFl)
        {
            printf("%s  ",visibCode[sat[j].eclCode]);

            printf("%3.0f %3.0f ",sat[j].satAzi*CRD,sat[j].satEle*CRD);

            if (firstMultiSatDisp)
                printf("_");

            else
            {
                if (sat[j].satLastEle < sat[j].satEle)
                    printf("/");
                else
                    printf("\\");
            }

            printf("%6.0f",sat[j].satRng);
        }

        if (gndTrkFlag && !sat[j].crashFl && !sat[j].invalidFl)
        {
            if (!strncmp(sat[j].model,TLEMEANMODEL,4))
                printf(" M ");

            if (!strncmp(sat[j].model,LOWEARTHMODEL,4))
                printf(" G ");

            if (!strncmp(sat[j].model,DEEPSPACEMODEL,4))
                printf(" D ");
        }

        if (!sat[j].preorbitFl && !sat[j].crashFl && !sat[j].invalidFl && 
            gndTrkFlag)
        {
            printf("%4.0f km %3s %-37s",
                sat[j].ctyDist,sat[j].ctyDir,sat[j].ctyStr);
        }

        if (reverseFlag)
        {
            normal();
            reverseFlag = FALSE;
        }
    }

    gotoXY(DISPCOLUMNS + n, numSatLines + 6);
    firstMultiSatDisp = FALSE;
    fflush(stdout);
    return;
}

/******************************************************************************/
/*                                                                            */
/*  initSingleSatLiveDisp: initializes live display (text strings)            */
/*                                                                            */
/******************************************************************************/

void initSingleSatLiveDisp()

{
    char stsUnit[6], str[80];

    launchMsgA         = FALSE;
    launchMsgB         = FALSE;
    launchMsgC         = FALSE;
    launchMsgD         = FALSE;
    trackDispFlag      = TRUE;
    elevationFlag      = FALSE;
    newTrackObjectFlag = TRUE;
    lastYearDay        = -1;

    system("clear");
    dispTitle();

    gotoXY(nx+1,ny+3);
    printf("Ground Stn : %15s",siteName);
    advCurs(3);
    printf("Date: ___  _______");
    advCurs(3);
    printf("Radio Beacon:  %9.3f MHz",beaconFreq*CHZMHZ);

    gotoXY(nx+1,ny+4);

    if (!strlen(satAlias))
        printf("Satellite  : %15s",satName);

    else
    {
        if (stsFlag)
            printf("Vehicle    : %15s",satAlias);
        else
            printf("Satellite  : %15s",satAlias);
    }

    advCurs(3);
    printf("%-4s: ___ __:__:__",timeZoneStr);
    advCurs(3);
    printf("Doppler Shft:    -__.___ kHz");

    gotoXY(nx+1,ny+5);
    printf("Inclination:     %7.3f deg",inclination*CRD);

    if (debugFlag)
    {
        advCurs(3);
        printf("UTC :     ________");
        advCurs(3);
    }

    else
        advCurs(24);

    printf("Frequency   :   ____.___ MHz");

    gotoXY(nx+1,ny+6);
    printf("Orbit      : _____ ___._ ");
    putc('%',stdout);
    printf(" _");

    if (debugFlag)
    {
        advCurs(3);
        printf("LAST:     __:__:__");
        advCurs(3);
    }

    else
    {
        advCurs(3);
        printf("Object  :   %6ld",satNum);
        advCurs(3);
    }

    printf("Path Loss   :    ___.___ dB");

    gotoXY(nx+1,ny+7);
    printf("Sun   Az/El: ___._ -__._ deg");
    advCurs(3);
    printf("Model   :%9s",propModel);
    advCurs(3);
    printf("Squint Angle:    ___.___ deg");

    gotoXY(nx+1,ny+8);
    printf("Moon  Az/El: ___._ -__._ deg");
    advCurs(3);
    printf("Tracking:     ____");
    advCurs(3);
    printf("Phase (%4.0f):    ___.___ __",maxPhase);

    gotoXY(nx+6,ny+10);
    printf("Azimuth   :      ___.___ deg");
    advCurs(12);
    printf("Latitude     :     __.___ deg _ _");

    gotoXY(nx+6,ny+11);
    printf("Elevation :      -__.___ deg _");
    advCurs(10);
    printf("Longitude    :    ___.___ deg _");

    sprintf(stsUnit,(stsFlag) ? "nm" : "km");

    gotoXY(nx+6,ny+12);
    printf("Range     :   ______.___ %2s",stsUnit);
    advCurs(13);
    printf("Height       : ______.___ %2s",stsUnit);

    gotoXY(nx+6,ny+13);
    printf("Range Rate:    -____.___ %2s/s",stsUnit);
    advCurs(11);
    printf("Velocity     :  _____.___ %2s/s",stsUnit);

/*
    gotoXY(nx+1,ny+14);
    printf("Site  Vector  X: -_____.___ %2s",stsUnit);
    advCurs(8);
    printf("Y: -_____.___ %2s",stsUnit);
    advCurs(8);
    printf("Z: -_____.___ %2s",stsUnit);
*/

    gotoXY(nx+1,ny+15);
    printf("State Vector  X: -_____.___ %2s",stsUnit);
    advCurs(8);
    printf("Y: -_____.___ %2s",stsUnit);
    advCurs(8);
    printf("Z: -_____.___ %2s",stsUnit);

/*
    gotoXY(nx+1,ny+16);
    printf("Norad Model   X: -_____.___ %2s",stsUnit);
    advCurs(8);
    printf("Y: -_____.___ %2s",stsUnit);
    advCurs(8);
    printf("Z: -_____.___ %2s",stsUnit);
*/

    gotoXY(nx+1,ny+16);
    printf("             VX: -_____.___ %2s/s",stsUnit);
    advCurs(5);
    printf("VY: -_____.___ %2s/s",stsUnit);
    advCurs(5);
    printf("VZ: -_____.___ %2s/s",stsUnit);

    if (geoSyncFlag == GEOSTAT)
    {
        gotoXY(nx+12,ny+19);
        printf("Geostationary Satellite --- No Calculation of AOS and LOS");
    }

    else
    {
        gotoXY(nx+6,ny+18);
        printf("Next AOS  : ___/__:__:__ %-4s",timeZoneStr);
        advCurs(11);
        printf("AOS Azimuth  :    ___.___ deg  _");

        gotoXY(nx+6,ny+19);
        printf("Duration  : ___/__:__:__");
        advCurs(16);
        printf("MEL Azimuth  :    ___.___ deg  _");

        gotoXY(nx+6,ny+20);
        printf("Next LOS  : ___/__:__:__ %-4s",timeZoneStr);
        advCurs(11);
        printf("LOS Azimuth  :    ___.___ deg  _");

        gotoXY(nx+6,ny+21);
        reverse();
        printf("Countdown :");
        normal();
        printf(" ___/__:__:__");
        advCurs(16);
        printf("Max Elevation:     __.___ deg");
    }

    if (launchFlag)
    {
        gotoXY(nx+6,ny+22);
        printf("MET       : ___/__:__:__");
    }

    if (geoSyncFlag != GEOSTAT)
    {
        gotoXY(nx+46,ny+22);
        printf("MEL Range    : ______.___ %s",stsUnit);
    }

    gotoXY(nx+1,ny+24);
    sprintf(str,"_____________________________________ ");
    printf("Ground Track Location: _____._ %2s ___ of %s",stsUnit,str);
    fflush(stdout);
    return;
}

/******************************************************************************/
/*                                                                            */
/*  dispTitle: displays title for the single satellite live display           */
/*                                                                            */
/******************************************************************************/

void dispTitle()

{
    int  i;
    char satDispName[20], str[80], satStr[20];

    if (xTermFlag && didMultDispFlag)
    {
        nx = 26;
        ny = 2;
    }

    else
    {
        nx = 0;
        ny = 0;
    }

    sprintf(satStr,"%s %s",sattrName,sattrVersion);
    sprintf(satDispName,"%s",satName);
    upperCase(satDispName);
    sprintf(header," %s %s ",satDispName,dispHeader);
    headerCol = 41 - strlen(header) / 2;

    if (nx && ny)
    {
        gotoXY(nx,ny);
        underline();
        printf("%82c",' ');

        gotoXY(nx,ny+25);
        printf("%82c",' ');
        normal();

        sprintf(str,"%51s"," ");
        gotoXY(nx,ny+1);
    }

    else
    {
        sprintf(str,"%49s"," ");
        gotoXY(1,1);
    }

    underline();
    printf("%-14s %s %15s",satStr,str,callSign);

    gotoXY(nx+headerCol,ny+1);
    printf("%s",header);
    normal();

    return;
}

/******************************************************************************/
/*                                                                            */
/*  initMultiSatLiveDisp: initializes multiple satellite live display         */
/*                                                                            */
/******************************************************************************/

void initMultiSatLiveDisp()

{
    int  i, n;
    char str[150], satStr[150];

    lastYearDay       = TRUE;
    trackDispFlag     = TRUE;
    reverseFlag       = FALSE;
    justLostFlag      = FALSE;
    didMultDispFlag   = TRUE;
    firstMultiSatDisp = TRUE;
    quickStartFlag    = FALSE;

    numSatLines = (xTermFlag) ? numLinesX - 6 : numLines - 6;

    if (numSats < numSatLines)
        numSatLines = numSats;

    n = (gndTrkFlag) ? XDISPCOLUMNS - DISPCOLUMNS : 0;

    system("clear");
    gotoXY(1,1);
    underline();
    sprintf(satStr,"");
    sprintf(satStr,"%8s %-6s",sattrName,sattrVersion);

    for (i = 0; i < 50 + n; i++)
        strcat(satStr," ");

    sprintf(satStr,"%s%15s",satStr,callSign);
    printf("%s",satStr);

    sprintf(header,"GROUP %s MULTISAT %s",satGroup,dispHeader);
    headerCol = (DISPCOLUMNS + n + 1 - strlen(header)) / 2;
    gotoXY(headerCol,1);
    printf("%s",header);
    normal();

    gotoXY(2,3);

    if (gndTrkFlag)
        printf("Ground Station: %-s",fullSiteName);
    else
        printf("Ground Station: %-s",siteName);

    if (xTermFlag)
    {
        gotoXY(59,3);
        printf("Tracking: ____");
    }

    gotoXY(40+n,3);
    printf("Date: ___  _______   %4s: ___  __:__:__",timeZoneStr);

    gotoXY(2,5);
    underline();
    printf("   Satellite  ");
    normal();
    bl();
    underline();
    printf(" Countdown ");
    normal();
    bl(); bl();
    underline();
    printf("Next AOS/LOS");
    normal();
    bl(); bl();
    underline();
    printf("Duration");
    normal();
    bl();
    underline();
    printf("MEL");
    normal();
    bl(); bl();
    underline();
    printf("Visib");
    normal();
    bl(); bl();
    underline();
    printf("Azi");
    normal();
    bl();
    underline();
    printf("Ele");
    normal();
    bl(); bl();
    underline();
    printf(" Range");
    normal();

    if (gndTrkFlag)
    {
        bl(); 
        underline();
        printf("M");
        normal();
        bl();
        underline();
        printf("              Ground Track Location              ");
        normal();
    }

    if (numSats > numSatLines)
    {
        gotoXY(3,6);
        printf("%3d-%3d (%3d)",satLinePtr+1,satLinePtr+numSatLines,numSats);
    }

    for (i = 0; i < numSatLines; i++)
    {
        gotoXY(2,i+7);
        printf("%-12s",sat[i].satellite);
    }

    gotoXY(DISPCOLUMNS + n, numSatLines + 6);
    fflush(stdout);
    return;
}

/******************************************************************************/
/*                                                                            */
/*  sortSats: orders satellites that are below the horizon by rise time and   */
/*            those above the horizon by reverse set time                     */
/*                                                                            */
/******************************************************************************/

void sortSats()

{
    int i, j, tmpIndex;

    for (i = 0; i < numSats; i++)
        satOrder[i] = i;

    for (i = 0; i < numSats; i++)
    {
        for (j = 0; j < numSats - 1; j++)
        {
            if (sat[satOrder[j]].nextRiseT > sat[satOrder[j+1]].nextRiseT)
            {
                tmpIndex      = satOrder[j];
                satOrder[j]   = satOrder[j+1];
                satOrder[j+1] = tmpIndex;
            } 
        }
    }

    for (i = 0; i < numSats; i++)
    {
        for (j = 0; j < numSats - 1; j++)
        {
            if (sat[satOrder[j]].satEle   > ZERO && 
                sat[satOrder[j+1]].satEle > ZERO && 
                sat[satOrder[j]].nextSetT < sat[satOrder[j+1]].nextSetT)
            {
                tmpIndex      = satOrder[j];
                satOrder[j]   = satOrder[j+1];
                satOrder[j+1] = tmpIndex;
            }
        }
    }

    return;
}

/******************************************************************************/
/*                                                                            */
/*  end of function satdisp.c                                                 */
/*                                                                            */
/******************************************************************************/
