/******************************************************************************/
/*                                                                            */
/*  Title       : satprop.c                                                   */
/*  Author      : Manfred Bester, DL5KR                                       */
/*  Date        : 20Nov93                                                     */
/*  Last change : 05Oct94                                                     */
/*                                                                            */
/*  Synopsis    : SGP4/SDP4 propagation routines (translated from FORTRAN)    */
/*                for reference see:                                          */
/*                "Spacetrack Report No. 3, Models for Propagation of         */
/*                NORAD Element Sets", F. R. Hoots and R. L. Roehrich,        */
/*                December 1980                                               */
/*                                                                            */
/*  REMARK: The SDP4 model implementation has not been completed in this      */
/*          release of SatTrack (V2.0). The full implementation of the SDP4   */
/*          model for deep space objects is expected to be part of a future   */
/*          release. For now, a simplified model is used for deep space       */
/*          objects (the so-called 'TLE Mean model').                         */
/*                                                                            */
/*                                                                            */
/*  Copyright (c) 1992, 1993, 1994 Manfred Bester. All Rights Reserved.       */
/*                                                                            */
/*  Permission to use, copy, modify, and distribute this software and its     */
/*  documentation for educational, research and non-profit purposes, without  */
/*  fee, and without a written agreement is hereby granted, provided that the */
/*  above copyright notice and the following three paragraphs appear in all   */
/*  copies.                                                                   */
/*                                                                            */
/*  Permission to incorporate this software into commercial products may be   */
/*  obtained from the author, Dr. Manfred Bester, 1636 M. L. King Jr. Way,    */
/*  Berkeley, CA 94709, USA.                                                  */
/*                                                                            */
/*  IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, */
/*  SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF   */
/*  THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE AUTHOR HAS BEEN ADVISED  */
/*  OF THE POSSIBILITY OF SUCH DAMAGE.                                        */
/*                                                                            */
/*  THE AUTHOR SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT      */
/*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A   */
/*  PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS"      */
/*  BASIS, AND THE AUTHOR HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT, */
/*  UPDATES, ENHANCEMENTS, OR MODIFICATIONS.                                  */
/*                                                                            */
/******************************************************************************/

#include <stdio.h>
#include <math.h>

#include "satglobalsx.h"
#include "satglobalsp.h"
#include "sattrack.h"
#include "satprop.h"

#ifndef SUNOS4
#include <stdlib.h>
#endif

/******************************************************************************/
/*                                                                            */
/* checkPropModelType: checks type of propagation model used                  */
/*                                                                            */
/******************************************************************************/

void checkPropModelType()

{
    getSemiMajorAxis();
    checkPerigeeHeight();

    if (deepSpaceFlag)
    {
        sprintf(propModel,"%s",TLEMEANMODEL);                 /* for now only */
        propModelType = TLEMEAN;

/*
        sprintf(propModel,"%s",DEEPSPACEMODEL);
        propModelType = DEEPSPACE;
*/
    }

    else
    {
        sprintf(propModel,"%s",LOWEARTHMODEL);
        propModelType = LOWEARTH;
    }

    if (modelFlag)
    {
        sprintf(propModel,"%s",TLEMEANMODEL);
        propModelType = TLEMEAN;
    }

    return;
}

/******************************************************************************/
/*                                                                            */
/* changeNoradUnits: changes units of fundamental elements for use with NORAD */
/*                   model                                                    */
/*                                                                            */
/******************************************************************************/

void changeNoradUnits()

{
    epoch0     = epochDay;
    ecc0       = eccentricity;

    inc0       = inclination;                                /* [rad]         */
    cosInc     = cosInclination;
    sinInc     = sinInclination;

    meanMot0   = epochMeanMotion * TWOPI / MPD;              /* [rad/min]     */
    meanAn0    = epochMeanAnomaly;                           /* [rad]         */
    argPer0    = epochArgPerigee;                            /* [rad]         */
    raan0      = epochRaan;                                  /* [rad]         */

    decRate    = decayRate * TWOPI / MPD2;                   /* [rad/min^2]   */
    decRateDot = decayRateDot * TWOPI / MPD3;                /* [rad/min^3]   */

    bStar      = bStarCoeff / AE;

    return;
}

/******************************************************************************/
/*                                                                            */
/* getSemiMajorAxis: recover original mean motion (meanMotDeep) and semimajor */
/*                   axis (sma0Deep) from input elements                      */
/*                                                                            */
/******************************************************************************/

void getSemiMajorAxis()

{
    changeNoradUnits();

    cos2Inc       = SQR(cosInc);
    cos4Inc       = SQR(cos2Inc);

    x3thm1        = 3.0 * cos2Inc - 1.0;
    x7thm1        = 7.0 * cos2Inc - 1.0;
    x1m1th        = 1.0 - cos2Inc;
    x1m5th        = 1.0 - 5.0 * cos2Inc;

    eccSQ         = SQR(ecc0);
    eccCB         = ecc0 * eccSQ;

    beta02        = 1.0 - eccSQ;                          /* needed elsewhere */
    beta0         = sqrt(beta02);                         /* needed elsewhere */
    beta03        = beta0 * beta02;                       /* needed elsewhere */

    sma1          = pow((KE/meanMot0),TWOTHIRDS);

    delta1        = 1.5 * CK2 * x3thm1 / (SQR(sma1) * beta03);
    delta12       = SQR(delta1);
    delta13       = delta1 * delta12;

    sma0          = sma1 * (1.0 - ONETHIRD * delta1 - delta12 
                                - 134.0/81.0 * delta13);

    delta0        = 1.5 * CK2 * x3thm1 / (SQR(sma0) * beta03);

    meanMotDeep   = meanMot0 / (1.0 + delta0);

    sma0Deep      = sma0 / (1.0 - delta0);
    semiMajorAxis = sma0Deep * EARTHRADIUS;

    deepSpaceFlag = (TWOPI / meanMotDeep >= 225.0) ? TRUE : FALSE;

    return;
}

/******************************************************************************/
/*                                                                            */
/* getTimeArgs: gets time arguments for the SGP4/SDP4 model calculations      */
/*                                                                            */
/******************************************************************************/

void getTimeArgs(tmArg)

double tmArg;

{
    tFP = (tmArg - epoch0) * MPD;                                    /* [min] */
    tSQ = tFP * tFP;
    tCB = tSQ * tFP;
    tQD = tCB * tFP;
    tQN = tQD * tFP;

    return;
}

/******************************************************************************/
/*                                                                            */
/* checkPerigeeHeight: checks perigee height and make modifications           */
/*                     accordingly                                            */
/*                                                                            */
/******************************************************************************/

void checkPerigeeHeight()

{
    perigeeHeight = (sma0Deep * (1.0 - ecc0) - AE) * EARTHRADIUS;

    truncFlag     = (perigeeHeight < 220.0) ? TRUE : FALSE;
    satCrashFlag  = (perigeeHeight <   0.0) ? TRUE : FALSE;

    sStar = S;
    qmst4 = Q0MS0T4;

    if (perigeeHeight < 156.0)
    {
        sStar = perigeeHeight - S0;

        if (perigeeHeight <= 98.0)
            sStar = 20.0;
    
        qmst4  = pow(((Q0 - sStar) * AE / EARTHRADIUS),4.0);
        sStar /= EARTHRADIUS;
        sStar += AE;
    }

    return;
}

/******************************************************************************/
/*                                                                            */
/* calcCommonPerturb: calculates quantities that are common to the SGP4 and   */
/*                    SDP4 models                                             */
/*                                                                            */
/******************************************************************************/

void calcCommonPerturb()

{
    pInvSQ = 1.0 / (SQR(sma0Deep) * SQR(beta02));
    xi     = 1.0 / (sma0Deep - sStar);
    eta    = sma0Deep * ecc0 * xi;
    etaSQ  = SQR(eta);
    eccEta = ecc0 * eta;
    psiSQ  = fabs(1.0 - etaSQ);
    coef0  = qmst4 * pow(xi,4.0);
    coef1  = coef0 / pow(psiSQ,3.5);

    c2     = coef1 * meanMotDeep * 
             (sma0Deep * (1.0 + 1.5 * etaSQ + eccEta * (4.0 + etaSQ)) + 
             0.75 * CK2 * xi / psiSQ * x3thm1 * 
             (8.0 + 3.0 * etaSQ * (8.0 + etaSQ)));

    c1     = bStar * c2;

    c3     = coef0 * xi * A30CK2 * meanMotDeep * AE * sinInc / ecc0;

    c4     = 2.0 * meanMotDeep * coef1 * sma0Deep * beta02 * 
             (eta * (2.0 + 0.5 * etaSQ) + ecc0 * (0.5 + 2.0 * etaSQ) - 
             2.0 * CK2 * xi / (sma0Deep * psiSQ) * 
             (-3.0 * x3thm1 * (1.0 - 2.0 * eccEta + etaSQ * 
             (1.5 - 0.5 * eccEta)) + 0.75 * x1m1th * 
             (2.0 * etaSQ - eccEta * (1.0 + etaSQ)) * cos(2.0 * argPer0)));

    c5     = 2.0 * coef1 * sma0Deep * beta02 * 
             (1.0 + 2.75 * (etaSQ + eccEta) + eccEta * etaSQ);

    clc1   = 3.0  * CK2 * pInvSQ * meanMotDeep;
    clc2   = clc1 * CK2 * pInvSQ;
    clc3   = 1.25 * CK4 * SQR(pInvSQ) * meanMotDeep;

    return;
}

/******************************************************************************/
/*                                                                            */
/* updateGravityAndDrag: updates quantities for secular gravity and           */
/*                       atmospheric drag                                     */
/*                                                                            */
/******************************************************************************/

void updateGravityAndDrag()

{
    meanAnDot = meanMotDeep + 0.5 * clc1 * beta0 * x3thm1 + 
                0.0625 * clc2 * beta0 * 
                (13.0 - 78.0 * cos2Inc + 137.0 * cos4Inc);

    argPerDot = -0.5 * clc1 * x1m5th + 0.0625 * clc2 * 
                (7.0 - 114.0 * cos2Inc + 395.0 * cos4Inc) + 
                clc3 * (3.0 - 36.0 * cos2Inc + 49.0 * cos4Inc);

    raanDot   = -clc1 * cosInc + (0.5 * clc2 * (4.0 - 19.0 * cos2Inc) + 
                2.0 * clc3 * (3.0 -  7.0 * cos2Inc)) * cosInc;

    meanAnDF  = meanAn0 + meanAnDot * tFP;
    argPerDF  = argPer0 + argPerDot * tFP;
    raanDF    = raan0   + raanDot   * tFP;

    delA      = 1.0 - c1 * tFP;
    delE      = bStar * c4 * tFP;
    delL      = 1.5 * c1 * tSQ;

    if (!truncFlag)
    {
        c1SQ       = SQR(c1);
        d2         = 4.0 * sma0Deep * xi * c1SQ;
        clc0       = d2 * xi * c1 / 3.0;
        d3         = (17.0 * sma0Deep + sStar) * clc0;
        d4         = 0.5 * clc0 * sma0Deep * xi * 
                     (221.0 * sma0Deep + 31.0 * sStar) * c1;

        t3Cof      = d2 + 2.0 * c1SQ;
        t4Cof      = 0.25 * (3.0 * d3 + c1 * (12.0 * d2 + 10.0 * c1SQ));
        t5Cof      = 0.2 * (3.0 * d4 + 12.0 * c1 * d3 + 
                     6.0 * SQR(d2) + 15.0 * c1SQ * (2.0 * d2 + c1SQ));

        argPerCof  = bStar * c3 * cos(argPer0);
        meanAnCof  = -TWOTHIRDS * coef0 * bStar * AE / eccEta;
        raanCof    = -3.5 * beta02 * clc1 * cosInc * c1;

        delArgPer  = argPerCof * tFP;
        delMeanAn  = meanAnCof * (pow(1.0 + eta * cos(meanAnDF),3.0) - 
                                  pow(1.0 + eta * cos(meanAn0),3.0));

        delAM      = delArgPer + delMeanAn;
        meanAnP    = meanAnDF + delAM;
        argPer     = argPerDF - delAM;

        sinMeanAn0 = sin(meanAn0);
        sinMeanAnP = sin(meanAnP);

        delA      -= d2 * tSQ - d3 * tCB - d4 * tQD;
        delE      += bStar * c5 * (sinMeanAnP - sinMeanAn0);
        delL      += t3Cof * tCB + t4Cof * tQD + t5Cof * tQN;
    }

    else
    {
        meanAnP    = meanAnDF;
        argPer     = argPerDF;
    }

    raan    = raanDF - raanCof * tSQ;

    smallE  = ecc0 - delE;
    smallA  = sma0Deep * SQR(delA);

    xL      = meanAnP + argPer + raan + meanMotDeep * delL;
    beta    = sqrt(1.0 - SQR(smallE));
    meanMot = KE / pow(smallA,1.5);

    perigeeHeight  = (smallA * (1.0 - ecc0) - AE) * EARTHRADIUS;   /* [km]    */
    curArgPerigeeX = argPer;                                       /* [rad]   */
    curRaanX       = raan;                                         /* [rad]   */
    curMotionX     = meanMot * MPD / TWOPI;                        /* [rev/d] */

    return;
}

/******************************************************************************/
/*                                                                            */
/* updateLongPeridics: updates long periodic quantities                       */
/*                                                                            */
/******************************************************************************/

void updateLongPeriodics()

{
    upd6  = 1.0 / (smallA * SQR(beta));

    axN   = smallE * cos(argPer);

    xLCof = 0.125 * A30CK2 * sinInc * (3.0 + 5.0 * cosInc) / (1.0 + cosInc);
    xLL   = upd6 * xLCof * axN;
    xLT   = xL + xLL;

    ayCof = 0.250 * A30CK2 * sinInc;
    ayNL  = upd6 * ayCof;
    ayN   = smallE * sin(argPer) + ayNL;

    return;
}

/******************************************************************************/
/*                                                                            */
/* updateShortPeridics: gets osculating quantities by adding delta terms      */
/*                                                                            */
/******************************************************************************/

void updateShortPeriodics()

{
    eCosE    = kep4 + kep5;
    eSinE    = kep2 - kep3;

    eL2      = SQR(axN) + SQR(ayN);
    upd1     = 1.0 - eL2;
    pL       = smallA * upd1;
    betaL    = sqrt(upd1);

    smallR   = smallA * (1.0 - eCosE);

    rDot     = KE * sqrt(smallA) / smallR * eSinE;
    rfDot    = KE * sqrt(pL) / smallR;

    upd2     = smallA / smallR;
    upd3     = 1.0 / (1.0 + betaL);

    cosU     = upd2 * (cosEpAP - axN + ayN * eSinE * upd3);
    sinU     = upd2 * (sinEpAP - ayN - axN * eSinE * upd3);

    sin2U    = 2.0 * sinU * cosU;
    cos2U    = 2.0 * cosU * cosU - 1.0;

    smallU   = atan2(sinU,cosU);
    smallU   = reduce(smallU,ZERO,TWOPI);

    upd4     = CK2 / SQR(pL);
    upd5     = CK2 / pL;

    delR     = 0.5 * upd5 * x1m1th * cos2U;
    delU     = -0.25 * upd4 * x7thm1 * sin2U;

    delRaan  = 1.5 * upd4 * cosInc * sin2U;
    delInc   = 1.5 * upd4 * cosInc * sinInc * cos2U;

    delRDot  = -meanMot * upd5 * x1m1th * sin2U;
    delRFDot = meanMot * upd5 * (x1m1th * cos2U + 1.5 * x3thm1);

    rk       = smallR * (1.0 - 1.5 * upd4 * betaL * x3thm1) + delR;
    uk       = smallU + delU;
    raanK    = raan   + delRaan;
    incK     = inc0   + delInc;
    rkDot    = rDot   + delRDot;
    rfkDot   = rfDot  + delRFDot;

    curArgNodeX = uk;

    return;
}

/******************************************************************************/
/*                                                                            */
/* keplerX: solve Kepler's equation                                           */
/*                                                                            */
/******************************************************************************/

void keplerX()

{
    register double rCapU, rCapEpAP, rSinEpAP, rCosEpAP, diffAnom, eccAnom;
    register double rKep1, rKep2, rKep3, rKep4, rKep5;

    rCapU    = reduce(xLT - raan,ZERO,TWOPI);
    rCapEpAP = rCapU;

    do
    {
        rKep1    = rCapEpAP;
        rSinEpAP = sin(rKep1);
        rCosEpAP = cos(rKep1);

        rKep2    = axN * rSinEpAP;
        rKep3    = ayN * rCosEpAP;
        rKep4    = axN * rCosEpAP;
        rKep5    = ayN * rSinEpAP;

        rCapEpAP = rKep1 + (rCapU - rKep1 + rKep2 - rKep3) / 
                           (1.0 - rKep4 - rKep5);

        diffAnom = rCapEpAP - rKep1;

    } while (fabs(diffAnom) >= ONEPPM);               /* 0.2 arcsec precision */

    sinEpAP = rSinEpAP;
    cosEpAP = rCosEpAP;

    kep1    = rKep1;
    kep2    = rKep2;
    kep3    = rKep3;
    kep4    = rKep4;
    kep5    = rKep5;

    eccAnom = rCapEpAP - argPer;

    if (fabs(eccAnom) < ONEPPM)
        trueAnomalyX = PI;
    else
        trueAnomalyX = 2.0 * atan(sqrt((1.0 + ecc0) /
                            (1.0 - ecc0)) * tan(eccAnom / 2.0));

    trueAnomalyX = reduce(trueAnomalyX,ZERO,TWOPI);

    return;
}

/******************************************************************************/
/*                                                                            */
/* getOrbitNumberX: gets orbit number and mean anomaly                        */
/*                                                                            */
/******************************************************************************/

void getOrbitNumberX(tArg)

double tArg;

{
    double dT, orbFract;

    dT          = tArg - epoch0;
    deltaTimeX  = dT;

    refOrbit    = (double) epochOrbitNum + epochMeanAnomaly / TWOPI; /* [rev] */
    curOrbit    = refOrbit + curMotionX * dT 
                           - (curMotionX - epochMeanMotion) * dT;

    orbitNum    = (long) curOrbit + 1L;

    orbFract    = modf(curOrbit,&dummyd);
    orbitFract  = orbFract * 100.0;
    meanAnomaly = orbFract * TWOPI;

    return;
}

/******************************************************************************/
/*                                                                            */
/* getStateVectorX: calculates position and velocity components               */
/*                                                                            */
/******************************************************************************/

void getStateVectorX()

{
    double mVec[3], nVec[3], uVec[3], vVec[3];
    double cosIK, cosRK, cosUK, sinIK, sinRK, sinUK;
    int    i;

    cosUK   = cos(uk);
    sinUK   = sin(uk);
    cosIK   = cos(incK);
    sinIK   = sin(incK);
    cosRK   = cos(raanK);
    sinRK   = sin(raanK);

    mVec[0] = -sinRK * cosIK;
    mVec[1] = cosRK * cosIK;
    mVec[2] = sinIK;

    nVec[0] = cosRK;
    nVec[1] = sinRK;
    nVec[2] = 0.0;

    for (i = 0; i <= 2; i++)
    {
        uVec[i]     = mVec[i] * sinUK + nVec[i] * cosUK;
        vVec[i]     = mVec[i] * cosUK - nVec[i] * sinUK;

        satPosS[i]  = rk * uVec[i];
        satVelS[i]  = rkDot * uVec[i] + rfkDot * vVec[i];

        satPosS[i] *= EARTHRADIUS;
        satVelS[i] *= EARTHRADIUS / 60.0;

        satPosGl[i] = satPosS[i];
        satVelGl[i] = satVelS[i];
    }

    return;
}

/******************************************************************************/
/*                                                                            */
/* calcSGP4: calculates SGP4 propagation model                                */
/*                                                                            */
/* For perigee less than 220 km, the truncFlag is set and the equations for   */
/* 'smallA' and 'xL' are truncated after the 'c1' term, and the terms         */
/*  involving 'c5', 'deltaArgPer' and 'deltaMeanAn' are dropped.              */
/*                                                                            */
/* For a perigee below 156 km, the values of S and Q0MS0T4 are altered.       */
/* (see 'checkPerigeeHeight()')                                               */
/*                                                                            */
/******************************************************************************/

void calcSGP4(tmArg)

double tmArg;

{
    checkPerigeeHeight();

    if (satCrashFlag)
        return;

    getTimeArgs(tmArg);
    calcCommonPerturb();
    updateGravityAndDrag();
    updateLongPeriodics();
    keplerX();
    updateShortPeriodics();
    getStateVectorX();
    getOrbitNumberX(tmArg);
    return;
}

/******************************************************************************/
/*                                                                            */
/* initDeepSpace:                                                             */
/*                                                                            */
/******************************************************************************/

void initDeepSpace()

{
    return;
}

/******************************************************************************/
/*                                                                            */
/* updateGravity:                                                             */
/*                                                                            */
/******************************************************************************/

void updateGravity()

{
    return;
}

/******************************************************************************/
/*                                                                            */
/* updateDrag:                                                                */
/*                                                                            */
/******************************************************************************/

void updateDrag()

{
    return;
}

/******************************************************************************/
/*                                                                            */
/* deepSpaceSecular:                                                          */
/*                                                                            */
/******************************************************************************/

void deepSpaceSecular()

{
    return;
}

/******************************************************************************/
/*                                                                            */
/* deepSpacePeriodics:                                                        */
/*                                                                            */
/******************************************************************************/

void deepSpacePeriodics()

{
    return;
}

/******************************************************************************/
/*                                                                            */
/* calcSDP4:                                                                  */
/*                                                                            */
/******************************************************************************/

void calcSDP4(tmArg)

double tmArg;

{
    checkPerigeeHeight();

    if (satCrashFlag)
        return;

    getTimeArgs(tmArg);
    calcCommonPerturb();
    initDeepSpace();
    updateGravity();
    updateDrag();
    deepSpaceSecular();
    updateLongPeriodics();
    keplerX();
    updateShortPeriodics();
    getStateVectorX();
    getOrbitNumberX(tmArg);
    return;
}

/******************************************************************************/
/*                                                                            */
/* getNoradStateVector: calculates state vector following the NORAD SGP4/SDP4 */
/*                      propagation models                                    */
/*                                                                            */
/* uses deep space model if orbital period is longer than 225 minutes; all    */
/* time arguments are in minutes                                              */
/*                                                                            */
/******************************************************************************/

void getNoradStateVector(timeArg)

double timeArg;

{
    if (modelFlag)                 /* TLE Mean Model already calculated       */
        return;                    /* else state vector is overwritten        */

    getSemiMajorAxis();

    if (deepSpaceFlag)
        return;                    /* for now; eventually call calcSDP4()     */
    else
        calcSGP4(timeArg);         /* overwrites TLE Mean Model state vector  */

    return;
}

/******************************************************************************/
/*                                                                            */
/* end of function block satprop.c                                            */
/*                                                                            */
/******************************************************************************/
