/******************************************************************************/
/*                                                                            */
/*  Title       : satread.c                                                   */
/*  Author      : Manfred Bester, DL5KR                                       */
/*  Date        : 03Mar92                                                     */
/*  Last change : 05Oct94                                                     */
/*                                                                            */
/*  Synopsis    : Routines for reading data from resource files and other     */
/*                I/O functions.                                              */
/*                                                                            */
/*                                                                            */
/*  Copyright (c) 1992, 1993, 1994 Manfred Bester. All Rights Reserved.       */
/*                                                                            */
/*  Permission to use, copy, modify, and distribute this software and its     */
/*  documentation for educational, research and non-profit purposes, without  */
/*  fee, and without a written agreement is hereby granted, provided that the */
/*  above copyright notice and the following three paragraphs appear in all   */
/*  copies.                                                                   */
/*                                                                            */
/*  Permission to incorporate this software into commercial products may be   */
/*  obtained from the author, Dr. Manfred Bester, 1636 M. L. King Jr. Way,    */
/*  Berkeley, CA 94709, USA.                                                  */
/*                                                                            */
/*  IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, */
/*  SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF   */
/*  THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE AUTHOR HAS BEEN ADVISED  */
/*  OF THE POSSIBILITY OF SUCH DAMAGE.                                        */
/*                                                                            */
/*  THE AUTHOR SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT      */
/*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A   */
/*  PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS"      */
/*  BASIS, AND THE AUTHOR HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT, */
/*  UPDATES, ENHANCEMENTS, OR MODIFICATIONS.                                  */
/*                                                                            */
/*                                                                            */
/*  Resources   : cities.dat, defaults.dat, modes.dat, satlist_nnn.dat,       */
/*                tlex.dat                                                    */
/*                                                                            */
/*  cities.dat:      contains information on various gorund stations          */
/*                   format: `name`,lat,long,alt                              */
/*                                                                            */
/*  defaults.dat:    contains default parameters                              */
/*                                                                            */
/*  modes.dat:       contains mode information for amateur satellites         */
/*                                                                            */
/*  satlist_nnn.dat: contains list of satellite names for multisat tracking   */
/*                   display ('nnn' denotes the satellite group)              */
/*                                                                            */
/*  tle.dat:         contains NORAD two-line orbital element sets             */
/*                   (also sts.dat, sts-55.dat, tvsat.dat, etc.)              */
/*                                                                            */
/*  tlex.dat:        contains concatenated NORAD two-line orbital element     */
/*                   sets                                                     */
/*                                                                            */
/******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "satglobalsx.h"
#include "sattrack.h"
#include "vt100.h"

#ifndef SUNOS4
#include <stdlib.h>
#endif

/******************************************************************************/
/*                                                                            */
/* checkTimeZone: enter time zone                                             */
/*                                                                            */
/******************************************************************************/

void checkTimeZone()

{
    if (timeZoneFlag)
    {
        nl();
        printf("Time zone                      ");
        printf("<%4s> : ",defaultTimeZoneStr);
        gets(timeZoneStr);
        upperCase(timeZoneStr);

        if (!strlen(timeZoneStr))
            strcpy(timeZoneStr,defaultTimeZoneStr);

        getTimeZone();
        printf("Time zone set to %s (%+.1f h)\n",
                timeZoneStr,timeZone*24.0);
    }

    if (batchModeFlag)
    {
        strcpy(timeZoneStr,batchTimeZone);
        getTimeZone();
    }

    return;
}

/******************************************************************************/
/*                                                                            */
/* getTimeZone: gets time zone [h] from the time zone string information      */
/*                                                                            */
/******************************************************************************/

void getTimeZone()

{
    int i;

    timeZone = 0.0;

    for (i = 0; i < timeZones; i++)
    {
        if (!strcmp(timeZoneList[i],timeZoneStr))
            timeZone = timeZoneHour[i] / 24.0;          /* gets time zone [d] */
    }

    if (fabs(timeZone) < ONEPPM)
    {
        if (strlen(timeZoneStr) && fabs((double) atof(timeZoneStr)) < 12.000001
                                && fabs((double) atof(timeZoneStr)) >  0.000001)
        {
            timeZone = (double) atof(timeZoneStr) / 24.0;
            sprintf(timeZoneStr,"LOC");
        }

        else
            sprintf(timeZoneStr,"UTC");
    }

    return;
}

/******************************************************************************/
/*                                                                            */
/* getDefaultData: reads default parameters from 'defaults.dat'               */
/*                                                                            */
/******************************************************************************/

void getDefaultData()

{
    char defaultLine[120], defaultDuration[60], defaultMinElev[60];
    char defaultTracking[60], defaultTerminal[60], defAntenna[60], defRadio[60];
    char dispLines[60], dispLinesX[60];
    FILE *defaultFile;

    sprintf(defaultsDat,"%s/%s/%s",strpHome,DATA,DEFAULTS);

    if ((defaultFile = fopen(defaultsDat,"r")) == 0)
    {
        alarm(); nl(); reverseBlink();
        printf(" Cannot open input file '%s' \n",defaultsDat);
        normal(); nl();
        exit(-1);
    }

    sprintf(callSign,"%s",CALLSIGN);
    sprintf(defaultSite,"the North Pole");
    sprintf(defaultSat,"Mir");
    sprintf(defaultSatGroup,"a");
    sprintf(defaultSetTLE,"000");
    sprintf(defaultSetSTS,"000");
    sprintf(defaultTimeZoneStr,"UTC");
    sprintf(defaultDuration,"6.0");
    sprintf(defaultMinElev,"0.0");
    sprintf(defaultTracking,"0");
    sprintf(antennaIO,"%s",ANTENNADEV);
    sprintf(radioIO,"%s",RADIODEV);
    sprintf(dispLines,"%d",DISPLINES);
    sprintf(dispLinesX,"%d",XDISPLINES);
    sprintf(termTypeX,"%s",XTERM);

    atmPressure    = ATMPRESS;
    ambTemperature = AMBTEMP;
    relHumidity    = RELHUMID;

    while (fgets(defaultLine,100,defaultFile))
    {
        defaultLine[strlen(defaultLine) - 1] = '\0';
        truncBlanks(defaultLine);

        if (!strncmp(defaultLine,"Site:",5))
        {
           clipString(defaultLine,5);
           sprintf(defaultSite,"%s",defaultLine);
        }

        if (!strncmp(defaultLine,"Callsign:",9))
        {
           clipString(defaultLine,9);
           sprintf(callSign,"%s",defaultLine);
        }

        if (!strncmp(defaultLine,"Satellite:",10))
        {
           clipString(defaultLine,10);
           sprintf(defaultSat,"%s",defaultLine);
        }

        sscanf(defaultLine,"Satellite group: %s",defaultSatGroup);
        sscanf(defaultLine,"Element set TLE: %s",defaultSetTLE);
        sscanf(defaultLine,"Element set STS: %s",defaultSetSTS);
        sscanf(defaultLine,"Element set type: %s",defaultType);
        sscanf(defaultLine,"Time zone: %s",defaultTimeZoneStr);
        sscanf(defaultLine,"Duration: %s",defaultDuration);
        sscanf(defaultLine,"Min elevation: %s",defaultMinElev);
        sscanf(defaultLine,"Tracking: %s",defaultTracking);
        sscanf(defaultLine,"Antenna: %s",antennaIO);
        sscanf(defaultLine,"Radio: %s",radioIO);
        sscanf(defaultLine,"Display lines: %s",dispLines);
        sscanf(defaultLine,"X display lines: %s",dispLinesX);
        sscanf(defaultLine,"X terminal type: %s",termTypeX);
        sscanf(defaultLine,"Pressure: %lf",&atmPressure);
        sscanf(defaultLine,"Temperature: %lf",&ambTemperature);
        sscanf(defaultLine,"Humidity: %lf",&relHumidity);
    }

    strcpy(timeZoneStr,defaultTimeZoneStr);
    getTimeZone();
    defDuration     = atof(defaultDuration);
    defMinElevation = atof(defaultMinElev);
    trackingFlag    = atoi(defaultTracking);
    numLines        = atoi(dispLines);
    numLinesX       = atoi(dispLinesX);

    fclose(defaultFile);

    if (verboseFlag)
    {
        nl();
        printf("Default parameters:\n\n");
        printf("Ground station    : %19s\n",defaultSite);
        printf("Call sign         : %19s\n",callSign);
        printf("Satellite         : %19s\n",defaultSat);
        printf("Element set TLE   : %19s\n",defaultSetTLE);
        printf("Element set STS   : %19s\n",defaultSetSTS);
        printf("Element set type  : %19s\n",defaultType);
        printf("Time zone         : %19s (%+.1f h)\n",
                timeZoneStr,timeZone*24.0);
        printf("Duration          : %19.1f d\n",defDuration);
        printf("Min elevation     : %19.1f deg\n",defMinElevation);
        printf("Tracking          : %19d\n",trackingFlag);
        printf("Antenna interface : %19s\n",antennaIO);
        printf("Radio interface   : %19s\n",radioIO);
        printf("Display lines     : %19s\n",dispLines);
        printf("X display lines   : %19s\n",dispLinesX);
        printf("X terminal type   : %19s\n",termTypeX);
        printf("Pressure          : %19.1f mmHg\n",atmPressure);
        printf("Temperature       : %19.1f C\n",ambTemperature);
        printf("Humidity          : %19.1f ",relHumidity);
        putchar('%');
        nl();
        nl();
    }

    return;
}

/******************************************************************************/
/*                                                                            */
/* getSatFileName: gets reduced satellite name for creating output file, i.e. */
/*                 replace blanks and "/" in file name with "_", delete "("   */
/*                 and ")"                                                    */
/*                                                                            */
/******************************************************************************/

void getSatFileName()

{
    int i, j;

    i = 0;
    j = 0;

    do
    {
        if (satName[i] != ' ' && satName[i] != '(' &&
            satName[i] != ')' && satName[i] != '"' &&
            satName[i] != '/')
        {
            satFileName[j] = satName[i];
            i++; j++;
        }

        if (satName[i] == ' ' || satName[i] == '/')
        {
            satFileName[j] = '_';
            i++; j++;
        }

        if (satName[i] == '(' || satName[i] == ')' || satName[i] == '"')
        {
            i++;
        }
    }
    while (satName[i] != '\0');

    satFileName[i] = '\0';
    return;
}

/******************************************************************************/
/*                                                                            */
/* getSatParams: reads satellite parameters from data file with the two-line  */
/*               orbital element sets                                         */
/*                                                                            */
/******************************************************************************/

void getSatParams()

{
    int    error, error1;
    char   str[80], stsName[20];

    error = TRUE;

    while (error)
    {
        if (batchModeFlag)
            strcpy(satName,batchSatName);

        else
        {
            if (quickStartFlag)
                sprintf(satName,"");

            else
            {
                printf("Satellite name <%20s> : ",defaultSat);
                gets(satName);
            }
        }

        if (!strlen(satName))
            sprintf(satName,"%s",defaultSat);

        cleanSatName(satName);
        strcpy(stsName,satName);
        upperCase(stsName);

        stsFlag = !strncmp(stsName,"STS",3) ? TRUE : FALSE;
        sprintf(defaultNORAD,"%s",(stsFlag) ? defaultSetSTS : defaultSetTLE);

        if (batchModeFlag)
            strcpy(elementFile,batchTleFile);

        else
        {
            if (quickStartFlag)
                sprintf(elementFile,"");

            else
            {
                printf("Two-line elements <%17s> : ",defaultNORAD);
                gets(elementFile);
            }
        }

        if (!strlen(elementFile))
            sprintf(elementFile,"%s",defaultNORAD);

        if (strcmp(TLEX,elementFile))
        {
            error = readTleFile(strpHome,elementFile,numTle);

            if (error)
            {
                alarm(); nl(); reverseBlink();
                printf(" File '%s' not found \n",elementFile);
                normal(); nl();

                if (batchModeFlag || quickStartFlag)
                    exit(-1);
            }
        }

        error = readTle(TRUE,satName,&satNum,&elementSet,
                   &epochDay,&inclination,&epochRaan, &eccentricity,
                   &epochArgPerigee,&epochMeanAnomaly,&epochMeanMotion,
                   &decayRate,&decayRateDot,&bStarCoeff,
                   &epochOrbitNum,&ephemerisType);

        if (error)
        {
            alarm(); nl(); reverseBlink();

            if (error == 1 || error == 2)
                printf(" %s: check sum error in line %d \n",satName,error);

            if (error == 3)
                printf(" %s: no matching satellite name \n",satName);

            if (error == 4)
                printf(" Specify one satellite \n");

            normal(); nl();

            if (batchModeFlag)
                exit(-1);
        }
    }

    strcpy(stsName,satName);
    upperCase(stsName);

    stsFlag = !strncmp(stsName,"STS",3) ? TRUE : FALSE;    /* restore stsFlag */

    if (stsFlag)
    {
        error1 = TRUE;

        while (error1)
        {
            if (batchModeFlag)
                strcpy(elementType,batchTleType);

            else
            {
                if (quickStartFlag)
                    sprintf(elementType,"");

                else
                {
                    printf("Element set type              ");
                    printf("<%5s> : ",defaultType);
                    gets(elementType);
                }
            }

            if (!strlen(elementType))
                sprintf(elementType,"%s",defaultType);
            else
                upperCase(elementType);

            error1 = (!strcmp(elementType,"NASA") || 
                      !strcmp(elementType,"NORAD")) ? FALSE : TRUE;

            if ((batchModeFlag || quickStartFlag) && error1)
            {
                alarm(); nl(); reverseBlink();
                printf(" Type of element set specified incorrectly \n");
                normal(); nl();
                exit(-1);
            }
        }
    }

    getSatFileName();                                   /* remove blanks etc. */
    changeElementUnits();
    checkPropModelType();

    if (realTime-elsetEpoch > 0.0)
        sprintf(updateString,"%.1f days ago",realTime-elsetEpoch);
    else
        sprintf(updateString,"%.1f days ahead",elsetEpoch-realTime);

    if (verboseFlag)
    {
        nl();
        printf("Satellite name    : %19s\n",satName);
        printf("Satellite number  : %19ld\n",satNum);
        printf("Element set       : %19d\n",elementSet);
        printf("Ephemeris type    : %19d\n",ephemerisType);
        printf("Epoch             : %19.9f d        ",elsetEpochDay);
        printf("%s\n",epochString);
        printf("Mean anomaly      : %19.9f deg\n",epochMeanAnomaly * CRD);
        printf("Arg of perigee    : %19.9f deg\n",epochArgPerigee * CRD);
        printf("RAAN              : %19.9f deg\n",epochRaan * CRD);
        printf("Inclination       : %19.9f deg\n",inclination * CRD);
        printf("Eccentricity      : %19.9f\n",eccentricity);
        printf("Semimajor axis    : %19.9f km\n",semiMajorAxis);
        printf("Perigee height    : %19.9f km\n",perigeeHeight);
        printf("Mean motion       : %19.9f rev/d\n",epochMeanMotion);
        printf("Decay rate        : %19.9f rev/d^2\n",decayRate);
        printf("Decay rate dot    : %19.9f rev/d^3\n",decayRateDot);
        printf("BSTAR drag coeff. : %19.9f\n",bStarCoeff);
        printf("Orbit             : %19ld\n",epochOrbitNum);
        printf("Propagation model : %19s\n",propModel);
        nl();

        if (geoSyncFlag)
        {
            if (geoSyncFlag == GEOSYNC)
                printf("Geosynchronous satellite\n");
            else
                printf("Geostationary satellite\n");

            nl();
        }

        if (satCrashFlag || tleInvalidFlag)
        {
            reverseBlink();

            if (satCrashFlag)
                printf(" Satellite has crashed already \n");

            if (tleInvalidFlag)
                printf(" TLE set is out of date \n");

            normal();
            nl();
        }
    }

    else
    {
        if (!batchModeFlag && !quickStartFlag)
        {
            sprintf(str,"%ld, %s",satNum,satName);
            printf("Last update for %-21s : ",str);
            printf("%s (#%d, Orbit %ld)\n\n",
                    updateString,elementSet,epochOrbitNum);
        }
    }

    return;
}

/******************************************************************************/
/*                                                                            */
/* changeElementUnits: changes units of Keplerian elements etc.               */
/*                                                                            */
/******************************************************************************/

void changeElementUnits()

{
    double epochDayTime, epochH, epochM, epochS, epochSec;
    int    epochYear, epochMonth, epochMonthDay, epochD, epochHour, epochMin;

    elsetEpochDay = epochDay;
    epochYear     = (int) (epochDay  / 1000.0);
    epochDay     -= epochYear * 1000.0;
    elsetEpoch    = epochDay + getDayNum(0,1,epochYear);
    epochD        = (int) epochDay;
    epochDayTime  = epochDay - (double) epochD;
    epochDay     += getDayNum(0,1,epochYear);

    epochH        = 24.0 * epochDayTime;
    epochM        = 60.0 * modf(epochH,&epochH);
    epochS        = 60.0 * modf(epochM,&epochM);

    if (fabs(epochS - 60.0) < 5.0e-4)
    {
        epochS  = 0.0;
        epochM += 1.0;

        if (fabs(epochM - 60.0) < 1.0e-4)
        {
            epochM  = 0.0;
            epochH += 1.0;

            if (fabs(epochH - 24.0) < 1.0e-4)
            {
                epochH  = 0.0;
                epochD += 1;
            }
        }
    }

    epochHour = (int) epochH;
    epochMin  = (int) epochM;
    epochSec  = epochS;

    calendar(epochYear,epochD,&epochMonth,&epochMonthDay);

    sprintf(epochString,"%02d%3s%02d  %02d:%02d:%06.3f UTC",
            epochMonthDay,monthName(epochMonth),epochYear,
            epochHour,epochMin,epochSec);

    inclination      *= CDR;
    epochRaan        *= CDR;
    epochArgPerigee  *= CDR;
    epochMeanAnomaly *= CDR;

    cosInclination    = cos(inclination);
    sinInclination    = sin(inclination);

    geoSyncFlag = (fabs(epochMeanMotion - 1.0) < GEOSYNCMOT) ? GEOSYNC : FALSE;

    if (geoSyncFlag)
        geoSyncFlag = (fabs(inclination) < GEOSTATINC) ? GEOSTAT : GEOSYNC;

    return;
}

/******************************************************************************/
/*                                                                            */
/* getSatModes: reads modes of the specified satellite                        */
/*                                                                            */
/******************************************************************************/

void getSatModes(satelliteName)

char *satelliteName[];

{
    double cosAttLat, sinAttLat, cosAttLong, sinAttLong;
    long   objNum, objNumber;
    int    readLines, foundSat, foundMode, gotMode;
    char   str[40], strSat[120], modeLine[120], lDate[12];
    FILE   *modeFile;

    foundSat     = FALSE;
    foundMode    = FALSE;
    gotMode      = FALSE;
    launchFlag   = FALSE;
    launchEpoch  = 0.0;
    attitudeFlag = FALSE;
    beaconFreq   = BEACONFREQ;
    attLat       = 0.0;
    attLong      = 0.0;
    perigeePhase = 0.0;
    maxPhase     = MAXPHASE;
    numModes     = 0;
    objNum       = 0L;
    readLines    = 0;

    sprintf(satAlias,"");
    sprintf(fullSatAlias,"");
    sprintf(modesDat,"%s/%s/%s",strpHome,DATA,MODES);
    sprintf(str,"%s",satelliteName);
    objNumber = atol(str);

    if ((modeFile = fopen(modesDat,"r")) != 0)
    {
        while (fgets(modeLine,100,modeFile) && !gotMode)
        {
            modeLine[strlen(modeLine) - 1] = '\0';
            truncBlanks(modeLine);

            if (!foundMode)
            {
                if (!strncmp(modeLine,"Satellite:",10))
                {
                    clipString(modeLine,10);
                    sprintf(strSat,"%s",modeLine);
                }

                if (!strncmp(modeLine,"Object:",7))
                {
                    clipString(modeLine,7);
                    sscanf(modeLine,"%ld",&objNum);
                }

                if (!strncmp(str,strSat,strlen(strSat)) || 
                    objNum == objNumber)
                {
                    readLines++;

                    if (objNum != objNumber)
                        readLines = 2;

                    foundSat = TRUE;
                }

                if (foundSat && readLines == 2)
                    foundMode = TRUE;
            }

            if (foundMode)
            {
                if (!strncmp(modeLine,"Alias:",6))
                {
                    clipString(modeLine,6);
                    sprintf(fullSatAlias,"%s",modeLine);
                    strncpy(satAlias,fullSatAlias,15);

                    if (verboseFlag)
                        printf("Alias name        : %19s\n",fullSatAlias);
                }

                if (!strncmp(modeLine,"Launch:",7))
                {
                    clipString(modeLine,7);
                    sscanf(modeLine,"%s %d:%d:%d",
                        lDate,&launchHour,&launchMin,&launchSec);

                    sprintf(launchString,"%s  %02d:%02d:%02d.000 UTC",
                        lDate,launchHour,launchMin,launchSec);

                    dummyi = decodeDate(lDate,
                                 &launchDay,&launchMonth,&launchYear);

                    launchFlag = TRUE;

                    if (verboseFlag)
                        printf("Launch date/time  :   %s\n",launchString);
                }

                if (!strncmp(modeLine,"Beacon:",7))
                {
                    clipString(modeLine,7);
                    sscanf(modeLine,"%lf",&beaconFreq);

                    if (verboseFlag)
                        printf("Beacon frequency  : %19.6f MHz\n",beaconFreq);
                }

                if (!strncmp(modeLine,"Perigee phase:",14))
                {
                    clipString(modeLine,14);
                    sscanf(modeLine,"%lf",&perigeePhase);

                    if (verboseFlag)
                        printf("Perigee phase     : %19.6f MHz\n",perigeePhase);
                }

                if (!strncmp(modeLine,"Attitude:",9))
                {
                    clipString(modeLine,9);
                    sscanf(modeLine,"%lf %lf",&attLong,&attLat);
                    attitudeFlag = TRUE;

                    if (verboseFlag)
                    {
                        printf("Attitude          :    ");
                        printf("%8.3f %7.3f deg\n",attLong,attLat);
                    }
                }

                if (!strncmp(modeLine,"Max phase:",10))
                {
                    clipString(modeLine,10);
                    sscanf(modeLine,"%lf",&maxPhase);

                    if (verboseFlag)
                        printf("Max phase         : %19.6f\n",maxPhase);
                }

                if (!strncmp(modeLine,"Mode:",5) && numModes < MAXMODES)
                {
                    clipString(modeLine,5);

                    sscanf(modeLine,"%s %lf - %lf",
                        modes[numModes].modeStr,
                        &modes[numModes].minPhase,
                        &modes[numModes].maxPhase);

                    if (verboseFlag)
                        printf("Mode              :  %6s   %3.0f - %3.0f\n",
                        modes[numModes].modeStr,
                        modes[numModes].minPhase,
                        modes[numModes].maxPhase);

                    numModes++;
                }

                if (strlen(modeLine) <= 2)
                    gotMode = TRUE;
            }
        }

        fclose(modeFile);

        if (verboseFlag) nl();

        beaconFreq /= CHZMHZ;

        if (launchFlag)
        {
            launchDate  = getDayNum(launchDay,launchMonth,launchYear);
            launchTime  = (double) (launchHour+launchMin/60.0+launchSec/3600.0);
            launchEpoch = (double) launchDate+launchTime/24.0;
        }

        if (attitudeFlag)
        {
            attLat     *= CDR;
            attLong    *= CDR;

            cosAttLat   = cos(attLat);
            sinAttLat   = sin(attLat);
            cosAttLong  = cos(attLong);
            sinAttLong  = sin(attLong);

            satAttGl[0] = cosAttLong * cosAttLat;
            satAttGl[1] = sinAttLong * cosAttLat;
            satAttGl[2] = sinAttLat;
        }
    }

    return;
}

/******************************************************************************/
/*                                                                            */
/* getSiteParams: reads site parameters from input file 'cities.dat'          */
/*                                                                            */
/******************************************************************************/
 
void getSiteParams()

{
    int  i, k, siteNameLen, foundSite;
    char siteNameStr[80], saveStr[80], ctyStr[80], str[80];

    foundSite = FALSE;

    while (!foundSite)
    {
        if (batchModeFlag)
            strcpy(siteNameStr,batchSiteName);

        else
        {
            if (quickStartFlag)
                sprintf(siteNameStr,"");

            else
            {
                nl();
                printf("Ground station <%20s> : ",defaultSite);
                gets(siteNameStr);
            }
        }

        if (!strlen(siteNameStr))
            sprintf(siteNameStr,"%s",defaultSite);

        upperCase(siteNameStr);
        siteNameLen = strlen(siteNameStr);

        k = 0;

        for (i = 0; i < numCities; i++)        /* check for ambiguous entries */
        {
            if (k == 1)
                strcpy(saveStr,ctyStr);

            strcpy(str,city[i].cty);
            upperCase(str);

            if (!strncmp(siteNameStr,str,siteNameLen))
            {
                strcpy(ctyStr,city[i].cty);

                if (k == 1)
                    printf("\n%s\n",saveStr);

                if (k >= 1)
                    printf("%s\n",city[i].cty);

                k++;
            }
        }

        i = 0;

        do
        {
            strcpy(str,city[i].cty);
            upperCase(str);

            if (k == 1 && !strncmp(str,siteNameStr,siteNameLen))
            {
                foundSite = TRUE;

                strcpy(fullSiteName,city[i].cty);
                strncpy(siteName,city[i].cty,15);
                shortString(siteName,fullSiteName);

                siteLat     = city[i].lat;
                siteLong    = city[i].lng;
                siteAlt     = city[i].alt * CMKM;

                cosSiteLat  = cos(siteLat);
                sinSiteLat  = sin(siteLat);
                cosSiteLong = cos(siteLong);
                sinSiteLong = sin(siteLong);

                siteCtyVecGl[0] = siteAlt*cosSiteLong*cosSiteLat;
                siteCtyVecGl[1] = siteAlt*sinSiteLong*cosSiteLat;
                siteCtyVecGl[2] = siteAlt*sinSiteLat;
            }

            i++;
        }
        while (i < numCities && !foundSite);
 
        if (!foundSite)
        {
            alarm(); nl(); reverseBlink();

            if (k == 0)
                printf(" %s: no matching ground station ",siteNameStr);
            else
                printf(" Specify one ground station ");

            normal(); nl();

            if (batchModeFlag)
            {
                nl(); exit(-1);
            }
        }
    }

    if (verboseFlag)
    {
        nl();
        printf("Ground station    :%20s\n",fullSiteName);
        printf("Latitude          :%20.6f deg ",fabs(siteLat*CRD));

        (siteLat >= 0.0) ? printf("N\n") : printf("S\n");

        printf("Longitude         :%20.6f deg ",fabs(siteLong*CRD));

        (siteLong >= 0.0) ? printf("W\n") : printf("E\n");

        printf("Altitude          :%20.6f m\n",siteAlt/CMKM);
        nl();
    }

    else
    {
        if (!batchModeFlag && strcmp(fullSiteName,defaultSite))
            printf("%s\n\n",fullSiteName);
    }

    return;
}
 
/******************************************************************************/
/*                                                                            */
/* getTimeParams: reads additional parameters for orbit calculations          */
/*                                                                            */
/******************************************************************************/

void getTimeParams(timeArg)

double timeArg;

{
    double hour;
    long   sysDayNum;
    int    sysYear, sysMonth, sysDay, sysYearDay, sysHour, sysMin, sysSec;
    int    i, error;
    char   sysDate[20], sysTime[20], dummys[40], checkStr[10], monthStr[10];
 
    timeArg  += timeZone;
    sysDayNum = (long) timeArg;
    convertTime(timeArg,&dummyi,&sysHour,&sysMin,&sysSec);
    getDate(sysDayNum,&sysYear,&sysMonth,&sysDay,&sysYearDay);
    sprintf(sysDate,"%02d%3s%02d",sysDay,monthName(sysMonth),sysYear%100);
    sprintf(sysTime,"%02d:%02d:%02d",sysHour,sysMin,sysSec);

    if (batchModeFlag)
    {
        if (!strcmp(batchStartDate,NOW))
            strcpy(dispStr,YES);
    }

    else
    {
        printf("Start time  %7s %8s %-4s <Y> ? ",sysDate,sysTime,timeZoneStr);

        gets(dispStr);
        upperCase(dispStr);
    }

    if (strlen(dispStr) && strncmp(dispStr,YES,1))
    {
        error = TRUE;

        while(error)
        {
            if (batchModeFlag)
                strcpy(dispStr,batchStartDate);

            else
            {
                printf("Start date (%-4s)           <%7s> : ",
                    timeZoneStr,sysDate);
                gets(dispStr);

                if (!strlen(dispStr))
                    sprintf(dispStr,"%s",sysDate);
            }

            error = decodeDate(dispStr,&sysDay,&sysMonth,&sysYear);
        }

        error = TRUE;

        while(error)
        {
            if (batchModeFlag)
                strcpy(dispStr,batchStartTime);

            else
            {
                printf("Start time (%-4s)          <%8s> : ",
                    timeZoneStr,sysTime);
                gets(dispStr);

                if (!strlen(dispStr))
                    sprintf(dispStr,"%s",sysTime);
            }

            error = decodeTime(dispStr,&sysHour,&sysMin,&sysSec);
        }
    }

    defStepTime = 1.0 / epochMeanMotion * SPD / 125.0;
    defStepTime = (double) (((int) (defStepTime / 10.0)) * 5.0);

    if (!shortPredFlag)
    {
        if (sunTransitFlag)
            stepTime = 0.25;

        else
        {
            if (batchModeFlag)
            {
                if (!strcmp(batchStepTime,AUTO))
                    sprintf(dispStr,"");
                else
                    strcpy(dispStr,batchStepTime);
            }

            else
            {
                printf("Time step [s]                 <%5.0f> : ",defStepTime);
                gets(dispStr);
            }

            stepTime = (!strlen(dispStr)) ? defStepTime : atof(dispStr);
        }
    }

    if (batchModeFlag)
        strcpy(dispStr,batchDuration);

    else
    {
        printf("Duration [d]                   <%4.1f> : ",defDuration);
        gets(dispStr);
    }

    duration   = (!strlen(dispStr)) ? defDuration : atof(dispStr);
    startTime  = (double) getDayNum(sysDay,sysMonth,sysYear);          /* [d] */
    hour       = (double) (sysHour + sysMin/60.0 + sysSec/3600.0);     /* [h] */
    startTime += hour / 24.0 - timeZone;
    stepTime  /= SPD;                                                  /* [d] */
    stopTime   = startTime + duration;                                 /* [d] */

    if (!shortPredFlag)
    {
        if (sunTransitFlag)
            minElevation = 0.0;

        else
        {
            if (batchModeFlag)
                strcpy(dispStr,batchMinElev);

            else
            {
                printf("Minimum elevation [deg]        ");
                printf("<%4.1f> : ",defMinElevation);
                gets(dispStr);
            }

            minElevation = (!strlen(dispStr)) ? defMinElevation : atof(dispStr);
            minElevation *= CDR;
        }
    }

    return;
}
 
/******************************************************************************/
/*                                                                            */
/* printMode: prints phase-dependent mode of satellite into output file       */
/*                                                                            */
/******************************************************************************/

void printMode(outFile)

FILE *outFile;

{
    getMode();
    fprintf(outFile," %2s",modeString);
    return;
}
 
/******************************************************************************/
/*                                                                            */
/*  readSatList: reads names of satellites from the input file                */
/*                                                                            */
/******************************************************************************/

void readSatList()

{
    int  i, j, error, foundSat;
    char data[80], satListName[80], satListFileName[80], inputFileName[120];
    char str1[80], str2[80];
    FILE *inputFile;

    error = TRUE;

    while (error)
    {
        if (quickStartFlag)
            sprintf(satGroup,"");

        else
        {
            printf("Enter satellite group        <%6s> : ",defaultSatGroup);
            gets(satGroup);
        }

        if (!strlen(satGroup))
            sprintf(satGroup,"%s",defaultSatGroup);

        lowerCase(satGroup);
        sprintf(data,"%s/%s",strpHome,DATA);
        sprintf(satListFileName,"satlist_%s.dat",satGroup);
        sprintf(inputFileName,"%s/%s",data,satListFileName);

        if ((inputFile = fopen(inputFileName,"r")) == 0)
        {
            alarm(); nl(); reverseBlink();
            printf(" File '%s' not found \n",satListFileName);
            normal(); nl();
            error = TRUE;
            fclose(inputFile);

            if (quickStartFlag)
                exit(-1);
        }

        else
            error = FALSE;
    }

    if (!quickStartFlag)
        printf("\nreading satellite names for group '%s' ...\n",satGroup);

    upperCase(satGroup);

    i = 0;

    while (fgets(satListName,80,inputFile))
    {
        satListName[strlen(satListName) - 1] = '\0';
        truncBlanks(satListName);
        foundSat = FALSE;

        for (j = 0; j < i; j++)                 /* filter out duplicate names */
        {
            strcpy(str1,satel[j].satelName);
            strcpy(str2,satListName);
            upperCase(str1);
            upperCase(str2);

            if (!strcmp(str1,str2))
            {
                printf("duplicate entry: %s\n",str1);
                foundSat = TRUE;
            }
        }

        if (!foundSat)
        {
            cleanSatName(satListName);
            strcpy(satel[i].satelName,satListName);
            i++;
        }
    }

    fclose(inputFile);
    numSats  = i;
    foundSat = FALSE;

    for (i = 0; i < numSats; i++)             /* check if menu sat is in list */
    {
        if (!strncmp(satName,satel[i].satelName,strlen(satel[i].satelName)) || 
             satNum == atol(satel[i].satelName))
            foundSat = TRUE;
    }

    if (!foundSat)                            /* if not, add menu sat to list */
    {
        sprintf(satel[numSats].satelName,"%ld",satNum);
        numSats++;
    }

    return;
}

/******************************************************************************/
/*                                                                            */
/*  loadSatElements: load structure with all Keplerian elements and other     */
/*                   parameters                                               */
/*                                                                            */
/******************************************************************************/

void loadSatElements()

{
    int  i, j, n, error, grandError;
    long satNumSelect;
    char str[80], stsStr[80];

    grandError   = FALSE;
    satNumSelect = satNum;
    n = 0;

    if (!quickStartFlag)
        printf("loading element sets for %d satellites ...\n",numSats);

    for (i = 0; i < numSats; i++)
    {
        strcpy(str,satel[i].satelName);
        strcpy(stsStr,str);
        upperCase(stsStr);

        stsFlag = !strncmp(str,"STS",3) ? TRUE : FALSE;

        if (stsFlag && !strcmp(str,satName))
            sprintf(elementFile,"%s",defaultSetSTS);
        else
            sprintf(elementFile,"%s",defaultSetTLE);

        error = readTle(FALSE,str,&satNum,&elementSet,
                   &epochDay,&inclination,&epochRaan,&eccentricity,
                   &epochArgPerigee,&epochMeanAnomaly,&epochMeanMotion,
                   &decayRate,&decayRateDot,&bStarCoeff,
                   &epochOrbitNum,&ephemerisType);

        /* if object number was specified on string 'str', readTle() comes    */
        /* back with the satellite name on the same variable 'str'            */

        strcpy(stsStr,str);          /* need to restore stsFlag, since object */
        upperCase(stsStr);           /* number might have been specified      */
                              
        stsFlag = !strncmp(str,"STS",3) ? TRUE : FALSE;

        if (error)
        {
            alarm(); reverseBlink(); nl();

            if (error == 1 || error == 2)
                printf(" %s: check sum error in line %d ",str,error);

            if (error == 3)
                printf(" %s: no matching satellite name ",str);

            if (error == 4)
                printf(" %s: multiple entries ",str);

            nl(); normal(); nl();

            grandError++;
        }

        else
        {
            getSatModes(satel[i].satelName);
            changeElementUnits();
            getSatPrecession();
            checkPropModelType();

            strcpy(sat[n].satellite,str);
            strcpy(sat[n].satelliteAlias,satAlias);
            strcpy(sat[n].model,propModel);

            sat[n].satIdNum     = satNum;
            sat[n].elSet        = elementSet;

            sat[n].epDay        = epochDay;
            sat[n].elSetEp      = elsetEpoch;
            sat[n].launchEp     = launchEpoch;

            sat[n].epInc        = inclination;
            sat[n].epRaan       = epochRaan;
            sat[n].epEcc        = eccentricity;
            sat[n].epArgPer     = epochArgPerigee;
            sat[n].epMeanAnom   = epochMeanAnomaly;
            sat[n].epMeanMot    = epochMeanMotion;
            sat[n].epDecRate    = decayRate;
            sat[n].epDecRateDot = decayRateDot;
            sat[n].epBstar      = bStarCoeff;
            sat[n].epOrbNum     = epochOrbitNum;
            sat[n].ephType      = ephemerisType;
            sat[n].modelType    = propModelType;

            sat[n].freq         = beaconFreq;
            sat[n].perigPh      = perigeePhase;
            sat[n].maxPh        = maxPhase;
            sat[n].numMd        = numModes;
            sat[n].attitudeFl   = attitudeFlag;
            sat[n].attLtd       = attLat;
            sat[n].attLng       = attLong;
            sat[n].attX         = satAttGl[0];
            sat[n].attY         = satAttGl[1];
            sat[n].attZ         = satAttGl[2];
            
            sat[n].semiMajAxis  = semiMajorAxis;
            sat[n].raanPr       = raanPrec;
            sat[n].periPr       = perigeePrec;
            sat[n].refOrb       = refOrbit;

            sat[n].cosInc       = cosInclination;
            sat[n].sinInc       = sinInclination;

            sat[n].stsFl        = stsFlag;
            sat[n].launchFl     = launchFlag;
            sat[n].prelaunchFl  = FALSE;
            sat[n].preorbitFl   = FALSE;
            sat[n].geosyncFl    = geoSyncFlag;
            sat[n].crashFl      = satCrashFlag;
            sat[n].invalidFl    = FALSE;

            sat[n].statusFl     = (satNumSelect == satNum) ? 
                                   SELECT : FALSE;

            if (numModes)
            {
                for (j = 0; j < numModes; j++)
                {
                    strcpy(sat[n].satModes[j].modeStr,modes[j].modeStr);
                    sat[n].satModes[j].minPhase = modes[j].minPhase; 
                    sat[n].satModes[j].maxPhase = modes[j].maxPhase;
                }
            }

            n++;
        }
    }

    numSats = n;

    if (grandError)
    {
        printf("\nhit RETURN to continue ");
        gets(str);
    }

    return;
}

/******************************************************************************/
/*                                                                            */
/* End of function block satread.c                                            */
/*                                                                            */
/******************************************************************************/
