/******************************************************************************/
/*                                                                            */
/*  Title       : sattele.c                                                   */
/*  Author      : Manfred Bester, DL5KR                                       */
/*  Date        : 14Feb94                                                     */
/*  Last change : 05Oct94                                                     */
/*                                                                            */
/*  Synopsis    : Interface routines for controlling the ISI telescopes.      */
/*                The main body of these routines will only be executed if    */
/*                'ISI' is defined in 'sattrack.h'. This also requires a      */
/*                corresponding change in the Makefile to allow the remote    */
/*                procedure call interface routines to be linked.             */
/*                                                                            */
/*                                                                            */
/*  Copyright (c) 1992, 1993, 1994 Manfred Bester. All Rights Reserved.       */
/*                                                                            */
/*  Permission to use, copy, modify, and distribute this software and its     */
/*  documentation for educational, research and non-profit purposes, without  */
/*  fee, and without a written agreement is hereby granted, provided that the */
/*  above copyright notice and the following three paragraphs appear in all   */
/*  copies.                                                                   */
/*                                                                            */
/*  Permission to incorporate this software into commercial products may be   */
/*  obtained from the author, Dr. Manfred Bester, 1636 M. L. King Jr. Way,    */
/*  Berkeley, CA 94709, USA.                                                  */
/*                                                                            */
/*  IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, */
/*  SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF   */
/*  THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE AUTHOR HAS BEEN ADVISED  */
/*  OF THE POSSIBILITY OF SUCH DAMAGE.                                        */
/*                                                                            */
/*  THE AUTHOR SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT      */
/*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A   */
/*  PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS"      */
/*  BASIS, AND THE AUTHOR HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT, */
/*  UPDATES, ENHANCEMENTS, OR MODIFICATIONS.                                  */
/*                                                                            */
/*                                                                            */
/*  If 'ISI' is undefined in 'sattrack.h', the parts of the following         */
/*  functions that require certain objects specific for the ISI telescopes    */
/*  are commented out. This way the code can be compiled elsewhere.           */
/*                                                                            */
/******************************************************************************/

#include <stdio.h>
#include <math.h>
#include <rpc/rpc.h>

#include "satglobalsx.h"
#include "sattrack.h"
#include "vt100.h"

#ifdef ISI
#include "/usr1/isisoft/src/VxSource/include/globalserver.h"
#endif

#ifndef SUNOS4
#include <stdlib.h>
#endif

/******************************************************************************/
/*                                                                            */
/*  global fields                                                             */
/*                                                                            */
/******************************************************************************/

char *cpu[] = { "vme0", "vme1", "vme2" };

/******************************************************************************/
/*                                                                            */
/*  external functions                                                        */
/*                                                                            */
/******************************************************************************/

#ifdef ISI
extern int remotevar(), trackControl();
#endif

/******************************************************************************/
/*                                                                            */
/*  initISI: initializes ISI telescope tracking control parameters            */
/*                                                                            */
/******************************************************************************/

int initISI()

{
    long clearFlag, solsysFlag, tracking, delayLine;
    int  i, error, trackFlag;
    char objName[80];

    error = FALSE;

#ifdef ISI

    if (trackObject == SUN)                         /* avoid tracking the Sun */
        return(-1);

    if (trackObject == MOON)
        strcpy(objName,"Moon");

    if (trackObject == SAT)
        strcpy(objName,satTrackName);

    clearFlag  = 1L;
    solsysFlag = 998L;
    tracking   = 1L;
    trackFlag  = 1;
    delayLine  = 1L;

    for (i = 0; i <= 2; i++)                      /* talk vme0, vme1 and vme2 */
    {
        error += remotevar(cpu[i], "srcname",   STRING | WRITE, objName);
        error += remotevar(cpu[i], "clearflag", LONG   | WRITE, &clearFlag);
        error += remotevar(cpu[i], "solsysfl",  LONG   | WRITE, &solsysFlag);
        error += remotevar(cpu[i], "tracking",  LONG   | WRITE, &tracking);

        error += trackControl(cpu[i],trackFlag);
    }

    error += remotevar(cpu[0],"dlyblockflag", LONG | WRITE, &delayLine);

#endif

    return(error);
}

/******************************************************************************/
/*                                                                            */
/*  stopISI: stops tracking with the ISI telescopes                           */
/*                                                                            */
/******************************************************************************/

int stopISI()

{
    long tracking;
    int  i, error, trackFlag;
    char objName[80];

    error = FALSE;

#ifdef ISI

    sprintf(objName,"%s","");

    tracking  = 0L;
    trackFlag = FALSE;

    for (i = 0; i <= 2; i++)                   /* talk to vme0, vme1 and vme2 */
    {
        error += remotevar(cpu[i], "srcname",   STRING | WRITE, objName);
        error += remotevar(cpu[i], "tracking",  LONG   | WRITE, &tracking);

        error += trackControl(cpu[i],trackFlag);
    }

#endif

    return(error);
}

/******************************************************************************/
/*                                                                            */
/*  controlISI: sends real-time tracking parameters to the ISI telescopes via */
/*              remote procedure calls                                        */
/*                                                                            */
/*              local geodetic coordinate system for the ISI:                 */
/*              left-handed                                                   */
/*              +X = north,  +Y = east,  +Z = up                              */
/*              therefore the x coordinate needs to be inverted               */
/*                                                                            */
/******************************************************************************/

int controlISI()

{
    double localVecX, localVecY, localVecZ, azimuth, elevation;
    int    i, index, error;
    char   varName[40];

    error = FALSE;

#ifdef ISI

    if (trackObject == SUN)                         /* avoid tracking the Sun */
        return(-1);

    if (trackObject == MOON)
    {
        localVecX = -localVecMoonGl[0];
        localVecY =  localVecMoonGl[1];
        localVecZ =  localVecMoonGl[2];

        azimuth   =  moonAzimuth;
        elevation =  moonElevation;
    }

    if (trackObject == SAT)
    {
        localVecX = -localVecSatGl[0];
        localVecY =  localVecSatGl[1];
        localVecZ =  localVecSatGl[2];

        azimuth   =  satAzimuth;
        elevation =  satElevation;
    }

    index = (int) (curYearSec % 20L);           /* ring buffer has 20 entries */

    for (i = 0; i <= 2; i++)                   /* talk to vme0, vme1 and vme2 */
    {
        sprintf(varName,"timeindexSat[%d]",index);
        error += remotevar(cpu[i], varName, LONG   | WRITE, &curYearSec);

        if (i == 0)                                     /* vme0 only          */
        {
            sprintf(varName,"azimuthSat[%d]",index);
            error += remotevar(cpu[i], varName, DOUBLE | WRITE, &azimuth);

            sprintf(varName,"altitudeSat[%d]",index);
            error += remotevar(cpu[i], varName, DOUBLE | WRITE, &elevation);
        }

        if (i > 0)                                      /* vme1 and vme2 only */
        {
            sprintf(varName,"scrxSat[%d]",index);
            error += remotevar(cpu[i], varName, DOUBLE | WRITE, &localVecX);

            sprintf(varName,"scrySat[%d]",index);
            error += remotevar(cpu[i], varName, DOUBLE | WRITE, &localVecY);

            sprintf(varName,"scrzSat[%d]",index);
            error += remotevar(cpu[i], varName, DOUBLE | WRITE, &localVecZ);
        }
    }

#endif

    return(error);
}

/******************************************************************************/
/*                                                                            */
/*  end of function block sattele.c                                           */
/*                                                                            */
/******************************************************************************/
