/******************************************************************************/
/*                                                                            */
/*  Title       : sattrack.c                                                  */
/*  Author      : Manfred Bester, DL5KR                                       */
/*  Date        : 24Feb92                                                     */
/*  Last change : 05Oct94                                                     */
/*                                                                            */
/*  Synopsis    : Satellite orbit prediction and live tracking program.       */
/*                                                                            */
/*  Resources   : cities.dat       modes.dat           tlex.dat               */
/*                defaults.dat     satlist_nnn.dat                            */
/*                                                                            */
/*                                                                            */
/*  Copyright (c) 1992, 1993, 1994 Manfred Bester. All Rights Reserved.       */
/*                                                                            */
/*  Permission to use, copy, modify, and distribute this software and its     */
/*  documentation for educational, research and non-profit purposes, without  */
/*  fee, and without a written agreement is hereby granted, provided that the */
/*  above copyright notice and the following three paragraphs appear in all   */
/*  copies.                                                                   */
/*                                                                            */
/*  Permission to incorporate this software into commercial products may be   */
/*  obtained from the author, Dr. Manfred Bester, 1636 M. L. King Jr. Way,    */
/*  Berkeley, CA 94709, USA.                                                  */
/*                                                                            */
/*  IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, */
/*  SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF   */
/*  THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE AUTHOR HAS BEEN ADVISED  */
/*  OF THE POSSIBILITY OF SUCH DAMAGE.                                        */
/*                                                                            */
/*  THE AUTHOR SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT      */
/*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A   */
/*  PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS"      */
/*  BASIS, AND THE AUTHOR HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT, */
/*  UPDATES, ENHANCEMENTS, OR MODIFICATIONS.                                  */
/*                                                                            */
/******************************************************************************/

#include <stdio.h>
#include <math.h>
#include <sys/types.h>
#include <sys/termios.h>
#include <sys/ioctl.h>

#include "satglobalsx.h"
#include "satglobals.h"
#include "satstrings.h"
#include "sattrack.h"
#include "vt100.h"

#ifndef SUNOS4
#include <stdlib.h>
#endif

/******************************************************************************/
/*                                                                            */
/*  list of flow control flags used in SatTrack:                              */
/*                                                                            */
/*  attitudeFlag       TRUE  if attitude is specified in modes.dat            */
/*  batchModeFlag      TRUE  if main() is invoked with '-b' argument          */
/*  checkCountdown     TRUE  after initRestart; used in singlesat live disp   */
/*  countdownFlag      TRUE  if time of AOS/LOS - current time < 1 minute     */
/*  debugFlag          TRUE  if main() is invoked with '-d' argument          */
/*  deepSpaceFlag      TRUE  if satellite is a deep space object (SDP4 model) */
/*  didMultDispFlag    TRUE  if multisat live display was selected once       */
/*  dispUpdateFlag     TRUE  if live display update interval has elapsed      */
/*  duplFlag           TRUE  if duplicate entry in TLE sets is discovered     */
/*  elevationFlag      TRUE  if sat elevation is > 0 (in live display)        */
/*  finishPass         TRUE  if sat elevation is > 0 (in prediction)          */
/*  firstMultiSat      TRUE  until first multi sat display cycle finished     */
/*  firstMultiSatCalc  TRUE  until 'calculating next pass' message displayed  */
/*  firstPage          TRUE  until prediction produces a form feed            */
/*  foundSatFlag       TRUE  if satellite found in multisat cursor pos mode   */
/*  foundRiseFlag      TRUE  if rise time was found (in prediction)           */
/*  foundSetFlag       TRUE  if set time was found (in prediction)            */
/*  geoSyncFlag        GEOSYNC if satellite is in geosynchronous orbit        */
/*                     GEOSTAT if satellite is in geostationary orbit         */
/*  gndTrkFlag         TRUE  if nearest city is displayed in multisat display */
/*  justAcquFlag       TRUE  if satellite just came above the horizon         */
/*  justLostFlag       TRUE  if satellite just went below the horizon         */
/*  launchFlag         TRUE  if launch date/time specified in modes.dat       */
/*  liveDispFlag       SINGLESAT or MULTISAT when in live display loop        */
/*  modelFlag          TRUE  if main() is invoked with '-m' argument          */
/*  newPassFlag        TRUE  if new pass was calculated but not saved yet     */
/*  newRiseFlag        TRUE  if new pass was calculated and not yet displayed */
/*  newTrackObjectFlag TRUE  if keyboard 0,1,2 and liveDispFlag = SINGLESAT   */
/*  noPassFlag         TRUE  if no pass found by getNextPass()                */
/*  objectNumFlag      TRUE  while display with object number is selected     */
/*  passDispFlag       TRUE  when new pass needs to be displayed              */
/*  predictionFlag     TRUE  if 'prediction' is chosen in main menu           */
/*  preLaunchFlag      TRUE  if launch date specified and not launched yet    */
/*  preOrbitFlag       TRUE  if current time is earlier than TLE epoch        */
/*  quickStartFlag     TRUE  if main() is invoked with '-q' argument          */
/*  reinitMultiSat     TRUE  when switched to multisat display                */
/*  reinitSingleSat    TRUE  when switched to singlesat display               */
/*  reverseFlag        TRUE  if display is set into reverse video mode        */
/*  sameOrbitPass      TRUE  if pass was found (needed for maybe second pass) */
/*  satCrashFlag       TRUE  if satellite has crashed (radius < Earth radius) */
/*  shortPredFlag      TRUE  if 'short prediction' is performed               */
/*  startFlag          TRUE  after initMain() until first initTerminal()      */
/*  stsFlag            TRUE  if satellite name begins with 'sts' or 'STS'     */
/*  sunProximity       TRUE  if satellite is close to the Sun                 */
/*  sunTransit         TRUE  if satellite is in front of the Sun              */
/*  sunTransitFlag     TRUE  if solar transits are to be predicted            */
/*  switchSatFlag      TRUE  when switching from single to multisat display   */
/*  timeZoneFlag       TRUE  if main() is invoked with '-t' argument          */
/*  tleInvalidFlag     TRUE  if TLE set is out of date                        */
/*  trackCtrlFlag      TRUE  after first tracking control loop                */
/*  trackDispFlag      TRUE  if trackingFlag was altered until display update */
/*  trackingFlag       TRUE  if tracking is enabled                           */
/*  transitFlag        TRUE  if batch mode prediction type is 'transit'       */
/*  verboseFlag        TRUE  if main() is invoked with '-v' argument          */
/*  writeFlag          TRUE  if predictions are written into a file           */
/*  xTermFlag          TRUE  if $term matches X terminal type in defaults.dat */
/*                                                                            */
/******************************************************************************/

/******************************************************************************/
/*                                                                            */
/*  global variables used with the main program                               */
/*                                                                            */
/******************************************************************************/

double riseAzimuth, setAzimuth;        /* [rad]                               */
double maxAzimuth;                     /* [rad]                               */
double maxElevation, relMaxElev;       /* [rad]                               */
double maxRange;                       /* range at MEL [km] or [nm]           */

double curTime, tmpTime, dT, dT2;      /* time [d]                            */
double predTime, predStopTime;         /* time [d]                            */
double fastStepTime, slowStepTime;     /* time [d]                            */
double medStepTime;
double nextRiseTime, nextSetTime;      /* rise and set times [d]              */
double nextMaxTime;
double maxDays;

double sunProxLimit;
double dAziSatSun, dEleSatSun;
double cUnit;

long   riseOrbitNum, printOrbitNum, lastYearSec;
long   trackInt, singleSatUpdateInt, multiSatUpdateInt;

int    sysYear, sysMonth, sysDay, sysYearDay, sysHour, sysMin, sysSec;
int    ns, sameOrbitPass, foundSatFlag, foundRiseFlag, foundSetFlag;
int    startFlag, predictionFlag, writeFlag, firstMultiSat, firstMultiSatCalc;
int    liveDispFlag, finishPass, newPassFlag, dispUpdateFlag, switchSatFlag;
int    reinitSingleSat, reinitMultiSat, key, keyBoard;

char   fileName[100], syscomm[120], buf[80], sUnit[10], fond[30];
char   str[80], satNameStr[80];

/******************************************************************************/
/*                                                                            */
/* makeOutputFile: makes output file for predictions                          */
/*                                                                            */
/******************************************************************************/

void makeOutputFile()

{
    if (predictionFlag)
        outFile = stdout;
 
    if (writeFlag)
    {
        sprintf(fileName,"%s/%s/%s",strpHome,PREDICTION,satFileName);

        if ((outFile = fopen(fileName,"w")) == NULL)
        {
            printf("\ncannot open output file %s\n\n",fileName);
            exit(-1);
        }

        if (!batchModeFlag)
            printf("\ndata are written into '%s'\n\n",fileName);
    }

    return;
}

/******************************************************************************/
/*                                                                            */
/* printPredHeader: prints prediction header                                  */
/*                                                                            */
/******************************************************************************/

void printPredHeader()

{
    checkPropModelType();

    fprintf(outFile,"\n%s %s %s (%s)\n\n",
        callSign,sattrName,predHeader,sattrVersion);
    fprintf(outFile,"Satellite #%-6ld : %s",satNum,satName);

    if (strlen(satAlias))
        fprintf(outFile," (%s)",fullSatAlias);

    fprintf(outFile,"\n");
    fprintf(outFile,"Data File         : %s\n",elementFile);

    if (stsFlag)
        fprintf(outFile,"Element Set Type  : %s\n",elementType);

    fprintf(outFile,"Element Set Number: %d ",elementSet);
    fprintf(outFile,"(Orbit %ld)\n",epochOrbitNum);
    fprintf(outFile,"Element Set Epoch : %s   ",epochString);
    fprintf(outFile,"(%s)\n",updateString);

    if (launchFlag)
        fprintf(outFile,"Launch Date/Time  : %s\n",launchString);

    fprintf(outFile,"Propagation Model : %s\n",propModel);
    fprintf(outFile,"Ground Station    : %s\n",fullSiteName);
    fprintf(outFile,"Time Zone         : %s (%+.1f h)\n",
            timeZoneStr,timeZone*24.0);

    if (!shortPredFlag && !sunTransitFlag)
        fprintf(outFile,"Doppler Reference : %.4f MHz\n",beaconFreq*CHZMHZ);

    if (!shortPredFlag && sunTransitFlag)
        fprintf(outFile,"\nPASSES NEAR OR ACROSS THE SOLAR DISK:\n\n");

    return;
}

/******************************************************************************/
/*                                                                            */
/* doLongPrediction: performs long form orbit prediction                      */
/*                                                                            */
/******************************************************************************/

void doLongPrediction()

{
    slowStepTime  = 1.0 / epochMeanMotion / 2000.0;
    medStepTime   = slowStepTime *  5.0;
    fastStepTime  = medStepTime * 20.0;
    firstLine     = TRUE;
    firstPage     = TRUE;
    sameOrbitPass = FALSE;

    if (curTime < realTime)
        predTime = curTime;
    else
        predTime = (elsetEpoch > curTime) ? elsetEpoch : curTime;

    predStopTime = predTime + (stopTime - startTime);

    do
    {
        getSiderealTime(predTime);
        getSiteVector();

        getNoradStateVector(predTime);
        getStateVector(predTime);

        getAziElev(SAT);

        if (satElevation >= minElevation && predTime >= startTime)
        {
            getRange();
            getShuttleOrbit();
            getSunVector(0);
            getAziElev(SUN);
            getSubSatPoint();
            getDoppler();
            getPathLoss();
            getSquintAngle();
            getPhase();
            getMode();

            if (sunTransitFlag)
            {
                dAziSatSun = sunAzimuth - satAzimuth;
                dEleSatSun = sunElevation - satElevation;

                sunProximity = (fabs(dAziSatSun) < sunProxLimit &&
                                fabs(dEleSatSun) < sunProxLimit) ? TRUE : FALSE;

                sunTransit   = (fabs(dAziSatSun) < SUNDISKRAD &&
                                fabs(dEleSatSun) < SUNDISKRAD) ? TRUE : FALSE;
            }

            if (!sameOrbitPass || firstLine)
            {
                if (stsFlag)
                {
                    cUnit = CKMNM;
                    sprintf(sUnit,"nm");
                    printOrbitNum = stsOrbitNum;
                }

                else
                {
                    cUnit = 1.0;
                    sprintf(sUnit,"km");
                    printOrbitNum = orbitNum;
                }

                if (printOrbitNum > 0 && 
                    (!sunTransitFlag || (sunTransitFlag && sunProximity)))
                {
                    if (!firstPage)
                        fprintf(outFile,"\f");

                    if (firstPage && !writeFlag || writeFlag)
                        printPredHeader();

                    firstPage = FALSE;

                    fprintf(outFile,"\n\n");
                    printDate(outFile,predTime+timeZone);
                    fprintf(outFile,"  ---  Orbit %ld\n\n",printOrbitNum);

                    fprintf(outFile,"  %4s",timeZoneStr);

                    if (launchFlag && stsFlag)
                        fprintf(outFile,"          MET ");

                    if (sunTransitFlag)
                    {
                        fprintf(outFile,"    Sat Azi  Sat Ele");
                        fprintf(outFile,"    Sun-Sat  Azi Ele");
                    }

                    else
                    {
                        fprintf(outFile,"    Azimuth  Elev");
                    }

                    fprintf(outFile,"    Range");

                    if (!sunTransitFlag)
                        fprintf(outFile,"  Doppler  Loss  SqAng  Phs Md V\n");
                    else
                        fprintf(outFile,"\n");

                    fprintf(outFile,"      ");

                    if (launchFlag && stsFlag)
                        fprintf(outFile,"              ");

                    if (sunTransitFlag)
                    {
                        fprintf(outFile,"     [deg]    [deg]");
                        fprintf(outFile,"         [arcmin]    ");
                    }

                    else
                    {
                        fprintf(outFile,"     [deg]  [deg]");
                    }

                    fprintf(outFile,"     [%2s]",sUnit);

                    if (!sunTransitFlag)
                        fprintf(outFile,"   [kHz]   [dB]  [deg]");

                    fprintf(outFile,"\n\n");
                    firstLine = FALSE;
                }

                if (writeFlag && !batchModeFlag && !sunTransitFlag)
                {
                    printf(".");
                    fflush(stdout);
                }
            }

            eclipseCode = satEclipse(satElevation);

            if (printOrbitNum > 0 && 
                (!sunTransitFlag || (sunTransitFlag && sunProximity)))
            {
                printTime(outFile,predTime+timeZone);

                if (launchFlag && stsFlag)
                {
                    fprintf(outFile,"  ");
                    printMET(outFile,predTime-launchEpoch);
                }

                if (sunTransitFlag)
                {
                    fprintf(outFile,"  %7.3lf",satAzimuth*CRD);
                    fprintf(outFile,"  %7.3lf",satElevation*CRD);
                    fprintf(outFile,"   %8.3lf",dAziSatSun*CRAM);
                    fprintf(outFile," %8.3lf",dEleSatSun*CRAM);
                }

                else
                {
                    fprintf(outFile,"   %5.1lf",satAzimuth*CRD);
                    fprintf(outFile,"  %5.1lf",satElevation*CRD);
                }

                fprintf(outFile," %8.1lf",satRange*cUnit);

                if (!sunTransitFlag)
                {
                    fprintf(outFile,"  %+6.2lf",dopplerShift*CHZKHZ);
                    fprintf(outFile,"  %5.1lf",pathLoss);
                    fprintf(outFile," %6.1lf",squintAngle*CRD);
                    fprintf(outFile,"  %3.0f",satPhase);
                    printMode(outFile);
                    fprintf(outFile," %s",visibCode[eclipseCode]);
                }

                else
                    fprintf(outFile,"  ");

                if (sunTransit)
                    fprintf(outFile,"  ###");

                fprintf(outFile,"\n");
                sameOrbitPass = TRUE;
            }
        }

        else
            sameOrbitPass = FALSE;

        predTime  += (satElevation < VLOWELEV) ? fastStepTime : stepTime;
        finishPass = (satElevation >= minElevation) ? TRUE : FALSE;

        if (tleInvalidFlag)
            fprintf(outFile,"\nTLE data set is out of date\n\n");
    }
    while ((predTime <= predStopTime || finishPass) && !tleInvalidFlag);

    return;
}

/******************************************************************************/
/*                                                                            */
/* doShortPrediction: performs short form orbit prediction                    */
/*                                                                            */
/******************************************************************************/

void doShortPrediction()

{
    printPredHeader();

    if (curTime < realTime)
        predTime = curTime;
    else
        predTime = (elsetEpoch > curTime) ? elsetEpoch : curTime;

    predStopTime = predTime + (stopTime - startTime);
    firstLine    = TRUE;
    lastDayNum   = 0;

    fprintf(outFile,"\n\n");
    fprintf(outFile," Date (%s)        ",timeZoneStr);
    if (strlen(timeZoneStr) == 3) fprintf(outFile," ");
    fprintf(outFile," Time (%s) of       ",timeZoneStr);
    if (strlen(timeZoneStr) == 3) fprintf(outFile," ");
    fprintf(outFile,"Duration   Azimuth at   Max  Vis Orbit");

    if (launchFlag && stsFlag && writeFlag)
        fprintf(outFile,"   MET at AOS");

    fprintf(outFile,"\n");
    fprintf(outFile,"                 AOS      MEL      LOS     ");
    fprintf(outFile,"of Pass  AOS MEL LOS  Elev\n");

    do
    {
        getNextPass();

        if (riseOrbitNum > 0)
        {
            printDate(outFile,nextRiseTime+timeZone);
            fprintf(outFile,"  ");
            printTime(outFile,nextRiseTime+timeZone);
            fprintf(outFile," ");
            printTime(outFile,nextMaxTime+timeZone);
            fprintf(outFile," ");
            printTime(outFile,nextSetTime+timeZone);
            fprintf(outFile,"  ");
            printTime(outFile,nextSetTime-nextRiseTime);
            fprintf(outFile,"  %3.0f %3.0f %3.0f",
                    riseAzimuth*CRD,maxAzimuth*CRD,setAzimuth*CRD);
            fprintf(outFile,"  %4.1f",maxElevation*CRD);
            if (maxElevation > MAXELEV)
                fprintf(outFile,"*");
            else
                fprintf(outFile," ");
            fprintf(outFile," %s%s%s",visibCode[eclipseRise],
                    visibCode[eclipseMax],visibCode[eclipseSet]);
            fprintf(outFile," %5ld",riseOrbitNum);

            if (launchFlag && stsFlag && writeFlag)
            {
                fprintf(outFile," ");
                printMET(outFile,nextRiseTime-launchEpoch);
            }

            fprintf(outFile,"\n");
            firstLine = FALSE;
        }

        predTime = tmpTime + fastStepTime;

        if (tleInvalidFlag)
            fprintf(outFile,"\nTLE data set is out of date\n\n");
    }
    while (predTime <= predStopTime && !tleInvalidFlag);

    return;
}

/******************************************************************************/
/*                                                                            */
/* doPrediction: run orbit prediction                                         */
/*                                                                            */
/******************************************************************************/

void doPrediction()

{
    makeOutputFile();

    if (predictionFlag || writeFlag)
    {
        getTimeParams(curTime);
        curTime = startTime;

        if (!batchModeFlag)
        {
            printf("\ncalculating passes of %s over %s ...\n",
                    satName,fullSiteName);
        }

        if (shortPredFlag)
            doShortPrediction();
        else
            doLongPrediction();

        if (writeFlag)
        {
            fclose(outFile);

            if (!satCrashFlag && !tleInvalidFlag)
            {
                if (batchModeFlag)
                {
                    upperCase(batchHardcopy);

                    if (!strcmp(batchHardcopy,HARDCOPY))
                        strcpy(dispStr,YES);
                    else
                        strcpy(dispStr,NO);
                }

                else
                {
                    printf("\nHardcopy (Y,N,Q)                  <Y> ? ");
                    gets(dispStr);
                    upperCase(dispStr);

                    if (!strncmp(dispStr,QX,1))
                        exit(-1);
                }

                if (!strlen(dispStr) || !strncmp(dispStr,YES,1))
                {
                    if (!batchModeFlag)
                        printf("\nprinting %s ...\n",fileName);

                    if (shortPredFlag && !launchFlag && !transitFlag)
                        sprintf(fond,"%s%s",FOND,SIZE10);

                    if (shortPredFlag && launchFlag && !transitFlag)
                        sprintf(fond,"%s%s",FOND,SIZE9);

                    if (!shortPredFlag && launchFlag && !transitFlag)
                        sprintf(fond,"%s%s",FOND,SIZE10);

                    if (!shortPredFlag && !launchFlag && !transitFlag)
                        sprintf(fond,"%s%s",FOND,SIZE10);

                    if (transitFlag)
                        sprintf(fond,"%s%s",FOND,SIZE9);

                    sprintf(syscomm,"%s %s %s %s",
                        PRINTCMD,PRINTOPT,fond,fileName);

                    system(syscomm);
                }
            }
        }

        if (!batchModeFlag)
            nl();
    }

    if (batchModeFlag)
        exit(-1);

    return;
}

/******************************************************************************/
/*                                                                            */
/* getNextPass: finds circumstances of the next pass                          */
/*                                                                            */
/******************************************************************************/

void getNextPass()

{
    tmpTime      = predTime;
    slowStepTime = 1.0 / epochMeanMotion / 2000.0;
    medStepTime  = 5.0 * slowStepTime;
    fastStepTime = 20.0 * medStepTime;
    riseOrbitNum = -1;

    maxDays = (geoSyncFlag == GEOSYNC) ? MAXDAYSSYNC : MAXDAYS;

    do
    {
        maxElevation  = -HALFPI;
        relMaxElev    = -HALFPI;
        riseAzimuth   = ZERO;
        foundRiseFlag = FALSE;
        foundSetFlag  = FALSE;
        noPassFlag    = FALSE;

        do
        {
            getSiderealTime(tmpTime);
            getSiteVector();

            getNoradStateVector(tmpTime);
            getStateVector(tmpTime);

            getAziElev(SAT);

            if (satElevation > ZERO && !foundRiseFlag)
            {
                getShuttleOrbit();
                getSunVector(0);
                getAziElev(SUN);

                riseOrbitNum  = (stsFlag) ? stsOrbitNum : orbitNum;
                nextRiseTime  = tmpTime;
                riseAzimuth   = satAzimuth;
                eclipseRise   = satEclipse(CASR);
                foundRiseFlag = TRUE;
            }

            if (foundRiseFlag && satElevation > maxElevation)
            {
                getRange();
                getSunVector(0);
                getAziElev(SUN);

                maxAzimuth    = satAzimuth;
                maxElevation  = satElevation;
                maxRange      = satRange;
                nextMaxTime   = tmpTime;
                eclipseMax    = satEclipse(satElevation);
            }

            if (satElevation < ZERO && foundRiseFlag)
            {
                getSunVector(0);
                getAziElev(SUN);

                nextSetTime   = tmpTime;
                setAzimuth    = satAzimuth;
                eclipseSet    = satEclipse(CASR);
                foundSetFlag  = TRUE;
            }

            if (satElevation < VLOWELEV || satElevation > VLOWELEV && 
                satElevation < relMaxElev && !foundRiseFlag)
                tmpTime += fastStepTime;

            if (satElevation > VLOWELEV && satElevation > relMaxElev && 
                satElevation < LOWELEV)
                tmpTime += medStepTime;

            if (satElevation > LOWELEV && satElevation > relMaxElev && 
                !foundRiseFlag)
                tmpTime += slowStepTime;

            if (foundRiseFlag && fabs(satElevation - maxElevation) < CASR || 
                satElevation > -LOWELEV)
                tmpTime += slowStepTime;

            if (satElevation > ZERO && satElevation < maxElevation && 
                satElevation < -LOWELEV)
                tmpTime += slowStepTime;

            if (tmpTime > predTime + maxDays)           /* stop after maxDays */
            {
                if (foundRiseFlag && !foundSetFlag)
                {
                    getSunVector(0);
                    getAziElev(SUN);

                    nextSetTime  = tmpTime;
                    setAzimuth   = satAzimuth;
                    eclipseSet   = satEclipse(CASR);
                    foundSetFlag = TRUE;
                }

                if (!foundRiseFlag && !foundSetFlag)
                {
                    nextRiseTime  = tmpTime;
                    nextSetTime   = tmpTime;
                    riseAzimuth   = ZERO;
                    maxAzimuth    = ZERO;
                    setAzimuth    = ZERO;
                    maxElevation  = ZERO;
                    maxRange      = ZERO;
                    eclipseRise   = BLANK;
                    eclipseMax    = BLANK;
                    eclipseSet    = BLANK;
                    foundRiseFlag = TRUE;
                    foundSetFlag  = TRUE;
                    noPassFlag    = TRUE;
                }
            }

            relMaxElev = satElevation;
        }
        while (!foundSetFlag);

        if (writeFlag && !batchModeFlag)
        {
            printf(".");
            fflush(stdout);
        }
    }
    while (riseOrbitNum <= 0 && !noPassFlag);

    newRiseFlag = TRUE;                            /* for controlling display */
    return;
}

/******************************************************************************/
/*                                                                            */
/* dispVersion: displays program version                                      */
/*                                                                            */
/******************************************************************************/

void dispVersion()

{
    if (!batchModeFlag && !quickStartFlag)
    {
        system("clear");
        printf("%s %s\n",sattrName,sattrVersion);
    }

    return;
}

/******************************************************************************/
/*                                                                            */
/* getEnvironment: gets the environment variables                             */
/*                                                                            */
/******************************************************************************/

void getEnvironment()

{
    strpHome   = getenv("HOME");
    strpTerm   = getenv("TERM");

    return;
}

/******************************************************************************/
/*                                                                            */
/* checkArguments: checks if command line arguments are passed to main        */
/*                                                                            */
/******************************************************************************/

void checkArguments(argC,argV)

int  argC;
char *argV[];

{
    int i, error;

    error          = FALSE;
    batchModeFlag  = FALSE;
    debugFlag      = FALSE;
    modelFlag      = FALSE;
    quickStartFlag = FALSE;
    timeZoneFlag   = FALSE;
    verboseFlag    = FALSE;

    if (argC >= 2)
    {
        for (i = 1; i < argC; i++)
        {
            upperCase(argV[i]);

            if (!strncmp(argV[i],BATCH,2))
                batchModeFlag = TRUE;

            if (!strncmp(argV[i],DEBUG,2))
                debugFlag = TRUE;

            if (!strncmp(argV[i],MODEL,2))
                modelFlag = TRUE;

            if (!strncmp(argV[i],QUICK,2))
                quickStartFlag = TRUE;

            if (!strncmp(argV[i],TIMEZONE,2))
                timeZoneFlag = TRUE;

            if (!strncmp(argV[i],VERBOSE,2))
                verboseFlag = TRUE;
        }
    } 

    if (batchModeFlag)
    {
        debugFlag      = FALSE;
        modelFlag      = FALSE;
        quickStartFlag = FALSE;
        timeZoneFlag   = FALSE;
        verboseFlag    = FALSE;

        if (argC != 14)
        {
            nl();
            printf("%s %s Batch Mode",sattrName,sattrVersion);
            nl(); alarm(); nl(); reverseBlink();
            printf(" Number of command line arguments is wrong \n");
            normal(); nl();

            error = TRUE;
        }

        if (error)
        {
            printf("Usage:    sattrack -b ");
            underline();
            printf("time-zone");
            normal();
            bl();
            underline();
            printf("ground-station");
            normal();
            bl();
            underline();
            printf("satellite-id");
            normal();
            bl();
            underline();
            printf("tle-file");
            normal();
            bl();
            underline();
            printf("tle-type");
            normal();
            nl();
            printf("          ");
            underline();
            printf("prediction-type");
            normal();
            bl();
            underline();
            printf("start-date");
            normal();
            bl();
            underline();
            printf("start-time");
            normal();
            bl();
            underline();
            printf("time-step (s)");
            normal();
            bl();
            underline();
            printf("duration (d)");
            normal();
            nl();
            printf("          ");
            underline();
            printf("min-elev (deg)");
            normal();
            bl();
            printf("[no]hardcopy");
            nl(); nl();

            printf("Examples: ");
            printf("sattrack -b UTC Berkeley ");
            printf("16609 tlex.dat nasa shortpr\n");
            printf("          ");
            printf("now now auto 6 0 hardcopy\n");
            nl();
            printf("          ");
            printf("sattrack -b EST Melbourne,_FL,_USA ");
            printf("sts-55 sts.dat norad longpr\n");
            printf("          ");
            printf("24Apr93 12:0:0 20 1.5 0 nohardcopy\n");
            nl();
            exit(-1);
        }

        else
        {
            strcpy(batchTimeZone,argV[2]);
            strcpy(batchSiteName,argV[3]);
            strcpy(batchSatName,argV[4]);
            strcpy(batchTleFile,argV[5]);
            strcpy(batchTleType,argV[6]);
            strcpy(batchPredType,argV[7]);
            strcpy(batchStartDate,argV[8]);
            strcpy(batchStartTime,argV[9]);
            strcpy(batchStepTime,argV[10]);
            strcpy(batchDuration,argV[11]);
            strcpy(batchMinElev,argV[12]);
            strcpy(batchHardcopy,argV[13]);

            replaceBlanks(batchSiteName);
            replaceBlanks(batchSatName);
        }

    }

    return;
}

/******************************************************************************/
/*                                                                            */
/* getRealTime: gets UTC date and time from the system clock                  */
/*                                                                            */
/*              If SatTrack is running on one machine and the real-time       */
/*              tracking control is done by another machine, the calculations */
/*              may have to be performed ahead of time. Therefore, the time   */
/*              offset TRACKTIMEAHEAD specified in sattrack.h is added to the */
/*              system time.                                                  */
/*                                                                            */
/*              If the machine SatTrack is running on is the one that also    */
/*              performs the tracking control, the TRACKTIMEAHEAD parameter   */
/*              has to be set to 0.                                           */
/*                                                                            */
/******************************************************************************/

void getRealTime()

{
    getUnixTime(&sysDay,&sysMonth,&sysYear,&sysYearDay,
                &sysHour,&sysMin,&sysSec);

    if (trackingFlag)
        sysSec += TRACKTIMEAHEAD;

    curYearSec = (long) 
                 ((sysYearDay - 1)*86400 + sysHour*3600 + sysMin*60 + sysSec);
    dayNumber  = getDayNum(sysDay,sysMonth,sysYear);
    utcTime    = (double) (sysHour + sysMin/60.0 + sysSec/3600.0) / 24.0;
    curTime    = (double) dayNumber + utcTime;
    realTime   = curTime;

    return;
}

/******************************************************************************/
/*                                                                            */
/*  disableTrack: disables tracking                                           */
/*                                                                            */
/******************************************************************************/

void disableTrack(disableFlag)

int disableFlag;

{
    int error;

    if (disableFlag || trackingFlag)
    {
        trackingFlag  = OFF;
        trackDispFlag = TRUE;

        if (ANTENNATYPE == ISITELESCOPE)
            error = stopISI();
    }

    return;
}

/******************************************************************************/
/*                                                                            */
/*  doSleep: sleeps for a while and checks the keyboard when waking up        */
/*                                                                            */
/******************************************************************************/

void doSleep()

{
    usleep(SLEEPTIME);
    key = checkKeyboard();

    switch (key)
    {
        case 0:                                                     /* no key */
            keyBoard = VOID;
            break;

        case 10: case 13: case 27:                             /* LF, CR, ESC */
            keyBoard = EXIT;
            disableTrack(0);
            break;

        case 48:                                                         /* 0 */
            if (liveDispFlag == SINGLESAT)
            {
                trackObject        = SAT;
                newTrackObjectFlag = TRUE;
                disableTrack(0);
            }
            break;

        case 49:                                                         /* 1 */
            if (liveDispFlag == SINGLESAT)
            {
                trackObject        = SUN;
                newTrackObjectFlag = TRUE;
                disableTrack(0);
            }
            break;

        case 50:                                                         /* 2 */
            if (liveDispFlag == SINGLESAT)
            {
                trackObject        = MOON;
                newTrackObjectFlag = TRUE;
                disableTrack(0);
            }
            break;

        case 65:                                                    /* A -> a */
            key = 97;
            break;

        case 70:                                                    /* F -> f */
            key = 102;
            break;

        case 77: case 109:                                            /* M, m */
            switchSatFlag = TRUE;
            keyBoard      = MULTISAT;
            break;

        case 78:                                                    /* N -> n */
            key = 110;
            break;

        case 79:                                                    /* O -> o */
            key = 111;
            break;

        case 81: case 113:                                            /* Q, q */
            keyBoard = QUIT;
            disableTrack(0);
            break;

        case 83: case 115:                                            /* S, s */
            keyBoard = SINGLESAT;
            break;

        default:
            keyBoard = VOID;
            break;
    }

    return;
}

/******************************************************************************/
/*                                                                            */
/*  checkToggleTrack: checks if tracking control is changed (t,T)             */
/*                                                                            */
/*  the functions of the tracking flags used are:                             */
/*                                                                            */
/*  trackingFlag:  main tracking control flag                                 */
/*                 TRUE when tracking is turned on, FALSE, when off           */
/*                                                                            */
/*  trackCtrlFlag: used to deal with azimuth and elevation rates properly;    */
/*                 TRUE after first antenna control command, so that rates    */
/*                 can be calculated                                          */
/*                                                                            */
/*  trackDispFlag: used to update tracking status in live display when it is  */
/*                 has changed                                                */
/*                 TRUE until live display has been updated                   */
/*                                                                            */
/******************************************************************************/

void checkToggleTrack()

{
    int i, m, error;

    if (key == 116 || key == 84)                                      /* t, T */
    {
        if (key == 116)
            trackingFlag = (trackingFlag) ? OFF : ON;

        if (key == 84)
        {
            switch (trackingFlag)
            {
                case OFF: case ON:
                    trackingFlag = (numSatsAuto > 1) ? AUTOTRK : ON;
                    break;

                case AUTOTRK:
                    trackingFlag = ON;
                    break;

                default:
                    break;
            }

            if (trackingFlag == AUTOTRK && numSatsAuto > 1)
            {
                i = getSatListPtr(ORDER);
                m = getSatListPtr(AUTOTRK);

                if (i > m)
                {
                    sat[satOrder[m]].statusFl = SELECT;
                    sat[satOrder[i]].statusFl = MARK;
                }
            }
        }

        trackDispFlag    = TRUE;
        trackCtrlFlag    = FALSE;
        lastTrackYearSec = -1L;

        if (!trackingFlag)
            lastYearSec = -1L;

        if (liveDispFlag == SINGLESAT)
            trackInt = (trackingFlag) ? (long) TRACKINT : singleSatUpdateInt;
        else
            trackInt = (trackingFlag) ? (long) TRACKINT : multiSatUpdateInt;

        if (trackingFlag)
        {
            if (ANTENNATYPE == ISITELESCOPE)
            {
                if (didMultDispFlag)
                {
                    i = getSatListPtr(ORDER);
                    strcpy(satTrackName,sat[satOrder[i]].satellite);
                }

                else
                    strcpy(satTrackName,satName);

                error = initISI();
            }
        }

        else
            disableTrack(1);
    }

    return;
}

/******************************************************************************/
/*                                                                            */
/*  checkSwitchSat: switches satellite to be tracked (u,U,d,D,a,A,n,N)        */
/*                  or finds satellite that is currently tracked (f,F)        */
/*                  or adds/deletes satellites to/from autotrack list (l,L)   */
/*                                                                            */
/******************************************************************************/

void checkSwitchSat()

{
    int i, j, k, m, n, error;

    i = getSatListPtr(ORDER);            /* gets presently selected satellite */

    if (trackingFlag == AUTOTRK)
    {
        if ((sat[satOrder[i]].satEle > TRACKLIMIT && 
             sat[satOrder[i]].nextRiseT - curTime > TENMIN) || switchSatFlag)
        {
            m = getSatListPtr(AUTOTRK);        /* gets first marked satellite */

            if (sat[satOrder[m]].statusFl != SELECT)
            {
                sat[satOrder[i]].statusFl = MARK;
                sat[satOrder[m]].statusFl = SELECT;

                if (ANTENNATYPE == ISITELESCOPE)
                {
                    strcpy(satTrackName,sat[satOrder[m]].satellite);
                    error = initISI();
                }

                trackCtrlFlag = FALSE;
                switchSatFlag = FALSE;
            }
        }
    }

    if (key == 117 && i > 0)                                             /* u */
    {
        if (sat[satOrder[i-1]].statusFl == MARK)
            numSatsAuto--;

        sat[satOrder[i]].statusFl   = FALSE;
        sat[satOrder[i-1]].statusFl = SELECT;

        disableTrack(0);
    }

    if (key == 85)                                                       /* U */
    {
        sat[satOrder[i]].statusFl = FALSE;

        if (i > 9)
        {
            if (sat[satOrder[i-10]].statusFl == MARK)
                numSatsAuto--;

            sat[satOrder[i-10]].statusFl = SELECT;
        }

        else
        {
            if (sat[satOrder[0]].statusFl == MARK)
                numSatsAuto--;

            sat[satOrder[0]].statusFl = SELECT;
        }

        disableTrack(0);
    }

    if (key == 100 && i < numSats - 1)                                   /* d */
    {
        if (sat[satOrder[i+1]].statusFl == MARK)
            numSatsAuto--;

        sat[satOrder[i]].statusFl   = FALSE;
        sat[satOrder[i+1]].statusFl = SELECT;

        disableTrack(0);
    }

    if (key == 68)                                                       /* D */
    {
        sat[satOrder[i]].statusFl = FALSE;

        if (i < numSats - 10)
        {
            if (sat[satOrder[i+10]].statusFl == MARK)
                numSatsAuto--;

            sat[satOrder[i+10]].statusFl = SELECT;
        }

        else
        {
            if (sat[satOrder[numSats-1]].statusFl == MARK)
                numSatsAuto--;

            sat[satOrder[numSats-1]].statusFl = SELECT;
        }

        disableTrack(0);
    }

    if (key == 97)                                                       /* a */
    {
        j = getSatListPtr(ELEV);

        if (sat[satOrder[j]].statusFl == MARK)
            numSatsAuto--;

        sat[satOrder[i]].statusFl = FALSE;
        sat[satOrder[j]].statusFl = SELECT;

        disableTrack(0);
    }

    if (key == 102)                                                      /* f */
    {
        j = i;
    }

    if (key == 108  && !objectNumFlag || key == 110)                  /* l, n */
    {
        n = (xTermFlag) ? numLinesX : numLines;
        n = (xTermFlag && numSats < n - 9) ? numSats + 9 : n;

        clearLine(1,n);
        clearLine(1,n-1);

        printf(" Enter satellite name: ");
        gets(satNameStr);

        j = getSatListPtr(NAME);

        if (j >= 0)
        {
            if (key == 108)
            {
                if (sat[satOrder[j]].statusFl == SELECT)
                {
                    if (numSatsAuto > 1)
                    {
                        sat[satOrder[j]].statusFl = FALSE;
                        numSatsAuto--;

                        m = getSatListPtr(AUTOTRK);
                        sat[satOrder[m]].statusFl = SELECT;
                    }
                }

                else
                {
                    if (sat[satOrder[j]].statusFl == MARK)
                    {
                        sat[satOrder[j]].statusFl = FALSE;
                        numSatsAuto--;
                    }

                    else
                    {
                        sat[satOrder[j]].statusFl = MARK;
                        numSatsAuto++;
                    }
                }
            }

            if (key == 110)
            {
                sat[satOrder[i]].statusFl = FALSE;
                sat[satOrder[j]].statusFl = SELECT;
            }

            disableTrack(0);
        }

        else
        {
            clearLine(1,n);

            if (!strlen(satNameStr))
                printf(" no change");
            else
                printf(" not found, no change");

            fflush(stdout);
            j = i;
        }

        objectNumFlag = FALSE;
    }

    if (key == 108 && objectNumFlag || key == 111)                    /* l, o */
    {
        n = (xTermFlag) ? numLinesX : numLines;
        n = (xTermFlag && numSats < n - 9) ? numSats + 9 : n;

        clearLine(1,n);
        clearLine(1,n-1);

        printf(" Enter object number: ");
        gets(satNameStr);

        j = getSatListPtr(OBJNUM);

        if (j >= 0)
        {
            if (key == 108)
            {
                if (sat[satOrder[j]].statusFl == SELECT)
                {
                    if (numSatsAuto > 1)
                    {
                        sat[satOrder[j]].statusFl = FALSE;
                        numSatsAuto--;

                        m = getSatListPtr(AUTOTRK);
                        sat[satOrder[m]].statusFl = SELECT;
                    }
                }

                else
                {
                    if (sat[satOrder[j]].statusFl == MARK)
                    {
                        sat[satOrder[j]].statusFl = FALSE;
                        numSatsAuto--;
                    }

                    else
                    {
                        sat[satOrder[j]].statusFl = MARK;
                        numSatsAuto++;
                    }
                }
            }

            if (key == 110)
            {
                sat[satOrder[i]].statusFl = FALSE;
                sat[satOrder[j]].statusFl = SELECT;
            }

            disableTrack(0);

            clearLine(1,n);
            printf(" %ld: %s",
                sat[satOrder[j]].satIdNum,sat[satOrder[j]].satellite);
            fflush(stdout);
        }

        else
        {
            clearLine(1,n);

            if (!strlen(satNameStr))
                printf(" no change");
            else
                printf(" not found, no change");

            fflush(stdout);
            j = i;
        }

        objectNumFlag = TRUE;
    }

    if (key == 76)                                                       /* L */
    {
        for (m = 0; m < numSats; m++)
        {
            if (sat[m].statusFl == FALSE)
                sat[m].statusFl = MARK;
            else
                if (sat[m].statusFl == MARK)
                    sat[m].statusFl = FALSE;
        }

        numSatsAuto = numSats;

        disableTrack(0);
    }

    if ((key == 97 || key == 102 || key == 110 || key == 111)   /* a, f, n, o */
         && numSats > numSatLines)
    {
        k = numSatLines / 2;

        if (j > k && j < numSats - k)
            satLinePtr = j - k;

        if (j <= k)
            satLinePtr = 0;

        if (j >= numSats - k)
            satLinePtr = numSats - numSatLines;
    }

    return;
}

/******************************************************************************/
/*                                                                            */
/*  checkSwitchPage: switches page in multisat display (j,J,k,K)              */
/*                                                                            */
/******************************************************************************/

void checkSwitchPage()

{
    if (key == 106 && numSats > numSatLines)                             /* j */
    {
        satLinePtr += numSatLines;

        if (satLinePtr > numSats - numSatLines)
            satLinePtr = numSats - numSatLines;
    }

    if (key == 74 && numSats > numSatLines)                              /* J */
        satLinePtr = numSats - numSatLines;

    if (key == 107 && numSats > numSatLines)                             /* k */
    {
        satLinePtr -= numSatLines;

        if (satLinePtr < 0)
            satLinePtr = 0;
    }

    if (key == 75 && numSats > numSatLines)                              /* K */
        satLinePtr = 0;

    return;
}

/******************************************************************************/
/*                                                                            */
/*  checkExit: checks if program should be exited                             */
/*                                                                            */
/******************************************************************************/

void checkExit()

{
    if (mainChoice == SINGLESAT || mainChoice == MULTISAT)
        mainChoice = RESTART;

    switch (keyBoard) 
    {
        case QUIT:                               /* quit sattrack on Q or q   */
            system("clear");
            exit(-1);

        case SINGLESAT:                          /* single sat disp on S or s */
            mainChoice = SINGLESAT;
            break;

        case MULTISAT:                           /* multisat disp on M or m   */
            mainChoice = (firstMultiSat) ? RESTART : MULTISAT;
            break;

        default:
            break;
    }

    return;
}

/******************************************************************************/
/*                                                                            */
/*  checkKeyboard: checks keyboard entry when live display is running;        */
/*                 it returns zero if no character was entered, and the       */
/*                 ASCII code otherwise                                       */
/*                                                                            */
/******************************************************************************/

int checkKeyboard()

{
    static int structValid = FALSE;
    static struct termios original, modified;
    int    retValue;

    if (!structValid)
    {
       ioctl(STDIN_FILENO,TIOCGETA,&original);
       bcopy((char *)&original, (char *)&modified, sizeof(struct termios));

       modified.c_lflag     = 0;
       modified.c_cc[VMIN]  = 0;              /* minimum number of characters */
       modified.c_cc[VTIME] = 0;              /* do not wait */

       structValid          = TRUE;
    }

    ioctl(STDIN_FILENO,TIOCSETA,&modified);
    retValue = getc(stdin);

    if (feof(stdin))
    {
       retValue = 0;
       clearerr(stdin);
    }

    ioctl(STDIN_FILENO,TIOCSETA,&original);

    if (retValue)
    {
        (xTermFlag) ? printf("\b ") : printf(" ");
        fflush(stdout);
    }

    return(retValue);
}

/******************************************************************************/
/*                                                                            */
/* initMain: initialize main program                                          */
/*                                                                            */
/******************************************************************************/

void initMain()

{
    int error;

    lastJulianDateSun  = 0.0;
    lastJulianDatePrec = 0.0;
    lastJulianDateNute = 0.0;

    dispVersion();
    getRealTime();
    getSiderealTime(curTime);
    readCities();
    calcCityVectors();
    getTleSets();

    mainChoice    = START;
    startFlag     = TRUE;
    objectNumFlag = FALSE;

    return;
}

/******************************************************************************/
/*                                                                            */
/* initRestart: inititialize main program at restart                          */
/*                                                                            */
/******************************************************************************/

void initRestart()

{
    if (mainChoice == RESTART)
        dispVersion();

    if ((satCrashFlag || tleInvalidFlag) && liveDispFlag != MULTISAT)
    {
        alarm(); nl(); reverseBlink();

        if (satCrashFlag)
            printf(" Satellite '%s' crashed already \n",satName);

        if (tleInvalidFlag)
            printf(" TLE set for satellite '%s' is out of date \n",satName);

        normal();
    }

    mainChoice        = RESTART;
    didMultDispFlag   = FALSE;
    firstMultiSat     = TRUE;
    firstMultiSatCalc = TRUE;
    satCrashFlag      = FALSE;
    tleInvalidFlag    = FALSE;
    preLaunchFlag     = FALSE;
    preOrbitFlag      = FALSE;
    predictionFlag    = FALSE;
    writeFlag         = FALSE;
    finishPass        = FALSE;
    newRiseFlag       = FALSE;
    passDispFlag      = TRUE;
    countdownFlag     = FALSE;
    checkCountdown    = TRUE;
    sunTransitFlag    = FALSE;
    trackObject       = SAT;
    switchSatFlag     = FALSE;
    key               = TRUE;
    satLinePtr        = FALSE;
    numSatsAuto       = 1;

    return;
}

/******************************************************************************/
/*                                                                            */
/* initTerminal: initializes terminal flag and sets display update intervals  */
/*                                                                            */
/******************************************************************************/

void initTerminal()

{
    xTermFlag  = (!strcmp(strpTerm,termTypeX)) ? TRUE : FALSE;
    gndTrkFlag = (xTermFlag) ? TRUE : FALSE;

    singleSatUpdateInt = (xTermFlag) ? (long) FASTUPDATEINT : 
                                       (long) MEDUPDATEINT;

    multiSatUpdateInt  = (xTermFlag) ? (long) FASTUPDATEINT : 
                                       (long) SLOWUPDATEINT;

    if (xTermFlag && verboseFlag && startFlag)
    {
        printf("%dx%d window required\n",XDISPCOLUMNS,numLinesX);
        startFlag = FALSE;
    }

    return;
}

/******************************************************************************/
/*                                                                            */
/* showMainMenu: presents main menu options                                   */
/*                                                                            */
/******************************************************************************/

void showMainMenu()

{
    if (batchModeFlag)
    {
        predictionFlag = TRUE;
        writeFlag      = TRUE;
        mainChoice     = START;
    }

    else
    {
        predictionFlag = FALSE;
        writeFlag      = FALSE;
        mainChoice     = START;
    }

    while (mainChoice == START)
    {
        if (batchModeFlag)
            strcpy(dispStr,PX);

        else
        {
            if (quickStartFlag)
                sprintf(dispStr,"");

            else
            {
                underline(); printf("D"); normal();
                printf("isplay ");
                underline(); printf("P"); normal();
                printf("rediction ");
                underline(); printf("R"); normal();
                printf("estart ");
                underline(); printf("Q"); normal();
                printf("uit   <D> ? ");

                gets(dispStr);
                upperCase(dispStr);
            }
        }

        if (!strlen(dispStr))       mainChoice = DISPLAY;          /* default */
        if (!strncmp(dispStr,DX,1)) mainChoice = DISPLAY;
        if (!strncmp(dispStr,PX,1)) mainChoice = PREDICT;
        if (!strncmp(dispStr,TX,1)) mainChoice = TEST;
        if (!strncmp(dispStr,RX,1)) mainChoice = RESTART;
        if (!strncmp(dispStr,QX,1)) mainChoice = QUIT;

        if (mainChoice == DISPLAY)
        {
            if (quickStartFlag)
                sprintf(dispStr,"M");

            else
            {
                underline(); printf("S"); normal();
                printf("ingle or ");
                underline(); printf("M"); normal();
                printf("ultiple satellite(s)   <M> ? ");
                gets(dispStr);
                upperCase(dispStr);
            }

            if (!strncmp(dispStr,RX,1))
            {
                mainChoice = RESTART;
                break;
            }

            if (!strncmp(dispStr,QX,1))
                exit(-1);

            mainChoice = (!strlen(dispStr) || 
                          !strncmp(dispStr,MX,1)) ? MULTISAT : SINGLESAT;
        }

        if (mainChoice == PREDICT)
        {
            transitFlag = FALSE;

            if (batchModeFlag)
            {
                if (!strcmp(batchPredType,LONGPR) || 
                    !strcmp(batchPredType,TRANSIT))
                    strcpy(dispStr,LX);
                else
                    strcpy(dispStr,SX);

                transitFlag = (!strcmp(batchPredType,TRANSIT)) ? TRUE : FALSE;
            }

            else
            {
                underline(); printf("S"); normal();
                printf("hort or ");
                underline(); printf("L"); normal();
                printf("ong prediction format   <S> ? ");
                gets(dispStr);
                upperCase(dispStr);

                if (!strncmp(dispStr,RX,1))
                {
                    mainChoice = RESTART;
                    break;
                }

                if (!strncmp(dispStr,QX,1))
                    exit(-1);
            }

            shortPredFlag = (!strlen(dispStr) || 
                             !strncmp(dispStr,SX,1)) ? TRUE : FALSE;

            if (!shortPredFlag)
            {
                if (batchModeFlag)
                    sunTransitFlag = FALSE;

                else
                {

#ifdef SUNTRANSITS
                    printf("Calculate solar transits          <N> ? ");
                    gets(dispStr);
#else
                    sprintf(dispStr,"");
#endif

                    upperCase(dispStr);

                    if (!strncmp(dispStr,RX,1))
                    {
                        mainChoice = RESTART;
                        break;
                    }

                    if (!strncmp(dispStr,QX,1))
                        exit(-1);

                    sunTransitFlag = (!strlen(dispStr) || 
                                      !strncmp(dispStr,NO,1)) ? FALSE : TRUE;
                }

                if (sunTransitFlag)
                {
                    printf("Enter approach limit [arcmin]  <%4.0f> : ",
                        SUNPROX);
                    gets(dispStr);

                    sunProxLimit = (!strlen(dispStr)) ? SUNPROX : atof(dispStr);
                    sunProxLimit *= CAMR;
                }

                if (transitFlag)
                {
                    sunProxLimit   = SUNPROX * CAMR;
                    sunTransitFlag = TRUE;
                }
            }

            if (batchModeFlag)
            {
                predictionFlag = TRUE;
                writeFlag      = TRUE;
            }

            else
            {
                printf("Output on ");
                underline(); printf("V"); normal();
                printf("ideo terminal or ");
                underline(); printf("F"); normal();
                printf("ile  <V> ? ");
                gets(dispStr);
                upperCase(dispStr);

                if (!strncmp(dispStr,RX,1))
                {
                    mainChoice = RESTART;
                    break;
                }

                if (!strncmp(dispStr,QX,1))
                    exit(-1);

                predictionFlag = (!strlen(dispStr) || 
                                  !strncmp(dispStr,VX,1)) ? TRUE : FALSE;
                writeFlag      = (!strncmp(dispStr,FX,1)) ? TRUE : FALSE;
            }
        }
    }

    return;
}

/******************************************************************************/
/*                                                                            */
/* getSatListPtr: finds pointer to the currently active satellite             */
/*                                                                            */
/******************************************************************************/

int getSatListPtr(ptrType)

int ptrType;

{
    long ll;
    int  i, j, k;

    i = 0;

    if (ptrType == LIST)
    {
        foundSatFlag = TRUE;

        while (sat[i].statusFl != SELECT && i < numSats)
            i++;
    }

    if (ptrType == ORDER)
    {
        foundSatFlag = TRUE;

        while (sat[satOrder[i]].statusFl != SELECT && i < numSats)
            i++;
    }

    if (ptrType == ELEV)
    {
        foundSatFlag = TRUE;

        while (sat[satOrder[i]].satEle > ZERO && i < numSats)
            i++;
    }

    if (ptrType == NAME)
    {
        foundSatFlag = FALSE;
        upperCase(satNameStr);
        k = strlen(satNameStr);
        j = 0;

        while (k > 0 && j < numSats)
        {
            strcpy(str,sat[satOrder[j]].satellite);
            upperCase(str);

            if (!strncmp(str,satNameStr,k))
            {
                foundSatFlag = TRUE;
                i = j;
                break;
            }

            j++;
        }
    }

    if (ptrType == OBJNUM)
    {
        foundSatFlag = FALSE;
        ll = atol(satNameStr);
        k  = strlen(satNameStr);
        j  = 0;

        while (k > 0 && j < numSats)
        {
            if (ll == sat[satOrder[j]].satIdNum)
            {
                foundSatFlag = TRUE;
                i = j;
                break;
            }

            j++;
        }
    }

    if (ptrType == AUTOTRK)
    {
        foundSatFlag = FALSE;
        j = 0;

        while (j < numSats)
        {
            if (sat[satOrder[j]].statusFl == MARK || 
                sat[satOrder[j]].statusFl == SELECT)
            {
                foundSatFlag = TRUE;
                i = j;
                break;
            }

            j++;
        }
    }

    i = (foundSatFlag) ? i : -1;
    return(i);
}

/******************************************************************************/
/*                                                                            */
/* initSingleSatTrackLoop: initializes tracking loop parameters for a single  */
/*                         satellite                                          */
/*                                                                            */
/******************************************************************************/

void initSingleSatTrackLoop()

{
    int ns;

    liveDispFlag  = SINGLESAT;
    trackObject   = SAT;
    prevDispEle   = -TWOPI;
    prevSatLat    = -TWOPI;
    lastYearSec   = -1L;
    oldCityNum    = -1;

    trackInt = (trackingFlag) ? (long) TRACKINT : singleSatUpdateInt;

    if (firstMultiSat)
    {
        nextRiseTime = ZERO;
        nextSetTime  = ZERO;
        riseAzimuth  = ZERO;
        maxAzimuth   = ZERO;
        setAzimuth   = ZERO;
        maxRange     = ZERO;
        newRiseFlag  = FALSE;
        passDispFlag = FALSE;
    }

    else
    {
        ns = getSatListPtr(LIST);
        loadSatData(ns);
        newRiseFlag  = TRUE;
        passDispFlag = TRUE;
    }

    checkPropModelType();

    return;
}

/******************************************************************************/
/*                                                                            */
/* doSingleSatTrackLoop: performs tracking loop with detailed live display    */
/*                       for a single satellite                               */
/*                                                                            */
/******************************************************************************/

void doSingleSatTrackLoop()

{
    if (reinitSingleSat)
    {
        initSingleSatTrackLoop();
        initSingleSatLiveDisp();
        reinitSingleSat = FALSE;
    }

    do
    {
        getRealTime();

        if (curYearSec >= lastYearSec + trackInt && curYearSec % trackInt == 0L)
        {
            doLiveCalcs();
            trackingControl();

            lastYearSec    = curYearSec;
            dispUpdateFlag = TRUE;
        }

        if (curYearSec % singleSatUpdateInt == 0L && dispUpdateFlag)
        {
            updateSingleSatLiveDisp(curTime,nextRiseTime,nextSetTime,
                riseAzimuth,maxAzimuth,setAzimuth,maxElevation,maxRange);

            dispUpdateFlag = FALSE;
        }

        doSleep();
        checkToggleTrack();
    }
    while (keyBoard == VOID && !satCrashFlag && !tleInvalidFlag);

    return;
}

/******************************************************************************/
/*                                                                            */
/* initMultiSatTrackLoop: initializes tracking loop parameters for multiple   */
/*                        satellites                                          */
/*                                                                            */
/******************************************************************************/

void initMultiSatTrackLoop()

{
    int i;

    liveDispFlag = MULTISAT;
    trackObject  = SAT;
    lastYearSec  = -1L;

    trackInt = (trackingFlag) ? (long) TRACKINT : multiSatUpdateInt;

    if (firstMultiSat)
    {
        newPassFlag = FALSE;

        for (i = 0; i < numSats; i++)
        {
            sat[i].satLastEle = -PI;
            sat[i].nextRiseT  = ZERO;
            sat[i].nextSetT   = ZERO;
            sat[i].crashFl    = FALSE;
            sat[i].invalidFl  = FALSE;
            sat[i].nopassFl   = FALSE;
        }
    }

    return;
}

/******************************************************************************/
/*                                                                            */
/* doMultiSatTrackLoop: performs tracking loop with live display for multiple */
/*                      satellites                                            */
/*                                                                            */
/******************************************************************************/

void doMultiSatTrackLoop()

{
    if (firstMultiSat)
    {
        readSatList();
        loadSatElements();
    }

    if (reinitMultiSat)
    {
        initMultiSatTrackLoop();
        initMultiSatLiveDisp();
        reinitMultiSat = FALSE;
    }

    do
    {
        getRealTime();

        if (firstMultiSat || 
           (curYearSec >= lastYearSec + trackInt &&
            curYearSec % trackInt == 0L))
        {
            for (ns = 0; ns < numSats; ns++)
            {
                if (firstMultiSat || 
                    (curYearSec % multiSatUpdateInt == 0L) ||
                    (curYearSec % multiSatUpdateInt != 0L && 
                     sat[ns].statusFl == SELECT && trackingFlag))
                {

                    if (firstMultiSat)
                    {
                        if (firstMultiSatCalc)
                        {
                            gotoXY(18,7);
                            printf("calculating next pass for satellite #");
                            firstMultiSatCalc = FALSE;
                        }

                        gotoXY(55,7);

                        if (xTermFlag)
                            printf("%-3d: %-14s",ns+1,sat[ns].satellite);
                        else
                            printf("%-d",ns+1);

                        if (ns+1 == numSats)
                            printf("  ---  standing by ");

                        fflush(stdout);
                    }

                    loadSatData(ns);
                    doLiveCalcs();

                    if (sat[ns].statusFl == SELECT)
                        trackingControl();

                    saveSatData(ns);
                }
            }

            firstMultiSat  = FALSE;
            dispUpdateFlag = TRUE;
            lastYearSec    = curYearSec;
        }

        if (curYearSec % multiSatUpdateInt == 0L && dispUpdateFlag)
        {
            updateMultiSatLiveDisp(curTime);
            dispUpdateFlag = FALSE;
        }

        doSleep();
        checkToggleTrack();
        checkSwitchSat();
        checkSwitchPage();
    }
    while (keyBoard == VOID);

    return;
}

/******************************************************************************/
/*                                                                            */
/* doLiveCalcs: performs the calculations for the live display loops          */
/*                                                                            */
/******************************************************************************/

void doLiveCalcs()

{
    if (curTime > nextSetTime && geoSyncFlag != GEOSTAT)
    {
        predTime     = (elsetEpoch > curTime) ? elsetEpoch : curTime;
        predStopTime = predTime + (stopTime - startTime);

        getNextPass();
        newPassFlag = TRUE;
    }

    preLaunchFlag = (launchEpoch > curTime) ? TRUE : FALSE;
    preOrbitFlag  = (elsetEpoch  > curTime) ? TRUE : FALSE;

    getSiderealTime(curTime);
    getSiteVector();

    getNoradStateVector(curTime);
    getStateVector(curTime);

    getAziElev(SAT);
    getRange();
    getShuttleOrbit();
    getSunVector(1);
    getAziElev(SUN);
    getAziElev(MOON);
    getSubSatPoint();
    getDoppler();
    getPathLoss();
    getSquintAngle();
    getPhase();
    getMode();

    eclipseCode = satEclipse(satElevation);
    return;
}

/******************************************************************************/
/*                                                                            */
/* saveSatData: saves satellite data in a structure                           */
/*                                                                            */
/******************************************************************************/

void saveSatData(satPtr)

int satPtr;

{
    if (gndTrkFlag)
        getGroundTrack(satLat,satLong,&gndTrkDist,gndTrkDir,gndTrkCity);

    else
    {
        gndTrkDist = ZERO;
        sprintf(gndTrkDir,"");
        sprintf(gndTrkCity,"");
    }

    sat[satPtr].satOrbNum   = orbitNum;
    sat[satPtr].stsOrbNum   = stsOrbitNum;

    sat[satPtr].satOrb      = curOrbit;
    sat[satPtr].stsOrb      = stsOrbit;
    sat[satPtr].satOrbFrac  = orbitFract;
    sat[satPtr].stsOrbFrac  = stsOrbitFract;

    sat[satPtr].satAzi      = satAzimuth;
    sat[satPtr].satLastEle  = sat[satPtr].satEle;
    sat[satPtr].satEle      = satElevation;
    sat[satPtr].satRng      = satRange;
    sat[satPtr].satLtd      = satLat;
    sat[satPtr].satLng      = satLong;
    sat[satPtr].satHgt      = satHeight;

    sat[satPtr].stsFl       = stsFlag;
    sat[satPtr].launchFl    = launchFlag;
    sat[satPtr].prelaunchFl = preLaunchFlag;
    sat[satPtr].preorbitFl  = preOrbitFlag;
    sat[satPtr].geosyncFl   = geoSyncFlag;
    sat[satPtr].crashFl     = satCrashFlag;
    sat[satPtr].invalidFl   = tleInvalidFlag;

    sat[satPtr].eclCode     = eclipseCode;
    sat[satPtr].ctyDist     = gndTrkDist;

    strcpy(sat[satPtr].ctyDir,gndTrkDir);
    strcpy(sat[satPtr].ctyStr,gndTrkCity);

    sat[satPtr].semiMajAxis = semiMajorAxis;;
    sat[satPtr].raanPr      = raanPrec;
    sat[satPtr].periPr      = perigeePrec;

    if (newPassFlag)
    {
        sat[satPtr].nextRiseT = nextRiseTime;
        sat[satPtr].nextSetT  = nextSetTime;
        sat[satPtr].riseAzi   = riseAzimuth;
        sat[satPtr].maxAzi    = maxAzimuth;
        sat[satPtr].setAzi    = setAzimuth;
        sat[satPtr].maxEle    = maxElevation;
        sat[satPtr].maxRng    = maxRange;
        sat[satPtr].eclRise   = eclipseRise;
        sat[satPtr].eclMax    = eclipseMax;
        sat[satPtr].eclSet    = eclipseSet;
        sat[satPtr].nopassFl  = noPassFlag;
    }

    if (curTime > sat[satPtr].nextRiseT && 
        fabs(sat[satPtr].nextSetT - curTime) < HALFMIN &&
        sat[satPtr].satEle < ZERO)
        sat[satPtr].satEle = ONEPPM;

    if (sat[satPtr].nopassFl && newPassFlag)
    {
        sat[satPtr].nextRiseT = curTime + maxDays;
        sat[satPtr].nextSetT  = curTime + maxDays;
    }

    newPassFlag = FALSE;

    if (sat[satPtr].geosyncFl == GEOSTAT && sat[satPtr].satEle <= ZERO)
        sat[satPtr].nextRiseT = ONEMEG + ONEPPM;

    if (sat[satPtr].invalidFl)
        sat[satPtr].nextRiseT = ONEMEG + TWOPPM;

    if (sat[satPtr].crashFl)
        sat[satPtr].nextRiseT = ONEMEG + ONEPPM + TWOPPM;

    if (sat[satPtr].geosyncFl == GEOSTAT && sat[satPtr].satEle > ZERO)
        sat[satPtr].nextSetT  = TWOMEG;

    return;
}

/******************************************************************************/
/*                                                                            */
/* loadSatData: loads satellite data from a structure                         */
/*                                                                            */
/******************************************************************************/

void loadSatData(satPtr)

int satPtr;

{
    int i;

    strcpy(satName,sat[satPtr].satellite);
    strcpy(satAlias,sat[satPtr].satelliteAlias);
    strcpy(propModel,sat[satPtr].model);

    satNum           = sat[satPtr].satIdNum;

    epochDay         = sat[satPtr].epDay;
    elsetEpoch       = sat[satPtr].elSetEp;
    launchEpoch      = sat[satPtr].launchEp;

    inclination      = sat[satPtr].epInc;
    epochRaan        = sat[satPtr].epRaan;
    eccentricity     = sat[satPtr].epEcc;
    epochArgPerigee  = sat[satPtr].epArgPer;
    epochMeanAnomaly = sat[satPtr].epMeanAnom;
    epochMeanMotion  = sat[satPtr].epMeanMot;
    decayRate        = sat[satPtr].epDecRate;
    decayRateDot     = sat[satPtr].epDecRateDot;
    bStarCoeff       = sat[satPtr].epBstar;
    epochOrbitNum    = sat[satPtr].epOrbNum;
    ephemerisType    = sat[satPtr].ephType;
    propModelType    = sat[satPtr].modelType;

    beaconFreq       = sat[satPtr].freq;
    perigeePhase     = sat[satPtr].perigPh;
    maxPhase         = sat[satPtr].maxPh;
    numModes         = sat[satPtr].numMd;
    attitudeFlag     = sat[satPtr].attitudeFl;

    if (attitudeFlag)
    {
        attLat       = sat[satPtr].attLtd;
        attLong      = sat[satPtr].attLng;
        satAttGl[0]  = sat[satPtr].attX;
        satAttGl[1]  = sat[satPtr].attY;
        satAttGl[2]  = sat[satPtr].attZ;
    }

    refOrbit         = sat[satPtr].refOrb;
    semiMajorAxis    = sat[satPtr].semiMajAxis;
    raanPrec         = sat[satPtr].raanPr;
    perigeePrec      = sat[satPtr].periPr;

    cosInclination   = sat[satPtr].cosInc;
    sinInclination   = sat[satPtr].sinInc;

    nextRiseTime     = sat[satPtr].nextRiseT;
    nextSetTime      = sat[satPtr].nextSetT;
    riseAzimuth      = sat[satPtr].riseAzi;
    maxAzimuth       = sat[satPtr].maxAzi;
    setAzimuth       = sat[satPtr].setAzi;
    maxElevation     = sat[satPtr].maxEle;
    maxRange         = sat[satPtr].maxRng;
    eclipseRise      = sat[satPtr].eclRise;
    eclipseMax       = sat[satPtr].eclMax;
    eclipseSet       = sat[satPtr].eclSet;

    stsFlag          = sat[satPtr].stsFl;
    launchFlag       = sat[satPtr].launchFl;
    preLaunchFlag    = sat[satPtr].prelaunchFl;
    preOrbitFlag     = sat[satPtr].preorbitFl;
    geoSyncFlag      = sat[satPtr].geosyncFl;
    satCrashFlag     = sat[satPtr].crashFl;
    tleInvalidFlag   = sat[satPtr].invalidFl;

    if (numModes)
    {
        for (i = 0; i < numModes; i++)
        {
            strcpy(modes[i].modeStr,sat[satPtr].satModes[i].modeStr);
            modes[i].minPhase = sat[satPtr].satModes[i].minPhase; 
            modes[i].maxPhase = sat[satPtr].satModes[i].maxPhase;;
        }
    }

    return;
}

/******************************************************************************/
/*                                                                            */
/* main program sattrack                                                      */
/*                                                                            */
/******************************************************************************/

main(argc,argv)

int  argc;
char *argv[];

{
    getEnvironment();
    checkArguments(argc,argv);
    initMain();

    while(TRUE)                                          /* main program loop */
    {
        if (mainChoice == START || mainChoice == PREDICT || 
            mainChoice == TEST  || mainChoice == RESTART)
        {
            initRestart();
            getRealTime();
            getDefaultData();
            initTerminal();
        }

        while (mainChoice == RESTART)
        {
            checkTimeZone();
            getSiteParams();
            getSatParams();
            getSatModes(satName);
            getSatPrecession();
            showMainMenu();
        }

        switch (mainChoice)
        {
            case SINGLESAT:
                reinitSingleSat = TRUE;
                doSingleSatTrackLoop();
                break;

            case MULTISAT:
                reinitMultiSat = TRUE;
                doMultiSatTrackLoop();
                break;

            case PREDICT:
                doPrediction();
                break;

            case TEST:
                doNoradTest();
                break;

            case RESTART:
                break;

            case QUIT:
                exit(-1);

            default:
                break;
        }

        checkExit();
    }
}

/******************************************************************************/
/*                                                                            */
/* end of program sattrack.c                                                  */
/*                                                                            */
/******************************************************************************/
